import os
import gc
import sys
import math
import hashlib
import asyncio
import pickle
import urllib3
import threading
import getpass
from abc import ABC
from time import sleep
import socketserver
import multiprocessing.pool
import multiprocessing as mp
import concurrent.futures
from bs4 import BeautifulSoup
from multiprocessing import Process
from sqlalchemy.pool import NullPool
from random import randint, randrange
from datetime import datetime, timedelta
from core_framework.crawlers import CrawlerBase
from core_framework.db_engine import DbEngine
from crawlers_sm.croatia.eoglasna.parser import *
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

from Scripts import filepath
storage_location = filepath.set_storage()
print("base storage_location",  storage_location)

crawlers_sm = 9
# crawlers_sm = 1


os.environ['NLS_LANG'] = 'CROATIAN_CROATIA.UTF8'
api_key = '7c315478c67a270a2d829f1a00f9096c'
site_key = '6LcJFLcUAAAAAAWb8bmUnRGwe9RBS4ADwlu2-VBp'

optimization_status = 'not optimized'
schema = 'eoglasna'
base = 'https://e-oglasna.pravosudje.hr'
objave_url = '/'.join([base, 'objave-sudova'])
court_list_url = '/'.join([objave_url, 'pregled-po-sudovima'])
captcha_api = f'http://2captcha.com/in.php?key={api_key}&method=userrecaptcha&googlekey={site_key}&pageurl={objave_url}'

# download  u upload dokumenata lokacije
# main_doc_path = 'E:\TuzbeDocs\\'
# main_doc_path = r'C:\Users\{}\Documents\eoglasna'.format(getpass.getuser())
# main_doc_path = r'E:\WebDocs\eoglasna'
main_doc_path = os.path.join(storage_location, 'eoglasna')

today = datetime.today().strftime('%Y%m%d')
current_hour = datetime.now().hour
now = datetime.now()
short_date = datetime(now.year, now.month, now.day, 5, 0)
short_date2 = datetime(now.year, now.month, now.day, 0, 0)

two_days = now.day-2
if two_days == 0:
    two_days = 1

short_date_rss = datetime(now.year, now.month, two_days, 5, 0)

# server_doc_path = os.path.join('Z:\\', 'oglastrgsud')
server_doc_path = os.path.join('//10.76.15.120', 'oglastrgsud')
xlsx_tables = os.path.join(main_doc_path, 'xlsx')
local_dir_path = os.path.join(main_doc_path, 'download')

fina_local_path = os.path.join(local_dir_path, str(99999))
fina_server_path = os.path.join(server_doc_path, 'nova', str(99999))


local_dir_path_day = os.path.join(main_doc_path, 'downloaddaily')
local_subdir_path_day = os.path.join(local_dir_path_day, today)

# directory is used for storing html that resulted in some kind of exception or html could not be parsed
parser_dir_errors = os.path.join(main_doc_path, 'parser_errors')

# directory is used save html that server returned
htmls_dir = os.path.join(main_doc_path, 'eoglasna_htmls')
htmls_subdir = os.path.join(htmls_dir, today)

detalji_dir = os.path.join(main_doc_path, 'eoglasna_detalji')
detalji_subdir = os.path.join(detalji_dir, today)

# search data for specified period of time, we look only for current day
date_to = datetime.now().strftime('%Y-%m-%d')
date_from = datetime.now().strftime('%Y-%m-%d')

lock = threading.Lock()


class OracleDbbisnode(DbEngine):
    def __init__(self):
        DbEngine.__init__(self)
        self.connect(4)

# for versions python < 3.6
# class NoDaemonProcess(mp.Process):
#     def _get_daemon(self):
#         return False
#
#     def _set_daemon(self, value):
#         pass
#     daemon = property(_get_daemon, _set_daemon)
#
#
# class MyPool(multiprocessing.pool.Pool):
#     Process = NoDaemonProcess


# for versions python 3.8+
class NoDaemonProcess(mp.Process):
    @property
    def daemon(self):
        return False

    @daemon.setter
    def daemon(self, value):
        pass


class NoDaemonContext(type(multiprocessing.get_context())):
    Process = NoDaemonProcess


class MyPool(multiprocessing.pool.Pool):
    def __init__(self, *args, **kwargs):
        kwargs['context'] = NoDaemonContext()
        super(MyPool, self).__init__(*args, **kwargs)


class Eoglasna(CrawlerBase, ABC):
    def __init__(self, recheck=None):
        CrawlerBase.__init__(self, base, proxy_type=2)
        self.dbbisnode = OracleDbbisnode()

        self.workers = 5
        self.courts = dict()
        self.court_groups = dict()
        self.flag_120 = False
        self.exit = False  # if this is True then program will exit because something went wrong during preload of app

        # runtime log initialization
        self.runtime_errors = 0
        self.runtime_status = 1
        self.runtime_start = datetime.now()
        self.runtime_id = hashlib.sha3_256(str(self.runtime_start).encode()).hexdigest()
        self.runtime_log = {'start_time': self.runtime_start, 'step_name': 'runtime'}

        self.crawler_errors = {}
        self.crawler_err_cnt = 0
        self.full_dl = False

        self.tokens = {}
        self.all_done = False
        self.day_download = True
        self.log_option = False
        self.daily_recaptcha = datetime(now.year, now.month, now.day-1)

    def crawler_error_combine(self, err_data, log_common, sha_data=None):
        self.crawler_err_cnt += 1
        log_copy = log_common.copy()
        log_copy.update(err_data)
        if sha_data is None:
            sha_data = err_data
        err_sha = hashlib.sha3_256(str(sha_data).encode()).hexdigest()
        err_sha_data = self.crawler_errors.get(err_sha)
        if err_sha_data is None:
            err_data.update({'err_cnt': 1})
            self.crawler_errors.update({err_sha: log_copy})
        else:
            err_cnt = err_sha_data.get('err_cnt')
            if err_cnt is None:
                err_cnt = 0
            err_cnt += 1
            err_sha_data.update({'err_cnt': err_cnt})

    def info_log(self, data):
        print("info_log", data)
        if self.log_option is True:
            sha = hashlib.sha3_256(str(data).encode()).hexdigest()
            data.update({'sha': sha, 'runtime_id': self.runtime_id, 'end_time': datetime.now()})
            self.insert('l_info', {0: data}, schema=schema)

    def error_log(self, data):
        print("error_log", data)
        self.runtime_status = 2
        self.runtime_errors += 1
        if self.log_option is True:
            sha = hashlib.sha3_256(str(data).encode()).hexdigest()
            data.update({'sha': sha, 'runtime_id': self.runtime_id, 'end_time': datetime.now()})
            self.insert('l_errors', {0: data}, schema=schema)

    def postgresql_conn(self):
        self.connect(crawlers_sm, connect_args={"application_name": "eoglasna/crawler.py"})

    def error_combine(self, log, data):
        log.update(data)
        return log

    def gather_courts(self):
        sha = None
        proc_id = randint(1, 20000)
        start_time = datetime.now()
        step_name = 'gather_courts'
        log_common = {'start_time': start_time, 'step_name': step_name}

        print(court_list_url)
        print('> gathering court list')
        try:
            sleep(randint(2, 5))
            sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)
            r = self.requests.get(proc_id)
            r.request_type = 1
            r.ses.verify = False

            r.go(court_list_url)

            html_bs = r.response.content
            if type(html_bs) is dict:
                err_desc = html_bs.get('RequestError')
                error = {'err_desc': err_desc}
                self.crawler_error_combine(error, log_common)
                return 400

            print('> parsing court list')
            court_list = court_list_parse(html_bs)
            if type(court_list) is dict:
                self.crawler_error_combine(court_list, log_common)
                return 400

            court_list = dict(enumerate(court_list))
            self.merge('court_list', court_list, filters={}, update=False, schema=schema)
            new_courts = self.select('court_list', filters={'court_id': None}, schema=schema)
            ora_courts = {row.get('naziv'): row.get('sud_id') for row in self.dbbisnode.select('SUD', schema='TUZBE')}
            print('> assigning court ids')
            for new_court in new_courts:
                court_name = new_court.get('court_name')
                row_sha = new_court.get('sha')
                if court_name is not None:
                    court_id = ora_courts.get(court_name)
                    self.update('court_list', {'sha': row_sha, 'court_id': None}, {'court_id': court_id}, schema=schema)
            return 200

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)}
            self.crawler_error_combine(error, log_common)
            return 400
        finally:
            # release proxy back in use if it is not released it will be not be available for that web_base next 30min
            self.release_proxy(sha)

    # =======================================================
    #                     Preload functions
    # =======================================================

    def top_last_cases(self):
        cases = self.select('search_results', filters={'date_created': f'>={short_date}'}, schema=schema)
        if not cases and now.hour < 19:
            self.full_dl = True

    def db_court_list(self):
        self.courts = {row.get('court_name'): row.get('court_id') for row in self.select('court_list', filters={'court_id': True}, schema=schema)}
        self.court_groups = {row.get('court_name'): row.get('court_group') for row in self.select('court_list', filters={'court_id': True}, schema=schema)}

    def dir_creation_check(self):
        start_time = datetime.now()
        step_name = 'dir_creation_check'
        print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}
        try:
            dirs = [main_doc_path, local_dir_path, local_dir_path_day, local_subdir_path_day,
                    parser_dir_errors, htmls_dir, htmls_subdir, detalji_dir, detalji_subdir, xlsx_tables,
                    fina_local_path]
            dirs.extend([os.path.join(local_dir_path, str(v)) for k,v in self.courts.items()])

            for dir_path in dirs:
                if os.path.isdir(dir_path) is False:
                    os.mkdir(dir_path)
            log_common.update({'status': 1, 'errors': 0})
            self.info_log(log_common)
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            log_common.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
            self.error_log(log_common)
            self.exit = True

    # =======================================================
    #                     Crawling functions
    # =======================================================
    def get_token(self, counter):
        proc_id = randint(1, 20000)
        step_name = 'get_token'
        start_time = datetime.now()
        log_common = {'start_time': start_time, 'step_name': step_name, 'gtoken': None, 'session_id': None}

        if len(self.tokens) > 4 or self.all_done is True:
            return 200

        try:
            for i in range(0, 10):
                sha = self.new_request(proc_id=proc_id, protocols=['http'], new=True)
                self.release_proxy(sha)
                r = self.requests.get(proc_id)
                r.request_type = 1
                r.go(captcha_api)
                html = r.response.content

                if type(html) is dict:
                    error = html.get('RequestError')
                    log_copy = log_common.copy()
                    log_copy.update({'err_desc': error, 'status': 2})
                    # self.token_log(log_copy)
                    continue
                else:
                    log_copy = log_common.copy()
                    id_search = re.search('(\d+)', html.decode())
                    if id_search is None:
                        error = {'err_desc': f'id_search is not given by 2captcha. RESPONSE: {html}'}
                        # print(error)
                        # log_copy.update(error)
                        # self.token_log(log_copy)
                        continue

                    request_id = id_search.group(1)
                    log_copy.update({'session_id': request_id})
                    sleep(randrange(15, 20))

                    # get results
                    for x in range(0, 20):
                        if self.all_done is True:
                            return 200

                        request_check = f"http://2captcha.com/res.php?key={api_key}&action=get&id={request_id}"
                        r.go(request_check)
                        html_bs = r.response.content

                        if type(html_bs) is dict:
                            err_desc = html_bs.get('RequestError')
                            # print(err_desc)
                            sleep(randrange(10, 15))
                            continue

                        html = html_bs.text

                        if html == 'CAPCHA_NOT_READY':
                            error = {'err_desc': 'Your captcha is not solved yet. Make 5 seconds timeout and repeat your request.'}
                            # print(error)
                            sleep(randrange(15, 20))
                            continue

                        if html == 'ERROR_CAPTCHA_UNSOLVABLE':
                            error = {'err_desc': '''We are unable to solve your captcha - three of our workers were unable solve it or we didnt get an answer within 90 seconds (300 seconds for ReCaptcha V2). We will not charge you for that request'''}
                            # print(error)
                            break

                        # major account errors
                        if html in ['ERROR_WRONG_USER_KEY', 'ERROR_KEY_DOES_NOT_EXIST']:
                            error = {'err_desc': 'Stop sending requests. Check your API key.'}
                            # print(error)
                            return 400

                        # major recaptcha errors
                        if html in ['ERROR_WRONG_ID_FORMAT', 'ERROR_WRONG_CAPTCHA_ID']:
                            error = {'err_desc': 'Check the ID of captcha or your code that gets the ID.'}
                            # print(error)
                            return 400

                        # minor errors
                        if html in ['ERROR_BAD_DUPLICATES', 'REPORT_NOT_RECORDED']:
                            error = {'err_desc': 'You can retry to send your captcha again.'}
                            # print(error)
                            break

                        # final results
                        if html.startswith('OK'):
                            html = html.replace('OK|','').strip()
                        else:
                            print("unhandled errors", html)
                            return 400
                        self.tokens.update({request_id: html})
                        return (html, request_id)
            return 400

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            log_copy = log_common.copy()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            log_copy.update(error)
            print(error)
            # self.token_log(log_copy)
        # finally:
            # release proxy back in use if it is not released it will be not be available for that web_base next 30min
            # self.release_proxy(sha)

    def fetch_tokens(self):
        while 1:
            if self.all_done is True:
                return
            if len(self.tokens) < 5:
                self.workers = 5
                pages = list(range(0, 100))
                with concurrent.futures.ThreadPoolExecutor(max_workers=self.workers) as executor:
                    # Start the load operations and mark each future with its URL
                    future_to_url = {executor.submit(self.get_token, link): link for link in pages}
                    for future in concurrent.futures.as_completed(future_to_url):
                        url = future_to_url[future]
                        try:
                            return_data = future.result()
                        except AttributeError as e:
                            pass
                        except Exception as e:
                            exc_type, exc_obj, exc_tb = sys.exc_info()
                            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)}
                            print(f'fetch_tokens error {error}')
            else:
                sleep(2)

    def token_server(self):
        thread = threading.Thread(target=self.fetch_tokens)
        thread.start()

    def forward_token(self):
        while True:
            token_size = len(self.tokens)
            # sys.stdout.write(f'\r tokens available:  {len(self.tokens)}')
            if token_size > 0:
                break
            sleep(1)

        lock.acquire()
        request_id = list(self.tokens.keys())[0]
        token = self.tokens.get(request_id)
        token_data = (request_id, token)
        self.tokens.pop(request_id)
        lock.release()
        return token_data

    @staticmethod
    def record_html(html, detalji=False):
        try:
            sha = hashlib.sha3_256(str(html).encode()).hexdigest()
            if detalji is False:
                doc_loc = '/'.join([htmls_subdir, f'{sha}.html'])
            else:
                doc_loc = '/'.join([detalji_subdir, f'{sha}.html'])
            with open(doc_loc, 'w') as f:
                f.write(str(html))
            return 200
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error_desc = f"Could not write html: reason {str(e)}"
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': error_desc}
            print(error)
            return error

    def gather_search_results(self):
        proc_id = 1
        max_workers = 3
        error_counter = 0
        start_time = datetime.now()
        step_name = 'gather_search_results'
        print(f"\n> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}

        # get results from first page
        vrste_objava = {1: 'noticeSourceType=COURT', 2: 'noticeSourceType=NOTARY_PUBLIC'}
        # vrste_objava = {2: 'noticeSourceType=NOTARY_PUBLIC'}
        for rid, vrsta_objave in vrste_objava.items():
            request_id, g_token = self.forward_token()
            sha = None
            html = ''
            last_page = 1
            total_records = 0
            no_results = None

            g_url = f'https://e-oglasna.pravosudje.hr/objave-sudova?{vrsta_objave}&datePublishedFrom={date_from}&datePublishedTo={date_to}&_g={g_token}'
            try:
                sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)
                r = self.requests.get(proc_id)
                r.request_type = 2
                r.ses.verify = False

                print(f'proxy {r.ip},{r.port}')
                r.go(g_url)
                html = r.response.content
                html_raw = r.response.content_raw
                if type(html) is dict:
                    err_desc = html.get('RequestError')
                    error = f'rid: {rid} desc: {err_desc} '
                    print(error)
                else:
                    # get last page
                    last_page_re = html.find('a', {'title': 'Poslijednja stranica', 'href': True})
                    if last_page_re is not None:
                        last_page_link = last_page_re['href']
                        last_page_n = re.search('page=(\d+)', last_page_link, re.DOTALL | re.MULTILINE | re.IGNORECASE)
                        if last_page_n is not None:
                            last_page = int(last_page_n.group(1))

                    # get total number of court cases specified for current rid search
                    rows_search = re.search('Pronađeno rezultata:\s+<strong>(\d+)</strong>', html.decode(), re.DOTALL | re.MULTILINE | re.IGNORECASE)
                    if rows_search is not None:
                        rows_search = rows_search.group(1)
                        total_records = rows_search

                    # check is there any data
                    no_results = html.find('em', text=re.compile('Nema rezultata.'))

            except Exception as e:
                exc_type, exc_obj, exc_tb = sys.exc_info()
                log_copy = log_common.copy()
                error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
                log_copy.update(error)
                print(error)
            finally:
                # release proxy back in use if it is not released it will be not be available for that web_base next 30min
                self.release_proxy(sha)

            print(rid, last_page, "total_records",  total_records)

            if no_results is not None:
                print(vrsta_objave, 'no results')
                # log_common.update({'status': 1, 'errors': 99})
                # self.info_log(log_common)
                continue

            # parse first page
            parsed_data = parse2(html, self.courts, self.court_groups, rid, parser_dir_errors)

            # if it is dictionary then some Error occurred during parsing
            # html is saved and it's location is under error desc
            if type(parsed_data) is dict:
                log_copy = log_common.copy()
                log_copy.update(parsed_data)
                print("cp.4", log_copy)
                # self.error_log(log_copy)
                continue

            # if parsed_data is not dictionary record html
            self.record_html(html)

            cnt = 0
            sys.stdout.write(f"\r> inserting in database: {cnt}")
            for link, data in parsed_data[0].items():
                cnt += 1
                sys.stdout.write(f"\r> inserting in database: {cnt}")
                merge_status = self.merge('search_results', data, filters={'details_link': link, 'archive': None}, on=['sha'], schema=schema)
                if merge_status != 200:
                    print("merge_status error")
                    break

            checked_pages = []
            while True:
                pages = list(range(2, last_page + 1))
                # preparing number of pages to parse excluding page 1
                total_pages = len(pages)
                pages = [page for page in pages if page not in checked_pages]
                with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
                    # Start the load operations and mark each future with its URL
                    future_to_url = {executor.submit(self.load_page, page, rid, vrsta_objave): page for page in pages}
                    for future in concurrent.futures.as_completed(future_to_url):
                        url = future_to_url[future]
                        try:
                            data, checked_page = future.result()
                            if checked_page not in [-1]:
                                checked_pages.append(checked_page)
                        except AttributeError as e:
                            pass
                        except Exception as e:
                            exc_type, exc_obj, exc_tb = sys.exc_info()
                            log_copy = log_common.copy()
                            log_copy.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
                            print("-- 6", log_copy)
                            # self.error_log(log_copy)
                            error_counter += 1
                        else:
                            if type(data) is dict:
                                log_copy = log_common.copy()
                                log_copy.update(data)
                                print("-- 7", log_copy)
                                # self.error_log(log_copy)
                                error_counter += 1

                total_checked = len(checked_pages)
                print(f"\n> rid:{rid} checked: {total_checked}, total_pages: {total_pages}, tokens available: {len(self.tokens)}")

                if total_checked == total_pages:
                    break

        self.all_done = True

    def load_page(self, page, rid, vrsta_objave):
        sha = None
        exception = None
        proc_id = threading.current_thread().ident
        request_id, g_token = self.forward_token()

        sys.stdout.write(f"\r> load_page rid: {rid},  page:{page} , thread_id: {proc_id}, tokens available: {len(self.tokens)}")

        url = f'https://e-oglasna.pravosudje.hr/objave-sudova?{vrsta_objave}&datePublishedFrom={date_from}&datePublishedTo={date_to}&page={page}&sort=datePublished,desc&_g={g_token}'
        try:
            sleep(randrange(2, 5))
            sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)
            r = self.requests.get(proc_id)
            r.request_type = 2
            r.ses.verify = False

            r.go(url)
            html = r.response.content
            html_raw = r.response.content_raw

            if type(html) is dict:
                err_desc = html.get('RequestError')
                bad_gateway = re.search('502 Bad Gateway', str(err_desc))
                forbiden = re.search('Tunnel connection failed: 403 Forbidden', str(err_desc))
                if err_desc == 'Blocked' or bad_gateway is not None or forbiden is not None:
                # if err_desc == 'Blocked':
                    self.bad_proxy(sha, r.proxy_data)
                exception = {'err_desc': f"def load_page during geting search page {rid} data err:{err_desc}"}
                return exception, -1

            # parse first page
            parsed_data = parse2(html, self.courts, self.court_groups,rid, parser_dir_errors)

            # if it is dictionary then some Error occurred during parsing
            # html is saved and it's location is under error desc
            if type(parsed_data) is dict:
                # log_copy = log_common.copy()
                # log_copy.update(parsed_data)
                print("cp.5", parsed_data)
                # self.error_log(log_copy)
                return parsed_data, -1

            # if parsed_data is not dictionary record html
            self.record_html(html, detalji=True)

            for link, data in parsed_data[0].items():
                merge_status = self.merge('search_results', data, filters={'details_link': link, 'archive': None}, on=['sha'], schema=schema)
                if merge_status != 200:
                    print("merge_status error")
                    return {'err_desc': 'error occured during merging data in load_page'}, -1

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            exception = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)}
            return exception, -1
        finally:
            # release proxy back in use if it is not released it will be not be available for that web_base next 30min
            self.release_proxy(sha)

        return parsed_data, page

    def court_rss_gather(self):
        sha = None
        proc_id = 1
        start_time = datetime.now()
        step_name = 'court_rss_gather'
        print(f"\n> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}
        courts = self.select('court_list', filters={}, schema=schema)
        try:
            sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)
            r = self.requests.get(proc_id)
            r.request_type = 1
            r.ses.verify = False

            for court in courts:
                court_group = court.get('court_group')
                court_name = court.get('court_name')
                court_id = court.get('court_id')

                sys.stdout.write(f'\r + fetching rss data: {court_name}')
                court_url = base+'/'.join([court.get('court_url'), 'rss'])
                r.go(court_url)

                json_data = r.response.json
                items = json_data.get('items')

                for item in items:
                    title = item.get('title')
                    details_link = item.get('link')
                    description = item.get('description').get('value')
                    description_type = item.get('description').get('type')
                    announce_date = pd.to_datetime(item.get('pubDate'), errors='coerce', yearfirst=True)
                    erase_date = pd.to_datetime(item.get('expirationDate'), errors='coerce', yearfirst=True)
                    author = item.get('author')

                    if details_link is not None:
                        details_link = details_link.replace(base, '')
                        row = {'title': title, 'details_link': details_link, 'description': description, 'description_type': description_type,
                               'announce_date': announce_date, 'erase_date': erase_date, 'author': author,
                               'court_group': court_group, 'court_name': court_name, 'court_id': court_id}
                        self.merge('court_rss', {0: row}, filters={'details_link': details_link}, on=['details_link'], update=False, schema=schema)
            print('\n ... DONE')
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            # print(error)
        finally:
            # release proxy back in use if it is not released it will be not be available for that web_base next 30min
            self.release_proxy(sha)

    def gather_details(self, night_hour=4):
        start_time = datetime.now()
        step_name = 'gather_details'
        print(f"\n> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}

        max_workers = 4
        error_counter = 0

        if datetime.now().hour >= night_hour and  datetime.now().hour <= night_hour+1:
            details_links = self.select('search_results', filters={'status': None}, columns=['details_link', 'court_name', 'court_id', 'court_group'], schema=schema)
        else:
            details_links = self.select('court_rss', filters={'status': None}, schema=schema)

        # details_links = self.select('court_rss', filters={'details_link': '/objave/bce6fb37-909e-4d3a-96e8-e627964c1b60'}, schema=schema)
        # details_links = details_links[:3]
        # details_links = [detail_link for detail_link in details_links if detail_link.get('announce_date') < short_date2]

        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            # Start the load operations and mark each future with its URL
            future_to_url = {executor.submit(self.load_detail, details_link): details_link for details_link in details_links}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    data, checked_page = future.result()
                except AttributeError as e:
                    pass
                except Exception as e:
                    exc_type, exc_obj, exc_tb = sys.exc_info()
                    log_copy = log_common.copy()
                    log_copy.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
                    print("-- 6", log_copy)
                    # self.error_log(log_copy)
                    error_counter += 1
                else:
                    if type(data) is dict:
                        log_copy = log_common.copy()
                        log_copy.update(data)
                        print("-- 7", log_copy)
                        # self.error_log(log_copy)
                        error_counter += 1

    def load_detail(self, rss_data):
        detail_link_short = rss_data.get('details_link')

        # fetch or generate rbr for unique details_link
        rbr = self.select('details_rbr', filters={'details_link': detail_link_short}, schema=schema)
        if not rbr:
            self.insert('details_rbr', {0: {'details_link': detail_link_short}}, schema=schema)
            rbr = self.select('details_rbr', filters={'details_link': detail_link_short}, schema=schema)
        rbr = rbr[0].get('rbr')

        detail_link = ''.join([base, detail_link_short])

        sha = None
        proc_id = 1
        start_time = datetime.now()
        step_name = 'court_rss_gather'
        log_common = {'start_time': start_time, 'step_name': step_name}

        try:
            sleep(randrange(5, 8))
            sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)
            r = self.requests.get(proc_id)
            r.request_type = 2
            r.ses.verify = False

            r.go(detail_link)

            # check are details still available on page
            no_data = {'Oglas nije dostupan za pregled': 404, 'Resurs nije prona.en': 405}
            data_status = 200
            for k, v in no_data.items():
                check_existance = re.search(k, str(r.response_raw.content), re.DOTALL | re.MULTILINE | re.IGNORECASE)
                if check_existance is not None:
                    data_status = v
                    break

            sys.stdout.write(f'\r>{detail_link}, {data_status}')

            # if details exist parse data
            if data_status == 200:
                html = r.response.content
                if type(html) is dict:
                    err_desc = html.get('RequestError')
                    bad_gateway = re.search('502 Bad Gateway', str(err_desc))
                    forbiden = re.search('Tunnel connection failed: 403 Forbidden', str(err_desc))
                    if err_desc == 'Blocked' or bad_gateway is not None or forbiden is not None:
                        # if err_desc == 'Blocked':
                        self.bad_proxy(sha, r.proxy_data)
                    exception = {'err_desc': f"def load_page during geting load_detail data err:{err_desc}"}
                    sleep(10)
                    return exception, -1

                parsed_details_data, parsed_doc_data = details(html, detail_link_short, rss_data, rbr)
                parsed_paticipants_data = participants_details(html, detail_link_short, rbr)

                # print('parsed_details_data', parsed_details_data)
                # print('parsed_doc_data', parsed_doc_data)
                # print('parsed_paticipants_data', parsed_paticipants_data)

                self.merge('details', {0: parsed_details_data}, filters={'details_link': detail_link_short, 'archive': None}, schema=schema)
                self.merge('documents', {0: parsed_doc_data}, filters={'details_link': detail_link_short, 'archive': None}, schema=schema)
                self.merge('participants', parsed_paticipants_data, filters={'details_link': detail_link_short, 'archive': None}, schema=schema)

            for table in ['court_rss', 'search_results']:
                self.update(table,  {'details_link': detail_link_short, 'status': None}, {'status': data_status, 'status_date': datetime.now()}, schema=schema)


        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            # print(error)
        finally:
            # release proxy back in use if it is not released it will be not be available for that web_base next 30min
            self.release_proxy(sha)

        return proc_id, 200

    def download_docs(self):
        start_time = datetime.now()
        step_name = 'download_docs'
        print(f"\n> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}

        max_workers = 5
        error_counter = 0

        docs_data= self.select('documents', filters={'archive': None, 'dl_status': None}, schema=schema)
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            # Start the load operations and mark each future with its URL
            future_to_url = {executor.submit(self.download_doc, doc_data): doc_data for doc_data in docs_data}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    data = future.result()
                except AttributeError as e:
                    pass
                except Exception as e:
                    exc_type, exc_obj, exc_tb = sys.exc_info()
                    log_copy = log_common.copy()
                    log_copy.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
                    print("-- download_docs err1", log_copy)
                    # self.error_log(log_copy)
                    error_counter += 1
                else:
                    if type(data) is dict:
                        log_copy = log_common.copy()
                        log_copy.update(data)
                        print("-- download_docs err2", log_copy)
                        # self.error_log(log_copy)
                        error_counter += 1

    def download_doc(self, doc_data):
        detail_link = doc_data.get('details_link')
        doc_link = doc_data.get('doc_link')
        dl_retries = doc_data.get('dl_retries')
        court_id = doc_data.get('court_id')
        rbr = doc_data.get('rbr')
        court_id = int(court_id)

        local_path = os.path.join(local_dir_path, str(court_id))
        server_path = os.path.join(server_doc_path, 'nova', str(court_id))

        if dl_retries is None:
            dl_retries = 0

        if dl_retries > 4:
            self.update('documents', {'doc_link': doc_link}, {'dl_status': 999}, schema=schema)
            return 999

        sha = None
        proc_id = threading.current_thread().ident
        start_time = datetime.now()
        step_name = 'download_doc'
        log_common = {'start_time': start_time, 'step_name': step_name}

        try:
            sleep(randrange(3, 5))
            sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)
            r = self.requests.get(proc_id)
            if dl_retries == 3:
                r.request_type = 1
            else:
                r.request_type = 2
            r.ses.verify = False

            dl_data = r.go(doc_link, download_raw=True)

            if type(dl_data) is dict:
                err_desc = dl_data.get('RequestError')
                bad_gateway = re.search('502 Bad Gateway', str(err_desc))
                forbiden = re.search('Tunnel connection failed: 403 Forbidden', str(err_desc))
                if err_desc == 'Blocked' or bad_gateway is not None or forbiden is not None:
                # if err_desc == 'Blocked':
                    self.bad_proxy(sha, r.proxy_data)
                exception = {'err_desc': f"def download_doc during geting load_detail data err:{err_desc}"}
                self.update('documents',  {'details_link': detail_link}, {'dl_retries': dl_retries+1}, schema=schema)
                return exception

            if int(dl_data.status_code) != 200:
                self.update('documents', {'details_link': detail_link}, {'dl_retries': dl_retries+1}, schema=schema)
                return int(dl_data.status_code)

            filename, extension = doc_filename(dl_data.headers)
            save_name = '.'.join([str(rbr), extension])

            # local dir path
            filename_full = os.path.join(local_path, save_name)
            # local dir path today
            filename_full_today = os.path.join(local_subdir_path_day, save_name)
            # server dir path
            filename_server_full = os.path.join(server_path, save_name)

            byte_data = dl_data.content
            file_size = len(byte_data)
            sys.stdout.write("\r> downloading: {}, file lenght:{}".format(filename_full, file_size))

            if file_size > 266:
                # writing files - local copy
                writing = [filename_full, filename_full_today]
                for path in writing:
                    with open(path, 'wb') as f:
                        f.write(byte_data)

                # try to save file on server right now if it is not possible pass -- enable when production ready
                try:
                    with open(filename_server_full, 'wb') as f:
                        f.write(byte_data)
                    values = {'up_status': 200, 'up_date': datetime.now(), 'up_path': filename_server_full}
                    self.update('documents', {'doc_link': doc_link, 'archive': None}, values, schema=schema)
                except:
                    pass

                dl_retries += 1
                doc_sha = hashlib.sha3_256(byte_data).hexdigest()
                values = {'dl_status': 200, 'dl_retries': dl_retries, 'dl_date': datetime.now(),
                          'doc_size': file_size, 'doc_sha': doc_sha, 'doc_name': save_name,
                          'doc_name_raw': filename, 'doc_path': filename_full}
                self.update('documents', {'doc_link': doc_link, 'dl_status': None},values, schema=schema)

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
        finally:
            # release proxy back in use if it is not released it will be not be available for that web_base next 30min
            self.release_proxy(sha)

        return 200

    def upload_documents(self):
        up_docs = self.select('documents', filters={'up_status': None, 'dl_status': 200, 'archive': None}, schema=schema)
        self.update('documents', {'up_status': None, 'dl_status': 999}, {'up_status': 999}, schema=schema)
        for up_doc in up_docs:
            sha = up_doc.get('sha')
            court_id = up_doc.get('court_id')
            doc_name = up_doc.get('doc_name')

            local_path = os.path.join(local_dir_path, str(court_id))  # local dir path
            server_path = os.path.join(server_doc_path, 'nova', str(court_id)) # server dir path

            filename_full = os.path.join(local_path, doc_name)  # local file path
            filename_server_full = os.path.join(server_path, doc_name)  # server file path

            check_existance = os.path.exists(filename_server_full)
            values = {'up_status': 200, 'up_path': filename_server_full, 'up_date': datetime.now()}
            filters = {'sha': sha, 'up_status': None}
            if check_existance is True:
                self.update('documents', filters, values, schema=schema)
                continue
            try:
                with open(filename_full, 'rb') as f:
                    file_bytes = f.read()
                with open(filename_server_full, 'wb') as f:
                     f.write(file_bytes)
                self.update('documents', filters, values, schema=schema)
            except Exception as e:
                print(f"unable to upload document. reason: {str(e)}")

    def transfer_data(self):
        # not optimized!
        def participant_prepare(participant, stecaj=True):
            vat_number = participant.get('vat_number')
            if stecaj is True:
                cols = {'participant': 'NAZIV_ST_DUZ', 'address': 'ADRESA'}
                participant_transfer = {'OIB': vat_number, 'OIB_STEC_DUZ': vat_number}
            else:
                cols = {'participant': 'OGLAS_SE_ODNOSI'}
                participant_transfer = {'OIB': vat_number}

            for k, v in cols.items():
                k_data = participant.get(k)
                if v == 'OGLAS_SE_ODNOSI' and k_data is not None:
                    if len(k_data) > 350:
                        k_data = k_data[:349]
                participant_transfer.update({v: k_data})
            return participant_transfer

        def top_participant(det_link, stecaj=True):
            participant = None
            participants = self.select('participants', filters={'details_link': det_link, 'archive': None}, schema=schema)
            # simple solution if there is only one participant
            if len(participants) == 1:
                participant = participants[0]
                participant = participant_prepare(participant, stecaj=stecaj)

            if len(participants) > 1:
                # highest grade row is the one with most available data
                # participant_role contains word Dužnik, vat_number is not empty, address is not empty
                highest_grade_row = {'id': -1, 'grade': -1}
                participants = dict(enumerate(participants))
                for id, part in participants.items():
                    grade = 0
                    participant_role = part.get('participant_role')
                    if re.search('dužnik', str(participant_role), re.IGNORECASE) is not None:
                        grade += 1
                    if part.get('vat_number') is not None:
                        grade += 1
                    if part.get('address'):
                        grade += 1
                    current_grade = highest_grade_row.get('grade')
                    if current_grade < grade:
                        highest_grade_row.update({'id': id, 'grade': grade})

                participant = participants.get(highest_grade_row.get('id'))
                participant = participant_prepare(participant, stecaj=stecaj)
            return participant

        tuzbe_cols = {'court_id': 'SUD_ID', 'rbr': 'RBR', 'decision': 'OZN_SPIS', 'decision_title': 'NAZIV_OTPRAVKA',
                      'announce_date': 'DATUM_OBJAVE', 'erase_date': 'DATUM_SKIDANJA', 'doc_link': 'DOKUMENT_URL', 'details_link': 'SPIS_LINK',
                      'ecase': 'EPREDMET_LINK', 'rid': 'RID'}

        print("sync tuzbe")
        tuzbe_datas = self.select('details', filters={'transfer': None, 'archive': None, 'badge': ['Objava suda', 'Stečaj']}, schema=schema)
        for tuzbe_data in tuzbe_datas:
            rbr = tuzbe_data.get('rbr')
            details_link = tuzbe_data.get('details_link')
            documents = self.select('documents', filters={'details_link': details_link, 'dl_status': [200, 999], 'archive': None}, schema=schema)
            # if document is not downloaded yet skip
            if not documents:
                continue

            participant = top_participant(details_link, stecaj=False)

            # filter and convert stecaj data
            tuzbe_ready = {'NOVA_OGLASNA': 1}
            for k, v in tuzbe_cols.items():
                key_data = tuzbe_data.get(k)
                tuzbe_ready.update({v: key_data})

            if participant is not None:
                tuzbe_ready.update(participant)

            objava_check = self.dbbisnode.select('OBJAVA', sql=f"select count(*)cnt from TUZBE.OBJAVA where RBR = {rbr}")[0].get('cnt')
            insert_status = 200
            if objava_check == 0:
                insert_status = self.dbbisnode.insert('OBJAVA', {0: tuzbe_ready}, schema='TUZBE', primary_key='rbr')

            if insert_status == 200:
                filters = {'details_link': details_link, 'archive': None}
                values = {'transfer': 1, 'transfer_date': datetime.now()}
                self.update('details', filters, values, schema=schema)

        stecaj_cols = {'court_id': 'SUD_ID', 'rbr': 'RBR', 'decision': 'OZN_SPIS', 'decision_title': 'NAZIV_OTPRAVKA',
                       'announce_date': 'DATUM_OBJAVE', 'doc_link': 'DOKUMENT_URL', 'details_link': 'SPIS_LINK'}

        stecaji_data = self.select('details', filters={'transfer': 1, 'archive': None, 'badge': 'Stečaj'}, schema=schema)
        print("sync stecajevi")
        for stecaj_data in stecaji_data:
            rbr = stecaj_data.get('rbr')
            details_link = stecaj_data.get('details_link')

            documents = self.select('documents', filters={'details_link': details_link, 'dl_status': [200, 999], 'archive': None}, schema=schema)
            # if document is not downloaded yet skip
            if not documents:
                continue

            participant = top_participant(details_link)

            # filter and convert stecaj data
            stecaj_ready = {}
            for k, v in stecaj_cols.items():
                key_data = stecaj_data.get(k)
                stecaj_ready.update({v: key_data})

            if participant is not None:
                stecaj_ready.update(participant)

            stecaj_check = self.dbbisnode.select('STECAJ', sql=f"select count(*)cnt from TUZBE.STECAJ where RBR = {rbr}")[0].get('cnt')

            insert_status = 200
            if stecaj_check == 0:
                insert_status = self.dbbisnode.insert('STECAJ',{0: stecaj_ready}, schema='TUZBE', primary_key='rbr')

            if insert_status == 200:
                filters = {'details_link': details_link, 'archive': None}
                values = {'transfer': 2, 'transfer_date': datetime.now()}
                self.update('details', filters, values, schema=schema)

        docs_cols = {'doc_name': 'FILE_DOWNLOAD', 'dl_status': 'HTTPSTATUS', 'md5': 'HASHMD5',
                     'doc_size': 'LEN_DOWNLOAD', 'dl_date': 'DATE_CREATED'}

        print("sync documents")
        filter_vars = [{'transfer': 1, 'badge': 'Objava suda'}, {'transfer': 2, 'badge': 'Stečaj'}]
        for filter_var in filter_vars:
            docs_transfer = self.select('details', filters=filter_var, schema=schema)
            for doc_transfer in docs_transfer:
                rbr = doc_transfer.get('rbr')
                details_link = doc_transfer.get('details_link')

                docs_data = self.select('documents', filters={'details_link': details_link, 'dl_status': True}, schema=schema)
                if not docs_data:
                    continue
                for doc_data in docs_data:
                    doc_sha = doc_data.get('doc_sha')
                    dl_status = doc_data.get('dl_status')
                    filters = {'details_link': details_link, 'archive': None}
                    values = {'transfer': 3, 'transfer_date': datetime.now()}
                    if doc_sha is None and dl_status in [999]:
                        self.update('details', filters, values, schema=schema)
                    if doc_sha is not None:
                        md5 = hashlib.md5(doc_sha.encode()).hexdigest()
                        doc_data.update({'md5': md5})
                        file_id = self.dbbisnode.select('OBJAVA', sql=f"select id from TUZBE.OBJAVA where RBR = {rbr}")
                        if file_id:
                            file_id = file_id[0].get('id')

                            # filter and convert docs data
                            doc_ready = {'ID_OBJAVA': file_id}
                            for k, v in docs_cols.items():
                                key_data = doc_data.get(k)
                                doc_ready.update({v: key_data})

                            doc_check = self.dbbisnode.select('OBJAVA_FILES', sql=f"select count(*)cnt from TUZBE.OBJAVA_FILES where ID_OBJAVA = {file_id}")[0].get('cnt')

                            insert_status = 200
                            if doc_check == 0:
                                insert_status = self.dbbisnode.insert('OBJAVA_FILES', {0: doc_ready}, schema='TUZBE', primary_key='id_objava')

                            if insert_status == 200:
                                self.update('details', filters, values, schema=schema)

        # update stečaj potrošača
        self.update('details', {'transfer': None, 'badge': 'Stečaj potrošača'}, {'transfer': 0}, schema=schema)

    def loop_download(self):
        for i in range(10):
            docs_data = self.select('documents', filters={'archive': None, 'dl_status': None}, schema=schema)
            if not docs_data:
                break
            self.download_docs()

    def fetch_archive_doc(self):
        archive_url = 'https://e-oglasna.pravosudje.hr/dokumenti/arhiva-objava'
        sha = None
        proc_id = 1
        start_time = datetime.now()
        step_name = 'fetch_archive_doc'
        log_common = {'start_time': start_time, 'step_name': step_name}

        column_repl = {'Vrsta objave': 'badge', 'Sud': 'court_name', 'Oznaka spisa': 'decision',
                       'Naslov': 'decision_title', 'Stečajni dužnik': 'debtor',
                       'Adresa stečajnog dužnika': 'debtor_address', 'OIB stečajnog dužnika': 'vat_number',
                       'Sudionici': 'participants', 'Početni dan objave': 'announce_date',
                       'Posljednji dan objave': 'erase_date', 'Oglas (link)': 'details_link',
                       'Dokumenti (datoteke)': 'doc_name', 'Dokumenti (link)': 'doc_link', 'e-Predmet (link)': 'ecase'}

        try:
            for i in range(8):
                sleep(randrange(5, 8))
                sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)
                r = self.requests.get(proc_id)
                self.release_proxy(sha)
                r.request_type = 1
                r.ses.verify = False

                r.go(archive_url)

                html = r.response.content

                if type(html) is dict:
                    print(html)
                    sleep(60)
                    continue

                table_links = html.find_all('a', {'href': re.compile('oglasi_objavljeni')})
                for table_link in table_links:
                    dl_link = table_link['href']
                    dl_text = table_link.text.strip()
                    check_table = self.select('archive_tables',filters={'table_name': dl_text}, schema=schema)
                    if not check_table:
                        excel_bytes = r.go(dl_link, download=True)
                        excel_link = os.path.join(xlsx_tables, dl_text)
                        with open(excel_link, 'wb') as f:
                            f.write(excel_bytes)

                        df = pd.read_excel(excel_link)
                        if df.empty is False:
                            df.columns = [column_repl.get(col) for col in df.columns]
                            df['doc_name'] = [str(docn)[:1000] for docn in df['doc_name']]
                            df.replace(pd.NaT, None, inplace=True)
                            df = df.where((pd.notnull(df)), None)
                            df['table_name'] = [dl_text]*len(df)
                            df['sha'] = [hashlib.sha3_256(str(self.order_data(row)).encode()).hexdigest() for row in df.values]
                            df['details_link'] = [row.replace('https://e-oglasna.pravosudje.hr', '') for row in df['details_link']]
                            merge_data = df.to_dict(orient='index')
                            self.merge('archive_tables', merge_data, filters={'table_name': dl_text}, schema=schema)
                break

            the_day = datetime.now()-timedelta(days=1)
            last_table = f'oglasi_objavljeni_{the_day.year}-{the_day.month}-{the_day.day}.xlsx'
            archive_datas = self.select('archive_tables', filters={'table_name': last_table}, schema=schema)

            for archive_data in archive_datas:
                details_link = archive_data.get('details_link')
                decision_title = archive_data.get('decision_title')
                archive_data.update({'title': decision_title, 'description': decision_title})
                archive_data.pop('decision_title', None)
                court_name = archive_data.get('court_name')
                court_group = self.court_groups.get(court_name)
                court_id = self.courts.get(court_name)
                archive_data.update({'court_group': court_group, 'court_id': court_id, 'author': 'XLSX'})
                self.merge('court_rss', {0: archive_data}, filters={'details_link': details_link}, on=['details_link'], update=False, schema=schema)

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)


def rss_start():
    print('rss_start')
    while True:
        day_name = datetime.now().strftime("%A")
        if day_name not in ['Saturday', 'Sunday']:
            current_hour = datetime.now().hour
            if current_hour < 21 and current_hour > 5:
                api = Eoglasna()
                api.postgresql_conn()
                api.court_rss_gather()
                gc.collect()
        if day_name  in ['Saturday', 'Sunday']:
            sleep(3600)
        sleep(30)


def crawler_start():
    print('crawler_start')
    start_time = datetime.now()
    api = Eoglasna()
    api.postgresql_conn()
    api.db_court_list()

    list_size_check = 0
    current_date = datetime(now.year, now.month, now.day)

    while True:
        day_name = datetime.now().strftime("%A")
        diff_time = (datetime.now() - start_time).total_seconds()
        current_hour = datetime.now().hour
        if current_hour < 21 and current_hour > 5 and day_name not in ['Saturday', 'Sunday']:
            if diff_time > 300:
                list_size = len(api.select('court_rss', filters={'status': None}, schema=schema))
                if list_size > 100 or list_size_check > 2:
                    list_size_check = 0
                    api.dir_creation_check()
                    api.gather_details()
                    api.loop_download()
                    api.upload_documents()
                    api.transfer_data()
                if list_size > 0 and list_size < 101:
                    list_size_check += 1
                start_time = datetime.now()

        if current_hour > 1 and current_hour < 9:
            api.dir_creation_check()
            api.fetch_archive_doc()
            api.gather_details()
            api.loop_download()
            api.upload_documents()
            api.transfer_data()
            sleep(3600)

        # if current_hour > 19 and current_hour < 23 and current_date < datetime(now.year, now.month, now.day):
        #     api.token_server()
        #     api.gather_search_results()
        #     api.gather_details()
        #     api.loop_download()
        #     api.transfer_data()
        #     current_date = datetime(now.year, now.month, now.day)
        if day_name == 'Monday' and current_hour >= 0 and current_hour <6:
            sleep(3600)
        gc.collect()
        sleep(10)


def simple_night_check():
    api = Eoglasna()
    api.postgresql_conn()
    api.db_court_list()
    api.dir_creation_check()

    api.token_server()
    api.gather_search_results()

    api.gather_details(night_hour=19)
    api.download_docs()
    api.upload_documents()
    api.transfer_data()


def task_handler(task):
    task()


if __name__ == '__main__':
    # api = Eoglasna()
    # api.postgresql_conn()
    # api.db_court_list()
    # api.dir_creation_check()
    # api.transfer_data()

    pool = MyPool(2)
    pool.map(task_handler, [rss_start, crawler_start])
    pool.close()
    pool.join()
