"""
+-------------------------------------------+
    Program: EOGLASNA CRAWLER 2019/10|
    Developed by: Dragan Matesic|
    Version: Optimized for Oracle database
+-------------------------------------------+
"""
import re
import os
import sys
import random
import hashlib
import urllib3
import threading
import pandas as pd
from abc import ABC
from time import sleep
from sys import stdout
import concurrent.futures
from random import randrange
from datetime import datetime, timedelta
from core_framework.crawlers import CrawlerBase
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
from crawlers_sm.croatia.eoglasna.parser import parse, detalji_sudionik, detalji, doc_filenames

os.environ['NLS_LANG'] = 'CROATIAN_CROATIA.UTF8'
api_key = '7c315478c67a270a2d829f1a00f9096c'
site_key = '6LcJFLcUAAAAAAWb8bmUnRGwe9RBS4ADwlu2-VBp'

base = 'https://e-oglasna.pravosudje.hr'
objave_url = '/'.join([base, 'objave-sudova'])
captcha_api = f'http://2captcha.com/in.php?key={api_key}&method=userrecaptcha&googlekey={site_key}&pageurl={objave_url}'


base = 'https://e-oglasna.pravosudje.hr'

# download  u upload dokumenata lokacije
main_doc_path = 'E:\TuzbeDocs\\'
today = datetime.today().strftime('%Y%m%d')
current_hour = datetime.now().hour

server_doc_path = os.path.join('Z:\\', 'oglastrgsud')

local_dir_path = os.path.join(main_doc_path, 'download')
local_dir_path_fina = os.path.join(main_doc_path, 'download', 'fina')

local_dir_path_day = os.path.join(main_doc_path, 'downloaddaily')
local_subdir_path_day = os.path.join(main_doc_path, 'downloaddaily', today)

server_dir_path = server_doc_path
server_dir_path_fina = os.path.join(server_doc_path, 'FINA')

# directory is used for storing html that resulted in some kind of exception or html could not be parsed
parser_dir_errors = os.path.join(main_doc_path, 'parser_errors')

# directory is used save html that server returned
htmls_dir = os.path.join(main_doc_path, 'eoglasna_htmls')
htmls_subdir = os.path.join(main_doc_path, 'eoglasna_htmls', today)

# search data for specified period of time, we look only for current day
date_to = datetime.now().strftime('%Y-%m-%d')
date_from = datetime.now().strftime('%Y-%m-%d')

# date_to = "2019-10-25"
# date_from = "2019-10-25"

# securing ourself from endleess loop
loop_limiter = 5

# logging
error_log_table = 'EOGLASNA_ERROR_LOG'
info_log_table = 'EOGLASNA_INFO_LOG'  # status 1 = no errors, 2 = done with errors, 99 = no results
token_log_table = 'EOGLASNA_TOKENS_LOG'  # status 1 = we did get token, 2 = error during geting token, 3 = we did get token but error occured during reques HInt check EOGLASNA_ERROR_LOG over session_id


class Eoglasna(CrawlerBase, ABC):
    def __init__(self, recheck=None):
        CrawlerBase.__init__(self, base, proxy_type=2)
        self.workers = 5
        self.courts = dict()
        self.flag_120 = False
        self.exit = False  # if this is True then program will exit because something went wrong during preload of app
        self.predmeti_last = self.fetch_last_case()

        # runtime log initialization
        self.runtime_errors = 0
        self.runtime_status = 1
        self.runtime_start = datetime.now()
        self.runtime_id = hashlib.sha3_256(str(self.runtime_start).encode()).hexdigest()
        self.runtime_log = {'start_time': self.runtime_start, 'step_name': 'runtime'}

        # during day we want to limit download to list until it hits last court case link we have in db
        self.day_download = False
        self.doc_update = False
        self.download_attrs = {'start_time': datetime.now(), 'step_name': 'download_docs', 'error_counter': 0}

        # during daily run we want to limit number of pages because it can happen that one of the pages
        # inside that range isn't been able to be processed and that could result in crawling all pages
        # and that would be unnecessary expense sine in 20 h we check all pages to get all data that is missing
        self.page_limit_default = 5
        # if 13 < current_hour < 19:
        #     self.day_download = True

        self.last_page = {1: None, 2: None}

    def fetch_last_case(self):
        self.connect(4)
        rid_predmeti_last = {}
        predmeti_last = self.select('V_EOGLASNA_ZADNJI', view=True)
        for row in predmeti_last:
            rid = row.get('rid')
            link = row.get('detalji_objave_link')
            rid_data = rid_predmeti_last.get(rid)
            if rid_data is None:
                rid_predmeti_last.update({rid: [link]})
            else:
                rid_data.append(link)
        self.disconnect()
        return rid_predmeti_last

    def eval_last_link(self,parsed_data, rid, page):
        last_link = False
        if parsed_data is not None and parsed_data != []:
            for d in parsed_data:
                detalji_objave_link = d.get('detalji_objave_link')
                rid_data = self.predmeti_last.get(rid)
                # print(rid, page, detalji_objave_link, self.last_page.get(rid))
                if rid_data is not None and self.last_page.get(rid) is None:
                    if detalji_objave_link in rid_data:
                        last_link = True
                        break
        return last_link

    def info_log(self, data):
        if data is not None:
            sha = hashlib.sha3_256(str(data).encode()).hexdigest()
            data.update({'sha': sha, 'runtime_id': self.runtime_id, 'end_time': datetime.now()})
            self.connect(4)
            self.insert(info_log_table, {0: data})

    def error_log(self, data):
        if data is not None:
            self.runtime_status = 2
            self.runtime_errors += 1
            sha = hashlib.sha3_256(str(data).encode()).hexdigest()
            data.update({'sha': sha, 'runtime_id': self.runtime_id, 'end_time': datetime.now()})
            self.connect(4)
            err_cnt = self.select(error_log_table, filters={'sha': sha}, columns=['err_cnt'])
            if not err_cnt:
                data.update({'err_cnt': 1})
                self.insert(error_log_table, {0: data})
            else:
                err_cnt = err_cnt[0].get('err_cnt') + 1
                self.update(error_log_table, {'sha': sha}, {'err_cnt': err_cnt})

    def token_log(self, data):
        sha = hashlib.sha3_256(str(data).encode()).hexdigest()
        data.update({'sha': sha, 'runtime_id': self.runtime_id, 'end_time': datetime.now()})
        self.connect(4)
        self.insert(token_log_table, {0: data})
        self.disconnect()

    def get_court_list(self):
        start_time = datetime.now()
        step_name = 'get_court_list'
        print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}
        try:
            self.connect(4)
            self.courts = self.select('SUD', schema='TUZBE')
            self.courts = {row.get('naziv'): row for row in self.courts}
            log_common.update({'status': 1, 'errors': 0})
            self.info_log(log_common)
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            log_common.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
            self.error_log(log_common)
            self.exit = True
        return self.courts

    def dir_creation_check(self):
        start_time = datetime.now()
        step_name = 'dir_creation_check'
        print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}
        try:
            dirs = [local_dir_path, local_dir_path_fina, local_dir_path_day, local_subdir_path_day, parser_dir_errors, htmls_dir, htmls_subdir]
            for dir_path in dirs:
                if os.path.isdir(dir_path) is False:
                    os.mkdir(dir_path)
            log_common.update({'status': 1, 'errors': 0})
            self.info_log(log_common)
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            log_common.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
            self.error_log(log_common)
            self.exit = True

    def get_token(self, proc_id=1):
        try:
            exit = False
            token_id = None
            request_id = None
            step_name = 'get_token'
            start_time = datetime.now()
            log_common = {'start_time': start_time, 'step_name': step_name, 'gtoken': None, 'session_id': None}
            self.connect(1, connect_args={"application_name": "crawlers_sm/croatia/eoglasna/crawler.py/Eoglasna"})

            sha = self.new_request(proc_id=proc_id, protocols=['http'], new=True)
            r = self.requests.get(proc_id)
            r.request_type = 1
            self.disconnect()

            try:
                r.go(captcha_api)
                html = r.response.content
                if not type(html) is dict:

                    id_search = re.search('(\d+)', html.decode())
                    if id_search is not None:
                        request_id = id_search.group(1)
                    log_common.update({'session_id': request_id})
                    sleep(randrange(10, 15))

                    loop_error = None
                    for i in range(10):
                        request_check = f"http://2captcha.com/res.php?key={api_key}&action=get&id={request_id}"
                        r.go(request_check)
                        html_bs = r.response.content
                        if type(html_bs) is dict:
                            loop_error = html_bs.get('RequestError')
                            sleep(20)
                            continue

                        html = html_bs.text
                        if html == 'CAPCHA_NOT_READY':
                            sleep(10)
                        else:
                            loop_error = None
                            token_id = html.replace("OK|", "")
                            log_common.update({'gtoken': token_id, 'status': 1})
                            self.token_log(log_common)
                            break

                    if loop_error is not None:
                        log_common.update({'err_desc': f"during waiting token {loop_error}", 'status': 2})
                        self.token_log(log_common)

                if type(html) is dict:
                    error = html.get('RequestError')
                    log_common.update({'err_desc': error, 'status': 2})
                    self.token_log(log_common)

            except Exception as e:
                exc_type, exc_obj, exc_tb = sys.exc_info()
                log_common.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2})
                self.token_log(log_common)
                exit = True

            finally:
                # release proxy back in use if it is not released it will be not be available for that web_base next 30min
                self.release_proxy(sha)

            if token_id is None or token_id == '' or token_id == 'ERROR_CAPTCHA_UNSOLVABLE':
                if token_id == 'ERROR_CAPTCHA_UNSOLVABLE':
                    sleep(20)
                if token_id == '':
                    token_id = None
                if exit is True:
                    error = 'could not get token from captcha service'
                    log_common.update({'err_desc': error, 'status': 2, 'gtoken':token_id})
                    self.token_log(log_common)
                    self.disconnect()
                return 404, 404
            else:
                # close open db connection
                self.disconnect()
                return request_id, token_id

        except Exception as e:
            err_desc = f'get_token {str(e)}'
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': err_desc}
            print(error)
            return 404, 404

    def search_crawler(self):
        if self.exit is True:
            return 404

        start_time = datetime.now()
        step_name = 'search_crawler'
        print(f"\n> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}

        proc_id = 1
        error_counter = 0
        no_results = None
        rows_search = None

        vrste_objava = {1: 'noticeSourceType=COURT', 2: 'noticeSourceType=NOTARY_PUBLIC'}
        # vrste_objava = {1: 'noticeSourceType=COURT'}
        # vrste_objava = {2: 'noticeSourceType=NOTARY_PUBLIC'}
        # rid is source of information it can be eather public notary or court

        for rid, vrsta_objave in vrste_objava.items():
            all_data = []
            last_page = 1
            html = ''
            major_error = False
            sha = None

            for i in range(5):
                self.page_limit = self.page_limit_default

                print(f"> {step_name} rid: {rid}")
                all_data = []
                last_page = 1
                html = ''
                major_error = False

                self.last_page.update({rid: None})

                # first we going to get google token for this page
                loop_counter = 0
                while True:
                    token_data = self.get_token()
                    if type(token_data) is tuple or loop_counter > loop_limiter:
                        session_id, g_token = token_data
                        break
                    loop_counter += loop_limiter
                    print("token_data", token_data)
                if loop_counter > 5 or g_token == 404:
                    print(f"loop counter > 5 , g_token: {g_token}")
                    continue

                # get all changes from today
                new_sha = None
                self.connect(1, connect_args={"application_name": "crawlers_sm/croatia/eoglasna/crawler.py/Eoglasna"})
                if sha is None:
                    sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)

                r = self.requests.get(proc_id)
                r.request_type = 2
                print(f'proxy {r.ip},{r.port}')
                try:
                    g_url = f'https://e-oglasna.pravosudje.hr/objave-sudova?{vrsta_objave}&datePublishedFrom={date_from}&datePublishedTo={date_to}&_g={g_token}'
                    r.ses.verify = False
                    r.go(g_url)
                    html = r.response.content
                    if type(html) is dict:
                        err_desc = html.get('RequestError')
                        error = f'rid: {rid} desc: {err_desc} '
                        log_common.update({'err_desc': error, 'token_ses_id': session_id})
                        print("-- 1", log_common)
                        self.error_log(log_common)
                        error_counter += 1
                        new_sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)
                        major_error = True
                    else:
                        # get last page
                        last_page_re = html.find('a', {'title': 'Poslijednja stranica', 'href': True})
                        if last_page_re is not None:
                            last_page_link = last_page_re['href']
                            last_page_n = re.search('page=(\d+)', last_page_link, re.DOTALL | re.MULTILINE | re.IGNORECASE)
                            if last_page_n is not None:
                                last_page = int(last_page_n.group(1))

                        # get total number of court cases specified for current rid search
                        rows_search = re.search('Pronađeno rezultata:\s+<strong>(\d+)</strong>', html.decode(), re.DOTALL | re.MULTILINE | re.IGNORECASE)
                        if rows_search is not None:
                            rows_search = rows_search.group(1)

                        # check is there any data
                        no_results = html.find('em', text=re.compile('Nema rezultata.'))

                except Exception as e:
                    exc_type, exc_obj, exc_tb = sys.exc_info()
                    log_common.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
                    print("\n-- 2",log_common)
                    self.error_log(log_common)
                    error_counter += 1
                    major_error = True

                finally:
                    # release proxy back in use if it is not released it will be not be available for that web_base next 30min
                    self.release_proxy(sha)
                    self.disconnect()
                    sha = None

                if no_results is not None:
                    log_common.update({'status': 1, 'errors': 99})
                    print("-- 3",log_common)
                    self.info_log(log_common)
                    continue

                # if some major error occured we go on the next rid
                if major_error is True:
                    print("> major_error", major_error)
                    sha = new_sha
                    sleep(60)
                    continue
                break
            if major_error is True:
                break

            print("> rid", rid, "last_page", last_page, 'rows', rows_search)

            # parse first page
            parsed_data = parse(html, self.courts, rid, parser_dir_errors)

            # if it is dictionary then some Error occurred during parsing
            # html is saved and it's location is under error desc
            if type(parsed_data) is dict:
                log_copy = log_common.copy()
                log_copy.update(parsed_data)
                print("-- 4",log_copy)
                self.error_log(log_copy)
                error_counter += 1
                continue

            # if parsed_data is not dictionary record html
            self.record_html(html)

            all_data.extend(parsed_data)

            # check if it reached daily page for this rid
            if self.day_download is True:
                last_link = self.eval_last_link(parsed_data, rid, 1)
                if last_link is True:
                    self.database_insert(parsed_data)
                    continue

            checked_pages = []
            while True:
                pages = list(range(2, last_page + 1))
                # preparing number of pages to parse excluding page 1
                total_pages = len(pages)
                pages = [page for page in pages if page not in checked_pages]
                with concurrent.futures.ThreadPoolExecutor(max_workers=self.workers) as executor:
                    # Start the load operations and mark each future with its URL
                    future_to_url = {executor.submit(self.load_page, page, rid, vrsta_objave): page for page in pages}
                    for future in concurrent.futures.as_completed(future_to_url):
                        url = future_to_url[future]
                        try:
                            data, checked_page = future.result()
                            if checked_page not in [-1]:
                                checked_pages.append(checked_page)
                        except AttributeError as e:
                            print("-- 5", str(e))
                            pass
                        except Exception as e:
                            exc_type, exc_obj, exc_tb = sys.exc_info()
                            log_copy = log_common.copy()
                            log_copy.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
                            print("-- 6",log_copy)
                            self.error_log(log_copy)
                            error_counter += 1
                        else:
                            # insert parsed data
                            if type(data) is list:
                                if data:
                                    all_data.extend(data)

                            if type(data) is dict:
                                log_copy = log_common.copy()
                                log_copy.update(data)
                                print("-- 7", log_copy)
                                self.error_log(log_copy)
                                error_counter += 1

                total_checked = len(checked_pages)
                print(f"\n> rid:{rid} checked: {total_checked}, total_pages: {total_pages}, last page: {self.last_page.get(rid)}")

                if total_checked == total_pages and self.last_page.get(rid) is None and self.day_download is True:
                    check_more = 5
                    uncheck_pages = list(range(self.page_limit+1, self.page_limit+check_more+1+2))
                    uncheck_pages = [p for p in uncheck_pages if p <= last_page]
                    print(f"> last url not find spreading search on next {check_more} pages: {uncheck_pages}")
                    if not uncheck_pages:
                        break

                    for uncheck_page in uncheck_pages:
                        checked_pages.remove(uncheck_page)
                    self.page_limit += check_more
                    continue

                if total_checked == total_pages:
                    break

            self.disconnect()
            self.database_insert(all_data)

        log_type = log_common.copy()
        status = 1
        if error_counter > 0:
            status = 2
        log_type.update({'status': status, 'errors': error_counter})
        self.info_log(log_type)

    def load_page(self, page, rid, vrsta_objave):
        try:
            sha = ''
            last_link = False
            exception = None
            parsed_data = []
            proc_id = threading.current_thread().ident
            rid_last_page = self.last_page.get(rid)

            if rid_last_page is not None and self.day_download is True:
                if page > rid_last_page:
                    return parsed_data, page

            if self.page_limit < page and self.day_download is True and rid_last_page is None:
                return parsed_data, page

            stdout.write(f"\r> load_page rid: {rid},  page:{page} , thread_id: {proc_id}, last_page: {rid_last_page}")

            # first we going to get google token for this page
            loop_counter = 0
            while True:
                token_data = self.get_token()
                if type(token_data) is tuple or loop_counter > loop_limiter:
                    session_id, g_token = token_data
                    break
                loop_counter += loop_limiter
            if loop_counter > 5:
                return {'err_desc': "def load_page token loop counter limit reached"}, -1

            if g_token == 404:
                return parsed_data, -1

            try:
                sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)
                r = self.requests.get(proc_id)
                r.request_type = 2
                r.ses.verify = False

                url = f'https://e-oglasna.pravosudje.hr/objave-sudova?{vrsta_objave}&datePublishedFrom={date_from}&datePublishedTo={date_to}&page={page}&sort=datePublished,desc&_g={g_token}'
                r.go(url)
                html = r.response.content
                if type(html) is dict:
                    err_desc = html.get('RequestError')
                    self.bad_proxy(sha, r.proxy_data)
                    # raise Exception(f"def load_page during geting search page data err:{err_desc}")
                    exception = {'err_desc': f"def load_page during geting search page data err:{err_desc}"}
                else:
                    parsed_data = parse(html, self.courts, rid)
                    # daily filter
                    if self.day_download is True:
                        last_link = self.eval_last_link(parsed_data, rid, page)

            except Exception as e:
                exc_type, exc_obj, exc_tb = sys.exc_info()
                exception = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)}
            finally:
                # release proxy back in use if it is not released it will be not be available for that web_base next 30min
                self.release_proxy(sha)

            if exception is not None:
                return exception, -1

            if self.day_download is True and last_link is True:
                self.last_page.update({rid: page})
            return parsed_data, page

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error_desc = f"major page_load error: reason {str(e)}"
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': error_desc}
            print(error)
            return error, -1

    @staticmethod
    def record_html(html):
        try:
            sha = hashlib.sha3_256(str(html).encode()).hexdigest()
            doc_loc = '/'.join([htmls_subdir, f'{sha}.html'])
            with open(doc_loc, 'w') as f:
                f.write(str(html))
            return 200
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error_desc = f"Could not write html: reason {str(e)}"
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': error_desc}
            print(error)
            return error

    def database_insert(self, data):
        cnt = 0
        stdout.write(f"\r> inserting in database: {cnt}")
        if data is not None and data != []:
            self.connect(4)
            for d in data:
                detalji_objave_link = d.get('detalji_objave_link')
                if detalji_objave_link is not None:
                    cnt += 1
                    stdout.write(f"\r> inserting in database: {cnt}")
                    # assign new rbr to new court case link
                    self.merge('EOGLASNA_RBR', {0: {'detalji_objave_link': detalji_objave_link}}, filters={'detalji_objave_link': detalji_objave_link}, on=['detalji_objave_link'])
                    sha = d.get('sha')
                    rbr = self.select('EOGLASNA_RBR', filters={'detalji_objave_link': detalji_objave_link})[0].get('rbr')
                    d.update({'rbr': rbr})
                    # insert results of search in table
                    self.merge('EOGLASNA_PRETRAGA', {0: d}, filters={'sha': sha})
            self.disconnect()

    def transfer_data(self):
        start_time = datetime.now()
        step_name = 'transfer_data'
        log_common = {'start_time': start_time, 'step_name': step_name}
        print("\n> transfering data to tuzba schema tables")
        self.connect(4)
        self.proc('eoglasna_insert_data'.upper())
        log_common.update({'status': 1, 'errors': 0})
        self.info_log(log_common)

    @staticmethod
    def filemd5(filename, block_size=2 ** 20):
        f = open(filename, errors='ignore')
        md5 = hashlib.md5()
        while True:
            data = f.read(block_size)
            if not data:
                break
            data = data.encode()
            md5.update(data)
        f.close()
        return md5.hexdigest()

    def download_docs(self, run_cnt=1):
        step_name = 'download_docs'
        if run_cnt == 1:
            print(f"> {step_name}")
            self.download_attrs.update({'start_time': datetime.now()})
        start_time = self.download_attrs.get('start_time')
        error_counter = self.download_attrs.get('error_counter')
        log_common = {'start_time': start_time, 'step_name': step_name}

        all_data, all_updates, all_errors = [], [], []

        # download part
        self.connect(4)
        download_lista = self.select('v_tuzbe_download', view=True)

        if not download_lista:
            # dodano kako bi za sinisu se dohvatili iz headera nazivi dokumenata
            self.doc_update = True
            download_lista2 = self.select('v_tuzbe_doc_dl2', view=True)
            download_lista.extend(download_lista2)

        if not download_lista:
            log_type = log_common.copy()
            status = 1
            if error_counter > 0:
                status = 2
            log_type.update({'status': status, 'errors': error_counter})
            self.info_log(log_type)
            return 200

        # We can use a with statement to ensure threads are cleaned up promptly
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.workers) as executor:
            # Start the load operations and mark each future with its URL
            future_to_url = {executor.submit(self.download, dl_data): dl_data for dl_data in download_lista}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    data = future.result()
                except AttributeError:
                    pass
                except Exception as e:
                    exc_type, exc_obj, exc_tb = sys.exc_info()
                    log_copy = log_common.copy()
                    log_copy.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
                    self.error_log(log_copy)
                    error_counter += 1
                    self.download_attrs.update({'error_counter': error_counter})
                else:
                    # parsed data
                    if type(data) is list:
                        if data:
                            all_data.extend(data)

                    # data with errors or doesnt exist
                    if type(data) is dict:
                        update_type = data.get('update')
                        if update_type is not None:
                            all_updates.append(update_type)
                        else:
                            log_copy = log_common.copy()
                            log_copy.update(data)
                            all_errors.append(log_copy)

        print("\n> database")
        e_cnt, d_cnt, u_cnt = 0, 0, 0
        for error in all_errors:
            e_cnt += 1
            stdout.write(f"\r> error {e_cnt}, return_data {d_cnt}, update_data {u_cnt}")
            self.error_log(error)
            error_counter += 1
            self.download_attrs.update({'error_counter': error_counter})

        for return_data in all_data:
            d_cnt += 1
            stdout.write(f"\r> error {e_cnt}, return_data {d_cnt}, update_data {u_cnt}")
            filters, values, insert_data = return_data
            # print("return_data", filters, values, insert_data)
            self.update('EOGLASNA_PRETRAGA', filters, values)
            self.insert('OBJAVA_FILES', insert_data, primary_key='id_objava', schema='tuzbe')

        for update_data in all_updates:
            u_cnt += 1
            stdout.write(f"\r> error {e_cnt}, return_data {d_cnt}, update_data {u_cnt}")
            filters, values = update_data
            # print("update_data",  filters, values)
            self.update('EOGLASNA_PRETRAGA', filters, values)

    def download(self, data):
        if self.flag_120 is True:
            return

        proc_id = threading.current_thread().ident

        id = data.get('id2')
        rbr = data.get('rbr')
        sud_id = data.get('sud_id')
        url = data.get('dokument_url')

        # get today changes
        sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)
        r = self.requests.get(proc_id)
        r.request_type = 2
        r.ses.verify = False

        final_data = []
        try:
            doc_data = r.go(url, download_raw=True)

            if type(doc_data) is not dict:
                if int(doc_data.status_code) > 499:
                    final_data = {'update': ({'rbr': rbr}, {'doc_dl': doc_data.status_code})}
                else:
                    content_disposition, content_type = None, None
                    filename, filename2 = doc_filenames(doc_data.headers)
                    doc_data = doc_data.content

                    # print(f"doc name: {filename}, content_disposition: {filename} content_type: {content_type} filename2: {filename2}")

                    if sud_id == 99999:
                        local_path = local_dir_path_fina
                        server_path = server_dir_path_fina
                    else:
                        local_path = os.path.join(local_dir_path, 'nova', str(sud_id))
                        server_path = os.path.join(server_dir_path, 'nova', str(sud_id))

                    if os.path.exists(local_path) is False:
                        os.makedirs(local_path)

                    # stara, filename_full2, local_path2 = False, None, None
                    # if sud_id in (list(range(99001, 99013+1))):
                    #     local_path2 = os.path.join(local_dir_path, 'stara', str(sud_id))
                    #     server_path2 = os.path.join(server_dir_path, 'stara', str(sud_id))
                    #     if os.path.exists(local_path2) is False:
                    #         os.makedirs(local_path2)
                    #     stara = True

                    if filename is not None:
                        docExtension_idx = filename.rfind(".")
                        if docExtension_idx == -1:
                            print('Nema ekstenzija ' + filename, url)
                            filters = {'rbr': rbr}

                            values = {'doc_name': filename, 'doc_filename': filename2, 'doc_ext': 'no ext', 'doc_dl': 902}
                            final_data = {'update': (filters, values)}

                        else:
                            docExtension = filename[docExtension_idx + 1:]

                            if len(docExtension) < 6:

                                filename_short = str(id) + '.' + docExtension

                                # local dir path
                                filename_full = os.path.join(local_path, filename_short)
                                # local dir path today
                                filename_full_today = os.path.join(local_subdir_path_day, filename_short)
                                # server dir path
                                filename_server_full = os.path.join(server_path, filename_short)

                                # local dir path2
                                # if stara is True and local_path2 is not None:
                                #     filename_full2 = os.path.join(local_path2, filename_short)

                                fileLen = len(doc_data)

                                stdout.write("\r> downloading: {}, file lenght:{}, only_update:{}".format(filename_full, fileLen, self.doc_update))

                                if fileLen > 266:
                                    # writing files
                                    # writing = [filename_full, filename_full_today, filename_server_full]
                                    if self.doc_update is False:
                                        writing = [filename_full, filename_full_today]
                                        for path in writing:
                                            with open(path, 'wb') as f:
                                                f.write(doc_data)

                                        # if stara is True and filename_full2 is not None:
                                        #     with open(filename_full2, 'wb') as f:
                                        #         f.write(doc_data)

                                        hashmd5 = self.filemd5(filename_full)
                                        insert_data = {0: {'date_created': datetime.now(), 'len_download': fileLen,
                                                           'file_download': filename_short, 'httpstatus': 200,
                                                           'id_objava': id, 'hashmd5': hashmd5}}

                                        filters = {'rbr': rbr}
                                        values = {'doc_name': filename, 'doc_filename': filename2, 'doc_ext': docExtension, 'doc_dl': 200}
                                        final_data.append((filters, values, insert_data) )
                                    else:
                                        filters = {'rbr': rbr}
                                        values = {'doc_name': filename, 'doc_filename': filename2, 'doc_ext': docExtension, 'doc_dl': 200}
                                        final_data = {'update': (filters, values)}
                            else:
                                filters = {'rbr': rbr}
                                values = {'doc_name': filename, 'doc_filename': filename2, 'doc_ext': docExtension, 'doc_dl': 901}
                                final_data = {'update': (filters, values)}

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            error_description = str(e)
            print("\n", "error_description", error_description)
            if error_description.find('No space left on device'):
                self.flag_120 = True

        finally:
            # release proxy back in use if it is not released it will be not be available for that web_base next 30min
            self.release_proxy(sha)

        return final_data

    def dohvat_detalja(self):
        start_time = datetime.now()
        step_name = 'dohvat_detalja'
        print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}
        error_counter = 0

        proc_id = 123

        sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)
        r = self.requests.get(proc_id)
        r.request_type = 2
        r.ses.verify = False

        self.connect(4)
        detalji_lista_unique = []
        detalji_lista = self.select('eoglasna_pretraga'.upper(), filters={'detalji': None}, columns=['rbr', 'detalji_objave_link', 'rid'])
        detalji_sha = []
        for dl in detalji_lista:
            sha = hashlib.sha3_256(str(dl).encode()).hexdigest()
            if sha not in detalji_sha:
                detalji_lista_unique.append(dl)
                detalji_sha.append(sha)

        for data in detalji_lista_unique:
            rbr = data.get('rbr')
            rid = data.get('rid')
            detalj_url = data.get('detalji_objave_link')
            stdout.write(f"\r {detalj_url}")
            try:
                if detalj_url is not None:
                    detalji_url = ''.join([base, detalj_url])
                    r.go(detalji_url)
                    html = r.response.content
                    if type(r.response_raw) is dict:
                        continue
                    status_code = r.response_raw.status_code
                    ress_not_found = re.search('Resurs nije pronađen', r.response_raw.content.decode(), re.MULTILINE | re.DOTALL | re.IGNORECASE)
                    if status_code == 404 and ress_not_found is not None :
                        self.update('eoglasna_pretraga'.upper(), filters={'rbr': rbr}, values={'detalji': 404})
                    elif type(html) is dict:
                        err_desc = html.get('RequestError')
                        # print("detalji_url", detalji_url, "load_page", err_desc)
                        self.release_proxy(sha)
                        sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)
                    else:
                        # e-oglasna
                        e_oglasna_detalji = detalji(html, rbr, detalj_url, rid)
                        # print(e_oglasna_detalji)
                        # key za merge je detalj_url
                        e_oglasna_sudionici = detalji_sudionik(html, rbr, detalj_url)

                        self.merge('eoglasna_detalji'.upper(), {0: e_oglasna_detalji}, filters={'rbr': rbr})
                        if e_oglasna_sudionici:
                            # print(e_oglasna_sudionici)
                            self.merge('eoglasna_sudionici'.upper(), e_oglasna_sudionici, filters={'rbr': rbr})
                        self.update('eoglasna_pretraga'.upper(), filters={'rbr': rbr}, values={'detalji': 200})

            except Exception as e:
                print("error", str(e))
                # release proxy back in use if it is not released it will be not be available for that web_base next 30min
                self.release_proxy(sha)
                sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)

        log_type = log_common.copy()
        status = 1
        if error_counter > 0:
            status = 2
        log_type.update({'status': status, 'errors': error_counter})
        self.info_log(log_type)


def manager(command):

    api = Eoglasna()
    if command == 'pretraga':
        api.fetch_last_case()
        api.get_court_list()
        api.dir_creation_check()
        api.search_crawler()
        api.transfer_data()
    if command == 'download':
        status = None
        while status is None and api.flag_120 is False:
            status = api.download_docs()
            if api.flag_120 is True:
                break
    if command == 'detalji':
        api.dohvat_detalja()

    # final runtime log
    api.runtime_log.update({'status': api.runtime_status,'errors': api.runtime_errors})
    api.info_log(api.runtime_log)


if __name__ == '__main__':
    if len(sys.argv) > 1:
        command = sys.argv[1]
    else:
        api = Eoglasna()

        api.fetch_last_case()
        api.get_court_list()
        api.dir_creation_check()
        api.search_crawler()
        api.transfer_data()

        status = None
        loop_counter = 1
        while status is None and api.flag_120 is False:
            status = api.download_docs(loop_counter)
            loop_counter += 1
            if api.flag_120 is True or loop_counter > 10:
                break

        api.dohvat_detalja()

        # final runtime log
        api.runtime_log.update({'status': api.runtime_status,'errors': api.runtime_errors})
        api.info_log(api.runtime_log)