import os
import gc
import re
import sys
import getpass
import threading
from abc import ABC
from time import sleep
import concurrent.futures
from datetime import datetime
from random import randint, randrange
from core_framework.db_engine import DbEngine
from core_framework.crawlers import CrawlerBase
from crawlers_sm.croatia.fina.predstecaj_nagodbe.parser import *

from Scripts import filepath
storage_location = filepath.set_storage()

lock = threading.Lock()

schema = 'fina'
base = 'http://predstecajnenagodbe.fina.hr'
sub_url = ['pn-public-web', 'predmet']
search_url = '/'.join([base, *sub_url, 'search'])
post_url = '/'.join([base, *sub_url, 'search_post'])

os.environ['NLS_LANG'] = 'CROATIAN_CROATIA.UTF8'
api_key = '7c315478c67a270a2d829f1a00f9096c'
site_key = '6LcD6ToUAAAAACaDHK4ZIUS4DuN0X9FPNblQHxiC'
captcha_api = f'http://2captcha.com/in.php?key={api_key}&method=userrecaptcha&googlekey={site_key}&pageurl={search_url}'

# fina_folder = r'C:\Users\{}\Documents\fina_predstec_nagodbe'.format(getpass.getuser())
# fina_folder = r'E:\WebDocs\fina_predstec_nagodbe'
fina_folder = os.path.join(storage_location, 'fina_predstec_nagodbe')
fina_docs_folder = '/'.join([fina_folder, 'documents'])
fina_html_folder = '/'.join([fina_folder, 'html'])
fina_today_html_folder = '/'.join([fina_html_folder, datetime.now().strftime('%Y%m%d')])
server_doc_path = os.path.join('//10.76.15.120', 'predstecaj', 'sve')

crawlers_sm = 9


class OracleDbbisnode(DbEngine):
    def __init__(self):
        DbEngine.__init__(self)
        self.connect(4)


class OracleImenik(DbEngine):
    def __init__(self):
        DbEngine.__init__(self)
        self.connect(7)


class PredstecajneNagodbe(CrawlerBase, ABC):
    def __init__(self):
        CrawlerBase.__init__(self, base)
        self.connect(crawlers_sm, connect_args={"application_name": "croatia/fina/predstecaj_nagodbe/crawler.py"})
        self.orabis = OracleDbbisnode()
        self.oraime = OracleImenik()

        self.tokens = {}
        self.all_done = False

    def dir_creation_check(self, folders=None):
        start_time = datetime.now()
        step_name = 'dir_creation_check'
        if folders is None:
            print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}
        try:
            if folders is None:
                folders = [fina_folder, fina_docs_folder, fina_html_folder, fina_today_html_folder]
            for folder in folders:
                if os.path.isdir(folder) is False:
                    os.mkdir(folder)
            log_common.update({'status': 1, 'errors': 0})
            # self.info_log(log_common)
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            log_common.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
            # self.error_log(log_common)

    def get_token(self, counter):
        proc_id = randint(1, 20000)
        step_name = 'get_token'
        start_time = datetime.now()
        log_common = {'start_time': start_time, 'step_name': step_name, 'gtoken': None, 'session_id': None}

        if len(self.tokens) > 4 or self.all_done is True:
            return 200

        try:
            for i in range(0, 10):
                sha = self.new_request(proc_id=proc_id, protocols=['http'], new=True)
                self.release_proxy(sha)
                r = self.requests.get(proc_id)
                r.request_type = 1
                r.go(captcha_api)
                html = r.response.content

                if type(html) is dict:
                    error = html.get('RequestError')
                    log_copy = log_common.copy()
                    log_copy.update({'err_desc': error, 'status': 2})
                    # self.token_log(log_copy)
                    continue
                else:
                    log_copy = log_common.copy()
                    id_search = re.search('(\d+)', html.decode())
                    if id_search is None:
                        error = {'err_desc': f'id_search is not given by 2captcha. RESPONSE: {html}'}
                        # print(error)
                        # log_copy.update(error)
                        # self.token_log(log_copy)
                        continue

                    request_id = id_search.group(1)
                    log_copy.update({'session_id': request_id})
                    sleep(randrange(15, 20))

                    # get results
                    for x in range(0, 20):
                        if self.all_done is True:
                            return 200

                        request_check = f"http://2captcha.com/res.php?key={api_key}&action=get&id={request_id}"
                        r.go(request_check)
                        html_bs = r.response.content

                        if type(html_bs) is dict:
                            err_desc = html_bs.get('RequestError')
                            # print(err_desc)
                            sleep(randrange(10, 15))
                            continue

                        html = html_bs.text

                        if html == 'CAPCHA_NOT_READY':
                            error = {'err_desc': 'Your captcha is not solved yet. Make 5 seconds timeout and repeat your request.'}
                            # print(error)
                            sleep(randrange(15, 20))
                            continue

                        if html == 'ERROR_CAPTCHA_UNSOLVABLE':
                            error = {'err_desc': '''We are unable to solve your captcha - three of our workers were unable solve it or we didnt get an answer within 90 seconds (300 seconds for ReCaptcha V2). We will not charge you for that request'''}
                            # print(error)
                            break

                        # major account errors
                        if html in ['ERROR_WRONG_USER_KEY', 'ERROR_KEY_DOES_NOT_EXIST']:
                            error = {'err_desc': 'Stop sending requests. Check your API key.'}
                            # print(error)
                            return 400

                        # major recaptcha errors
                        if html in ['ERROR_WRONG_ID_FORMAT', 'ERROR_WRONG_CAPTCHA_ID']:
                            error = {'err_desc': 'Check the ID of captcha or your code that gets the ID.'}
                            # print(error)
                            return 400

                        # minor errors
                        if html in ['ERROR_BAD_DUPLICATES', 'REPORT_NOT_RECORDED']:
                            error = {'err_desc': 'You can retry to send your captcha again.'}
                            # print(error)
                            break

                        # final results
                        if html.startswith('OK'):
                            html = html.replace('OK|','').strip()
                        else:
                            # print("unhandled errors", html)
                            return 400
                        self.tokens.update({request_id: html})
                        return (html, request_id)
            return 400

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            log_copy = log_common.copy()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            log_copy.update(error)
            print(error)
            # self.token_log(log_copy)
        # finally:
            # release proxy back in use if it is not released it will be not be available for that web_base next 30min
            # self.release_proxy(sha)

    def fetch_tokens(self):
        workers = 2
        while 1:
            if self.all_done is True:
                return
            if len(self.tokens) < 5:
                pages = list(range(0, 100))
                with concurrent.futures.ThreadPoolExecutor(max_workers=workers) as executor:
                    # Start the load operations and mark each future with its URL
                    future_to_url = {executor.submit(self.get_token, link): link for link in pages}
                    for future in concurrent.futures.as_completed(future_to_url):
                        url = future_to_url[future]
                        try:
                            return_data = future.result()
                        except AttributeError as e:
                            pass
                        except Exception as e:
                            exc_type, exc_obj, exc_tb = sys.exc_info()
                            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)}
                            print(f'fetch_tokens error {error}')
            else:
                sleep(2)

    def token_server(self):
        print("start token server")
        thread = threading.Thread(target=self.fetch_tokens)
        thread.start()

    def forward_token(self):
        while True:
            token_size = len(self.tokens)
            # sys.stdout.write(f'\r tokens available:  {len(self.tokens)}')
            if token_size > 0:
                break
            sleep(1)

        lock.acquire()
        request_id = list(self.tokens.keys())[0]
        token = self.tokens.get(request_id)
        token_data = (request_id, token)
        self.tokens.pop(request_id)
        lock.release()
        return token_data

    def gather_data(self):
        print("> gather_data")
        records = self.oraime.select('v_predst_pdf_dragan',columns=["SUB_ID", "OIB", "DATUM_ODLUKE", "NAZIV_ODLUKE", "KLASA", "URBROJ"], schema='PH', view=True)
        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            # Start the load operations and mark each future with its URL
            future_to_url = {executor.submit(self.gather_data_thread, record): record for record in records}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    data = future.result()
                except AttributeError as e:
                    pass
                except Exception as e:
                    exc_type, exc_obj, exc_tb = sys.exc_info()
                    error ={'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)}
                    print("gather_data", error)
                else:
                    if type(data) is dict:
                        print('err', data)
        self.all_done = True

    def gather_data_thread(self, record):
        sha = None
        oib = record.get('oib')
        proc_id = threading.current_thread().ident
        print(proc_id, record)
        request_id, g_token = self.forward_token()
        print(g_token)
        try:
            sleep(randrange(5, 8))
            sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)
            r = self.requests.get(proc_id)

            r.request_type = 1
            self.release_proxy(sha)

            r.go(search_url)
            main_cookies = r.ses.cookies.get_dict()
            html = r.response.content

            if type(html) is dict:
                err_desc = html.get('RequestError')
                print('gather_data_thread', err_desc)
                return err_desc

            hdiv_state = html.find('input', {'name': '_HDIV_STATE_', 'value': True})
            if hdiv_state is None:
                return {'err_desc': 'cannot find input element with name hdiv_state'}
            hdiv_state = hdiv_state['value']

            payload = {'oib': oib, 'name': '', 'since': '', 'until': '', '_includeOnlyOpenCases': 0,
                       'g-recaptcha-response': g_token, 'reCaptchaResponse': '',
                       '_HDIV_STATE_': hdiv_state}

            r.ses.headers.update({
                'Host': 'predstecajnenagodbe.fina.hr',
                'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
                'Accept-Language': 'en-GB,en;q=0.5',
                'Content-Type': 'application/x-www-form-urlencoded',
                'Origin': 'http://predstecajnenagodbe.fina.hr',
                'Referer': 'http://predstecajnenagodbe.fina.hr/pn-public-web/predmet/search',
                'Cookie': f'JSESSIONID={main_cookies.get("JSESSIONID")}; TS01398827={main_cookies.get("TS01398827")}',
                'Upgrade-Insecure-Requests': '1'})

            r.request_type = 3
            r.payload = payload
            r.go(post_url)
            html = r.response.content

            if type(html) is dict:
                print("post err", html)
                return

            # spremanje htmla
            # with open(f'predstec.html', 'wb') as f:
            #     f.write(r.response.content_raw)

            table_data = parse_table(html)
            if type(table_data) is dict:
                print("error occured during parsing search results")
                return
            table_data = table_data[0]

            pop_header_keys = ['Content-Type', 'Origin']
            for pop_header_key in pop_header_keys:
                r.ses.headers.pop(pop_header_key)

            for ur_number, class_data in table_data.items():
                search_data = class_data.search_data
                details_url = class_data.details_url
                details = ''.join([base, details_url])
                print(details)

                r.request_type = 1
                r.go(details)
                html = r.response.content

                if type(html) is dict:
                    print("fetching details error", html)
                    return

                vat_number = search_data.get('vat_number')
                ur_number_file = re.sub("\s+", "", ur_number.replace('-', '').replace("/",''))
                filename = f'{vat_number}_{ur_number_file}_{datetime.now().hour}_{datetime.now().minute}.html'
                html_file_path = os.path.join(fina_today_html_folder, filename)
                with open(html_file_path, 'wb') as f:
                    f.write(r.response.content_raw)

                paresed_data = parse_details(html)
                if type(paresed_data) is dict:
                    print("error during parsing details", paresed_data)
                    return

                details_data, documents_data = paresed_data
                print(details_data)

                vat_number = details_data.get('vat_number')
                doc_class = details_data.get('doc_class')
                filters = {'doc_class': doc_class, 'vat_number': vat_number}
                merge_status = self.merge('predstec_details', {0: details_data}, filters=filters, schema=schema)
                if merge_status != 200:
                    print("error during merge data in table predstec_details")

                # document download
                for row in documents_data:
                    row_copy = row.copy()
                    doc_id = row.get('doc_id')
                    vat_number = row.get('vat_number')
                    doc_class = row.get('doc_class')
                    doc_name = f'{vat_number}_{doc_id}.pdf'
                    doc_locdir_path = os.path.join(fina_docs_folder, vat_number[:2])
                    doc_locfile_path = os.path.join(doc_locdir_path, doc_name)

                    doc_servdir_path = os.path.join(server_doc_path, vat_number[:2])
                    doc_servfile_path = os.path.join(doc_servdir_path, doc_name)

                    self.dir_creation_check(folders=[doc_locdir_path, doc_servdir_path])

                    # check existence i database before proceeding with dl
                    filters = {'doc_id': doc_id, 'vat_number': vat_number, 'doc_class': doc_class, 'archive': None}
                    check_in_db = self.select('predstec_docs', filters=filters, schema=schema)
                    if not check_in_db:
                        doc_url = row.get('details_url')
                        if doc_url is not None:
                            doc_url = ''.join([base, doc_url])
                            row_copy.pop('details_url')

                            r.request_type = 1
                            data = r.go(doc_url, download=True)
                            if type(data) is dict:
                                print("error occured during download od document", data)
                                continue

                            doc_size = len(data)
                            doc_sha = hashlib.sha3_256(data).hexdigest()
                            row_copy.update({'doc_path': doc_locfile_path, 'doc_size': doc_size, 'doc_sha': doc_sha, 'up_path': doc_servfile_path})

                            print(row_copy)
                            check_loc_existance = os.path.exists(doc_locfile_path)
                            if check_loc_existance is False:
                                with open(doc_locfile_path, 'wb') as f:
                                    f.write(data)

                            check_serv_existance = os.path.exists(doc_servfile_path)
                            if check_serv_existance is False:
                                try:
                                    with open(doc_servfile_path, 'wb') as f:
                                        f.write(data)
                                except:
                                    # when server is not available or full
                                    pass

                            merge_status = self.merge('predstec_docs', {0: row_copy}, on=['sha'], filters=filters, schema=schema)
                            if merge_status != 200:
                                print("error during merge data in table predstec_docs")

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
        # finally:
            # release proxy back in use if it is not released it will be not be available for that web_base next 30min
            # self.release_proxy(sha)

    def upload_docs(self):
        print("> upload_docs")
        # rechecks all records that have up_status null if they exist on server side
        filters = {'up_status': None, 'archive': None}
        updoc_check = self.select('predstec_docs', filters=filters, schema=schema)
        for row in updoc_check:
            up_path = row.get('up_path')
            doc_path = row.get('doc_path')

            doc_id = row.get('doc_id')
            vat_number = row.get('vat_number')
            doc_class = row.get('doc_class')
            filters = {'doc_id': doc_id, 'vat_number': vat_number, 'doc_class': doc_class}

            check_servfile_existance = os.path.exists(up_path)
            if check_servfile_existance is True:
                self.update('predstec_docs', filters, {'up_status': 1, 'up_date': datetime.now()}, schema=schema)
            else:
                try:
                    with open(doc_path, 'rb') as f:
                        data = f.read()
                        with open(up_path, 'wb') as f:
                            f.write(data)
                    self.update('predstec_docs',  filters, {'up_status': 1, 'up_date': datetime.now()}, schema=schema)
                except:
                    # when server is not available or full
                    pass

    def transfer_data(self):
        print("> transfer_data")
        filters = {'transfer': None, 'archive': None}
        transfer_rows = self.select('predstec_docs', filters=filters, schema=schema)
        for transfer_row in transfer_rows:

            doc_id = transfer_row.get('doc_id')
            vat_number = transfer_row.get('vat_number')
            doc_class = transfer_row.get('doc_class')
            filters = {'urbroj': doc_id, 'oib': vat_number, 'klasa': doc_class}
            row_exist = self.oraime.select('predstecajna_dokumentacija', filters=filters, schema='PH')

            filters = {'doc_id': doc_id, 'vat_number': vat_number, 'doc_class': doc_class}
            if not row_exist:
                transfer_cols = {'doc_id': 'urbroj', 'vat_number': 'oib', 'doc_class': 'klasa', 'up_path': 'path', 'doc_type': 'naziv', 'doc_date': 'datumvrijeme'}
                transfer_row_copy = transfer_row.copy()
                transfer_row_copy = {transfer_cols.get(k):v for k,v in transfer_row_copy.items() if k in list(transfer_cols.keys())}
                path_col = transfer_row_copy.get('path').replace("//","\\\\")
                transfer_row_copy.update({'datum_upisa': datetime.now(), 'path': path_col})
                insert_status = self.oraime.insert('predstecajna_dokumentacija'.upper(), {0: transfer_row_copy}, schema='PH')
                if insert_status == 200:
                    self.update('predstec_docs', filters, {'transfer': 1, 'transfer_date': datetime.now()}, schema=schema)
            else:
                self.update('predstec_docs', filters, {'transfer': 1, 'transfer_date': datetime.now()}, schema=schema)


if __name__ == '__main__':
    while 1:
        print("Predstecajne Nagodbe")
        api = PredstecajneNagodbe()
        api.dir_creation_check()
        api.token_server()
        api.gather_data()
        api.upload_docs()
        api.transfer_data()
        gc.collect()
        sleep(3600)
