import os
import sys
import pickle
import getpass

from sqlalchemy import *
from subprocess import call
from sqlalchemy import event
from subprocess import Popen, PIPE
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base

if sys.version_info.major > 2:
    from core_framework.db_engine import DbEngine

schema = 'fina'

framework_folder = r'C:\Users\{}\Documents\crawler_framework'.format(getpass.getuser())
python_path = r'{}\py_config.pkl'.format(framework_folder)

BasePstg = declarative_base()

# creating schema if it doesn't exist
event.listen(BasePstg.metadata, 'before_create', DDL("CREATE SCHEMA IF NOT EXISTS {}".format(schema)))


# ======================================================================
#                               TABLES
# ======================================================================
class PredstecDetails(BasePstg):
    """All gathered detail results for specified vat_number"""
    __tablename__ = 'predstec_details'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('predstdet_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64))  # hash value of the row
    doc_class = Column(String(256), index=True)
    procedure = Column(String(1480))
    conditions = Column(String(2560))
    amount = Column(Numeric(18, 2))
    open_date = Column(DateTime())
    council = Column(String(4000))
    rc = Column(String(1480))
    vat_number = Column(String(20), index=True)
    cname = Column(String(4000))
    person_name = Column(String(4000))
    hq = Column(String(4000))
    business = Column(String(4000))
    residence = Column(String(4000))
    workers = Column(String(20))
    jurisdiction = Column(String(4000))
    phone = Column(String(4000))
    email = Column(String(500))

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class PredstecDocs(BasePstg):
    """All gathered documents for specified vat_number"""
    __tablename__ = 'predstec_docs'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('predstdoc_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        BigInteger, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    sha = Column(String(64))  # hash value of the row
    vat_number = Column(String(20), index=True)
    doc_class = Column(String(256), index=True)
    doc_id = Column(String(256), index=True)
    doc_type = Column(String(4000))
    doc_date = Column(DateTime())
    doc_size = Column(Integer())
    doc_sha = Column(String(64))
    doc_path = Column(String(2480))

    up_status = Column(Integer(), index=True)
    up_path = Column(String(2480))
    up_date = Column(DateTime)

    transfer = Column(Integer(), index=True)
    transfer_date = Column(DateTime)

    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


def deploy(conn_id):
    db = DbEngine()
    engine = db.connect(conn_id, echo=False)
    BasePstg.metadata.drop_all(engine)
    BasePstg.metadata.create_all(engine)


if __name__ == '__main__':
    deploy(1)