import re
import hashlib
import cyrtranslit
from time import sleep
from bs4 import BeautifulSoup
from datetime import datetime


def extract_links(html_raw, page):
    links = {}
    html = BeautifulSoup(html_raw, 'lxml')
    articles = html.find_all('article')
    for article in articles:
        link = article.find('a', {'href': True})
        if link is not None:
            link = link['href']
            link_reverse = link[::-1]
            link_id = link_reverse[0:link_reverse.find("/")][::-1].strip()
            links.update({len(links): {"page_link": link, 'link_id': int(link_id), 'page': page}})
    return links


def parse(html, link_id):
    html_lat = cyrtranslit.to_latin(html.decode())
    soup = BeautifulSoup(html_lat, 'lxml')
    section = soup.find('section', {'class': 'single_bcp'})
    [s.extract() for s in section('h3')]
    [s.extract() for s in section('div')]
    [s.extract() for s in section('a')]

    span_split = section.decode().split('<span')
    span_split = ['<span' + ss for ss in span_split if not ss.startswith('<')]
    for cleaning in ['</section>', '<hr/>']:
        span_split = [re.sub(cleaning, '', ss, re.MULTILINE | re.DOTALL | re.IGNORECASE) for ss in span_split]
    section = "\n".join(span_split) + "\n"

    records = re.findall(r'<span\s+class="info_title">(.*?)</span>(.*?)\n', section, re.MULTILINE | re.DOTALL | re.IGNORECASE)
    replace = {"Puno ime": 'company_name', 'Sud': 'court', 'Broj sudskog rešenja': 'decision', 'Datum otvaranja': 'decision_date',
               'Delatnost': 'activity', 'Matični broj': 'registry_number', 'Status stečajnog postupka': 'bankruptcy_status',
               'Opština': 'municipality', 'Adresa': 'address', 'Grad': 'city', 'Pravna forma': 'legal_form',
               'PIB': 'vat_number', 'Stečajni sudija': 'judge', 'Stečajni upravnik': 'manager'}

    final_data = {}
    pib = None

    for record in records:
        header, value = record
        header = header.strip(":").strip()
        header = replace.get(header)
        value = value.strip('\xa0').strip()
        if header == 'vat_number':
            pib = value
        if header is not None:
            if header == 'decision_date':
                date_format = re.search(r'\d+\.\d+\.\d+', value)
                if date_format is not None:
                    value = datetime.strptime(value, '%d.%m.%Y')
            final_data.update({header: value})

    if pib is not None:
        sha = hashlib.sha3_256(str(final_data.values()).encode()).hexdigest()
        final_data.update({'sha': sha, 'link_id': str(link_id)})
        return final_data

    return 400
