import os
import re
import gc
import sys
import zeep
import getpass
import random
import hashlib
import urllib3
import requests
import threading
import pandas as pd
from abc import ABC
from time import sleep
from zeep import Client
import concurrent.futures
from random import shuffle
from requests import Session
from bs4 import BeautifulSoup
from urllib3.util import parse_url
from zeep.transports import Transport
from datetime import datetime, timedelta
from zeep.wsse.username import UsernameToken
from core_framework.db_engine import DbEngine
from core_framework.request import Request
from core_framework import user_agent
from core_framework.crawlers import CrawlerBase
from crawlers_sm.serbia.apr.companies import parser
from crawlers_sm.serbia.apr.companies.codebook import change_ids
from crawlers_sm.serbia.apr.companies import deploy

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# works in production only
# from Scripts import filepath
# storage_location = filepath.set_storage()

# works in production and test env
storage_file = "storage.txt"
crawlers_sm_config_dir = r'C:\Users\{}\Documents\csm_conf'.format(getpass.getuser())
storage_path = os.path.join(crawlers_sm_config_dir, storage_file)
with open(storage_path, 'r') as f:
    storage_path = f.read()

# ------ paths
apr_dl_path = os.path.join(storage_path, 'apr')
apr_companies_path = os.path.join(apr_dl_path, 'companies')
apr_c_html_path = os.path.join(apr_companies_path, 'html')
apr_decisions = os.path.join(apr_companies_path, 'decisions')
folders = [storage_path, apr_dl_path, apr_companies_path, apr_c_html_path, apr_decisions]


log = False
debug = True
printing = True
schema = 'apr'
db_conn_id = None
base = 'apr.gov.rs'
lock = threading.RLock()

base_url = 'http://pretraga2.apr.gov.rs'
main_url = '/'.join([base_url,'unifiedentitysearch'])
post_url = '/'.join([base_url,'ObjedinjenePretrage/Search/SearchResult'])
company_types = ['Enterprise', 'Enterpreneur', 'SportAssociation', 'AssociationPublic', 'ForeignAssociation', 'Foundation', 'InsolventEstate', 'ForeignFAE']
skip_tabs = ['InsolvencyDocuments', 'PublishedDocuments', 'AnnouncementLiquidation', 'Statues',
             'LiquidationAnnouncement', 'BranchesAndObjectives', 'ObjavljeniDokumenti']


class DbAprConnection(DbEngine):
    def __init__(self):
        DbEngine.__init__(self, primary_key='id')
        self.connect(3)


class DbRsConnection(DbEngine):
    def __init__(self):
        DbEngine.__init__(self, primary_key='ENTRY_ID')
        self.connect(2)


def pprint(text):
    if printing is True:
        print(text)


class ParsedData:
    def __init__(self):
        self.search_info = dict()
        self.basic_info = dict()
        self.activity_pause = dict()
        self.address_data = dict()
        self.annotations = dict()
        self.association_extra = dict()
        self.bank_accounts = dict()
        self.bankruptcy = dict()
        self.branch_offices = dict()
        self.business_data = dict()
        self.business_name = dict()
        self.capital = dict()
        self.capital_amounts = dict()
        self.capital_groups = dict()
        self.capital_shares = dict()
        self.company_people = dict()
        self.decisions = dict()
        self.email = dict()
        self.erase_details = dict()
        self.fax = dict()
        self.limited_duration = dict()
        self.phone = dict()
        self.web_page = dict()


class AprCompanies(CrawlerBase, ABC):
    def __init__(self):
        CrawlerBase.__init__(self, base)
        # self.connect(1, connect_args={"application_name": "crawlers_sm/apr/crawler.py/AprCompanies"})  # test
        self.connect(9, connect_args={"application_name": "crawlers_sm/apr/crawler.py/AprCompanies"})  # production

        # runtime log initialization
        self.runtime_errors = 0
        self.runtime_status = 1
        self.runtime_start = datetime.now()
        self.runtime_id = hashlib.sha3_256(str(self.runtime_start).encode()).hexdigest()
        self.runtime_log = {'start_time': self.runtime_start, 'step_name': 'runtime'}

        # webservice initializations
        self.ws_session = Session()
        self.ws_session.verify = False
        self.client = self.create_ws_client()

        self.apr_conn = DbAprConnection()
        self.rs_conn = DbRsConnection()

        # webpage
        self.inuse = dict()
        self.inuse_id = list()
        self.apr_accounts = dict()

    def info_log(self, data):
        sha = hashlib.sha3_256(str(data).encode()).hexdigest()
        data.update({'sha': sha, 'runtime_id': self.runtime_id, 'end_time': datetime.now()})
        print(f"info_log: {data}")
        if log is True:
            self.insert('l_info', {0: data}, schema=schema)

    def error_log(self, data):
        self.runtime_status = 2
        self.runtime_errors += 1
        sha = hashlib.sha3_256(str(data).encode()).hexdigest()
        data.update({'sha': sha, 'runtime_id': self.runtime_id, 'end_time': datetime.now()})
        print(f"error_log: {data}")
        if log is True:
            self.insert('l_errors', {0: data}, schema=schema)

    def dir_creation_check(self, new_folder=None):
        start_time = datetime.now()
        step_name = 'dir_creation_check'
        log_common = {'start_time': start_time, 'step_name': step_name}
        try:
            if new_folder is not None:
                if os.path.isdir(new_folder) is False:
                    os.mkdir(new_folder)
            else:
                for folder in folders:
                    if os.path.isdir(folder) is False:
                        os.mkdir(folder)
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            log_common.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
            self.error_log(log_common)

    def html_loader(self):
        for registry_number in os.listdir(apr_c_html_path):
            dates = os.listdir(os.path.join(apr_c_html_path, registry_number))
            for date in dates:
                datet = datetime.strptime(date, '%Y%m%d')
                date_data = os.listdir(os.path.join(apr_c_html_path, registry_number, date))
                for dd in date_data:
                    doc_path = os.path.join(apr_c_html_path, registry_number, date, dd)
                    if dd != 'SearchInfo.html':
                        tab_name, registry = dd.split('_')
                        registry = registry.replace('.html', '')
                        proc_date = pd.to_datetime(datet.strftime('%Y-%m-%d'), yearfirst=True)
                        data = {'registry_number': registry_number, 'tab_name': tab_name, 'registry': registry, 'proc_date': proc_date}
                        row_sha = parser.row_sha(data, cols=['registry_number', 'tab_name', 'proc_date'])
                        data.update({'sha': row_sha})
                        with open(doc_path, 'rb') as f:
                            html_size = len(f.read())
                        data.update({'sha': row_sha, 'dl_path': doc_path, 'html_size': html_size, 'dl_date': datetime.now(), 'dl_stat': 1})
                        print(data)
                        merge_status = self.merge('html_download', {0: data}, filters={'registry_number': registry_number}, on=['sha'], update=False, schema=schema)
                        if merge_status != 200:
                            print("merge status failed in html_loader")
                            if debug:
                                exit()
                            else:
                                continue

    def create_ws_client(self):
        start_time = datetime.now()
        step_name = 'create_ws_client'
        log_common = {'start_time': start_time, 'step_name': step_name}
        try:
            # webservice client creation
            transport = Transport(session=self.ws_session)
            user_token = UsernameToken('plws.bisnode', ':v!-aQ')
            wsdl_server = 'https://service1.apr.gov.rs:4430/plws/PlService.svc?wsdl'
            wsdl_testserver = 'https://service1.apr.gov.rs:4430/plwstest/PlService.svc?wsdl'
            client = Client(wsdl_server, transport=transport, wsse=user_token)
            client.transport = Transport(session=self.ws_session)
            log_common.update({'status': 1, 'errors': 0})
            self.info_log(log_common)
            return client

        except zeep.exceptions.Fault as fault:
            error = f"zeep exception: msg={fault.message} | code={fault.code} | detail={fault.detail}"
            log_common.update({'err_desc': error})
            self.error_log(log_common)
            return 404

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            log_common.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
            self.error_log(log_common)
            return 404

    def update_apr_codebooks(self):
        start_time = datetime.now()
        step_name = 'update_apr_codebooks'
        log_common = {'start_time': start_time, 'step_name': step_name}
        try:
            # fetch codebooks from ws
            codebook_list = {'SviPodaciOTipovimaPoslovnogSubjekta':  deploy.CompanyTypes.__tablename__,
                             'SviPodaciOTipovimaNacinaPromene': deploy.CompanyChangeModes.__tablename__,
                             'SviPodaciOTipovimaRegistracionogPostupka': deploy.CompanyRegProcTypes.__tablename__}

            for codebookname, dbtablename in codebook_list.items():
                codebook = getattr(self.client.service, codebookname)
                data = codebook()
                parsed_data = parser.codebooks_type1(data, codebookname)

                if parsed_data:
                    merge_status = self.merge(dbtablename, parsed_data, filters={}, delete=True, schema=schema)
            log_common.update({'status': 1, 'errors': 0})
            self.info_log(log_common)

        except zeep.exceptions.Fault as fault:
            error = f"zeep exception: msg={fault.message} | code={fault.code} | detail={fault.detail}"
            log_common.update({'err_desc': error})
            self.error_log(log_common)

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            log_common.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
            self.error_log(log_common)

    def get_company_changes(self, date_from, date_to):
        start_time = datetime.now()
        step_name = 'get_company_changes'
        log_common = {'start_time': start_time, 'step_name': step_name}
        try:
            pppps = self.client.service.PreuzmiPromenePodatakaOPrivrednimSubjektima
            data = pppps({'datumOd': date_from.isoformat(), 'datumDo': date_to.isoformat()})

            # parsing data
            parsed_data = parser.company_changes(data)

            # fetch from database exiting data from last 8 days
            sql = f"select * from apr.company_changes where change_date >='{date_from}' and change_date <='{date_to}' "
            db_sha = self.select('company_changes', sql=sql, schema=schema)
            db_sha = [row.get('sha') for row in db_sha]

            # remove data that was already inserted in last 8 days to the database
            filtered_data = {}
            [filtered_data.update({len(filtered_data): row}) for row in parsed_data if not row.get('sha') in db_sha]

            # insert new data
            self.insert('company_changes', filtered_data, schema=schema)
            log_common.update({'status': 1, 'errors': 0})
            self.info_log(log_common)

        except zeep.exceptions.Fault as fault:
            error = f"zeep exception: msg={fault.message} | code={fault.code} | detail={fault.detail}"
            log_common.update({'err_desc': error})
            self.error_log(log_common)

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            log_common.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
            self.error_log(log_common)

    def acquire_account(self, thread_number):
        with lock:
            # filter available accounts
            acc_available = [row for row in self.apr_accounts if row.get('sha') not in self.inuse_id]
            if len(acc_available) == 0:
                pprint("there is no available apr accounts")
                return {'err_desc': "there is no available apr accounts"}

            account_data = acc_available[random.randint(0, len(acc_available))]
            account_id = account_data.get('sha')
            self.inuse_id.append(account_id)
            self.inuse.update({thread_number: account_data})

    def release_account(self, thread_number):
        with lock:
            thread_data = self.inuse.get(thread_number)
            account_id = thread_data.get('sha')
            self.inuse.pop(thread_number)
            self.inuse_id = [x for x in self.inuse_id if x != account_id]

    def start_crawl(self, registry_number=None, decisions=False):
        # get list of registry numbers for processing
        # companies = self.select('company_changes', filters={'processed': None},  schema=schema)
        companies = self.select('company_changes', sql="select * from apr.company_changes where processed is null limit 15")
        minutes_ago = datetime.now() - timedelta(minutes=20)
        update_status = self.update('search_info', {'link_status': 1, 'link_date': f'<={minutes_ago}'}, {'link_status': 0}, schema=schema)
        companies = self.select('search_info', filters={'archive': None}, columns=['sha', 'registry_number'], schema=schema)

        if registry_number is not None:
            companies = self.select('company_changes', filters={'registry_number': registry_number}, columns=['registry_number'], schema=schema)

        if decisions is True:
            companies = self.select('decisions', filters={'doc_link': None}, columns=['registry_number'], schema=schema)

        if registry_number is not None and decisions is True:
            companies = self.select('decisions', filters={'doc_link': None, 'registry_number': registry_number}, columns=['registry_number'], schema=schema)

        self.apr_accounts = self.select('c_apr_accounts',  schema=schema)
        shuffle(self.apr_accounts)

        # mt crawl data
        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            future_to_company = {executor.submit(self.crawl_data, company, decisions): company for company in companies}
            for future in concurrent.futures.as_completed(future_to_company):
                url = future_to_company[future]
                try:
                    data = future.result()
                except Exception as exc:
                    print('%r generated an exception: %s' % (url, exc))
                else:
                    if data is not None:
                        print('%r page is %d bytes' % (url, len(data)))

        # cleaning account usage history
        self.inuse, self.inuse_id = dict(), list()

    def search_info(self, registry_number, html, reg_date_folder, html_raw, r):
        # parsing and preparing data for database
        search_info_parser = parser.SearchInfo(registry_number, html)
        search_info_status = search_info_parser.parse()
        search_info_parsing_codes = search_info_parser.parsing

        if search_info_status != 200:
            return

        # registry_number, tab_name, registry, search_url, html, reg_date_folder,html_name, proxy_sha, r
        registry = search_info_parser.registry
        html_name = f'SearchInfo_{registry}'
        status_int = self.html_downloads(registry_number, 'SearchInfo', registry, None, html_raw, reg_date_folder, html_name, None, r)
        if status_int != 200:
            return

        # insert parsing results for search info
        si_sha = parser.row_sha({'data': str(list(sorted(search_info_parsing_codes)))})
        parsing_stats_data = {'registry_number': registry_number, 'parsing_stats': search_info_parser.parsing, 'sha': si_sha, 'tab_name': 'search_info'}
        insert_status = self.insert('parsing_results', {0: parsing_stats_data}, schema=schema)
        if insert_status != 200:
            error = {'err_desc': 'An error ocured during inserting into parsing_results for search_info'}
            self.error_log(error)
            return

        # check changes regarding blackout status of company we are searching
        blackout_check = self.select('blackout', filters={'registry_number': registry_number, 'archive': None}, schema=schema)
        if blackout_check is None:
            error = {'err_desc': 'Blackout, error during selecting from table'}
            self.error_log(error)
            return

        # if there is open, non-archived record in table and on webpage blackout is false, we will archive current open record in our table
        if len(blackout_check) > 0 and search_info_parser.blackout == 0:
            update_status = self.update('blackout', {'registry_number': registry_number, 'archive': None}, {'archive': datetime.now()}, schema=schema)
            if update_status != 200:
                error = {'err_desc': 'Blackout, error during updating record in table'}
                self.error_log(error)
                return

        # if there is no open record in table and on webpage is true, we will add new record for blackout
        if len(blackout_check) == 0 and search_info_parser.blackout == 1:
            insert_status = self.insert('blackout', {0: {'registry_number': registry_number}}, schema=schema)
            if insert_status != 200:
                error = {'err_desc': 'Blackout, error during inserting record in table'}
                self.error_log(error)
                return

        # if search_info_status is 400 that means that parsing didn't went well
        if search_info_status == 400:
            error = {'err_desc': "bad parsing - could not parse search info data"}
            self.error_log(error)
            return

        # if all went fine with previous steps regarding search info insert data
        merge_status = self.merge('search_info', {0: search_info_parser.data}, filters={'registry_number': registry_number, 'archive': None}, schema=schema, on=['sha'])
        if merge_status != 200:
            error = {'err_desc': 'search_info error occured during merging data'}
            self.error_log(error)
            return

        filters = {'registry_number': registry_number, 'archive': None}
        values = {'date_checked': datetime.now(), 'web_link': search_info_parser.details_link, 'link_date': datetime.now(), 'link_status': 1}
        update_status = self.update('search_info', filters, values, schema=schema)
        print(f"update_status search info link status {update_status}")

        return search_info_parser

    def html_downloads(self, registry_number, tab_name, registry, search_url, html, reg_date_folder,html_name, proxy_sha=None, r=None):
        if proxy_sha is not None and r is not None:
            html_raw = r.response.content_raw
        else:
            html_raw = html

        # check htmls for today in database
        proc_date = pd.to_datetime(datetime.now().strftime('%Y-%m-%d'), yearfirst=True)
        filters = {'registry_number': registry_number, 'proc_date': proc_date, 'tab_name': tab_name}
        reg_data = self.select('html_download', filters=filters, schema=schema)

        data = {'registry_number': registry_number, 'tab_name': tab_name, 'registry': registry, 'current_link': search_url, 'proc_date': proc_date}
        row_sha = parser.row_sha(data, cols=['registry_number', 'tab_name', 'proc_date'])
        data.update({'sha': row_sha})

        if not reg_data:
            # if it does not exist then add record
            self.merge('html_download', {0: data}, filters=filters, update=False, schema=schema)

        if type(html) is dict or re.search('HTTP/1.1 400 Bad Request', str(html)) is not None:
            err_desc = html.get('RequestError')
            error = {'err_desc': f"error happened trying fetching {tab_name} {err_desc}"}
            self.error_log(error)
            if proxy_sha is not None:
                self.bad_proxy(proxy_sha, r.proxy_data, web_page=base_url)
            return 400

        html_name = os.path.join(reg_date_folder, f"{html_name}.html")
        with open(html_name, 'wb') as f:
            f.write(html_raw)

        html_size = len(html_raw)
        data = {'dl_path': html_name, 'html_size': html_size, 'proc_date': proc_date, 'dl_date': datetime.now(), 'dl_stat': 1}
        self.update('html_download', {'sha': row_sha}, data, schema=schema)
        return 200

    def requests_loop(self, url, thread_number, html_name, reg_date_folder, registry_number,
                      tab_name, registry,
                      headers=None, request_type=2, payload=None, retries=3):
        locked_proxies = []
        for i in range(0, retries+1):
            sleep(random.randint(2, 5))
            sha = self.new_request(proc_id=thread_number, protocols=['http'], new=True)
            locked_proxies.append(sha)
            try:
                sleep(random.randint(5, 10))
                r = self.requests.get(thread_number)
                r.request_type = request_type
                print(r.ip, r.port, url)
                if headers is not None:
                    r.ses.headers.update(headers)
                if payload is not None:
                    r.payload = payload

                if html_name == 'doc':
                    resp = r.go(url, download_raw=True)
                    if type(resp) is dict:
                        self.bad_proxy(sha, r.proxy_data, web_page=base_url)
                        self.release_proxy(sha)
                        continue

                    status_code = resp.status_code
                    print(status_code)

                    # if link is not valid then remove doc link and set to regular download
                    if re.search(b'\xd0\x93\xd1\x80\xd0\xb5\xd1\x88\xd0\xba\xd0\xb0!', resp.content, re.MULTILINE) is not None:
                        return url, 404

                    if re.search(b'An error occurred while', resp.content, re.MULTILINE) is not None:
                        return url, 404

                    # checking for other errors that could be caused by proxy
                    errors = [b'The requested URL could not be retrieved', b'Proxy Error', b'Bad Gateway', b'Sorry',
                              b'Too Many Requests', b'Internal Server Error']
                    err_stat = 0
                    for error in errors:
                        if re.search(error, resp.content, re.MULTILINE) is not None:
                            err_stat = 1
                            break

                    if err_stat == 1:
                        self.bad_proxy(sha, r.proxy_data, web_page=base_url)
                        self.release_proxy(sha)
                        continue

                    if status_code != 200:
                        self.bad_proxy(sha, r.proxy_data, web_page=base_url)
                        self.release_proxy(sha)
                        continue

                    ext = 'doc'
                    content_disposition = resp.headers.get('Content-Disposition')
                    if content_disposition is not None:
                        dot_position = content_disposition.rfind('.')
                        if dot_position > 0:
                            ext = content_disposition[dot_position+1:]

                    html = resp.content
                    sha_doc = hashlib.sha3_256(str(html).encode(errors='ignore')).hexdigest()

                    html_name = os.path.join(reg_date_folder, f"{sha_doc}.{ext}")
                    with open(html_name, 'wb') as f:
                        f.write(html)

                    doc_size = len(html)
                    doc_data = {'doc_ext': ext, 'doc_size': doc_size, 'doc_dl': 1,
                                'doc_link': url, 'sha_doc': sha_doc, 'doc_loc': html_name}
                    return url, doc_data

                if html_name != 'doc':
                    r.go(url)
                    html = r.response.content
                    search_url = r.response.url
                    # registry_number, tab_name, registry, search_url, html, reg_date_folder,html_name, proxy_sha, r
                    status_int = self.html_downloads(registry_number, tab_name, registry, search_url, html, reg_date_folder, html_name, sha, r)
                    if status_int != 200:
                        continue

                    return search_url, r.response.content_raw

            # except Exception as e:
            #     print("requests_loop")
            #     exc_type, exc_obj, exc_tb = sys.exc_info()
            #     print({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
            finally:
                self.release_proxy(sha)
        return

    def details_crawl(self, details_link, thread_number, registry_number, registry, reg_date_folder, approach='account', decisions=False):
        print(f"approach over {approach} for details link {details_link}")
        sleep(random.randint(2, 5))

        # gathering data for already downloaded htmls for this day
        filters = {'registry_number': registry_number,
                   'proc_date': datetime.strptime(datetime.now().strftime('%Y-%m-%d'), '%Y-%m-%d'),
                   'dl_stat': True}
        not_parsed = self.select('html_download', filters=filters, schema=schema)
        yes_tab_names = [row.get('tab_name') for row in not_parsed]

        # getting basic info
        status_request = self.requests_loop(details_link, thread_number, f'BasicInfo_{registry}', reg_date_folder, registry_number, 'BasicInfo', registry)

        search_url, html = None, None
        if status_request is not None:
            search_url, html = status_request

        if html is None:
            error = {"err_desc": "Details link  could not be processed"}
            self.error_log(error)
            return

        # parsing data that can be found in basic info tab
        basic_info_parser = parser.BasicInfo(registry_number, html, registry)
        links = basic_info_parser.parse()
        basic_info_data = basic_info_parser.data
        pprint(f"basic_info_data {basic_info_data}")

        if type(links) is dict:
            self.error_log(links)
            return

        shuffle(links)
        pprint(f"links {links }")

        for link in links:
            try:
                for tab_name, tab_link in link.items():
                    # skipping tabs
                    if tab_name in skip_tabs:
                        continue

                    if tab_name in yes_tab_names:
                        if tab_name == 'Decrees' and decisions is True:
                            pass
                        else:
                            pprint(f"html for {tab_name} already downloaded")
                            continue

                    if re.search('ForeignFAE', details_link) is not None:
                        tab_link = tab_link.replace('..', '/ForeignFAEPublicWebSearch')

                    url = f"{base_url}{tab_link}"
                    pprint("{}".format(url))

                    # checking if the link is in known tabs for parsing
                    company_type = None
                    for ct in company_types:
                        if url.__contains__(ct):
                            company_type = ct

                    # if tab name is not recognised log error
                    if company_type is None:
                        error = {"err_desc", f"Parser not ready for this tab {tab_name}"}
                        self.error_log(error)
                        return

                    post_headers = {'Accept': '*/*',
                                    'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8',
                                    'Host': 'pretraga2.apr.gov.rs',
                                    'Referer': search_url,
                                    'X-Requested-With': 'XMLHttpRequest'}

                    payload = {'X-Requested-With': 'XMLHttpRequest'}

                    # getting further tabs
                    args = {'headers': post_headers, 'request_type': 4, 'payload': payload}

                    loop_status = self.requests_loop(url, thread_number, f'{tab_name}_{registry}', reg_date_folder, registry_number, tab_name, registry, **args)
                    if loop_status is None:
                        error = {'err_desc': 'To many bad proxies for details links'}
                        return error

                    current_tab_link, html = loop_status
                    print(tab_name, current_tab_link )

            except Exception as e:
                exc_type, exc_obj, exc_tb = sys.exc_info()
                pprint(f"error during tab crawling {str(e)} on line {exc_tb.tb_lineno}")

    def parse_data(self, registry_number=None):

        if registry_number is None:
            parsing_data = self.select('html_download', filters={'parse_stat': None, 'dl_stat': True}, schema=schema)
        else:
            parsing_data = self.select('html_download', filters={'parse_stat': None, 'registry_number': registry_number, 'dl_stat': True}, schema=schema)
        # parsing_data = self.select('html_download', filters={'registry_number': registry_number}, schema=schema)

        if parsing_data:
            df = pd.DataFrame.from_records(parsing_data)
            grouped = df.groupby('registry_number')

            for registry_number, reg_df in grouped:
                parsed_data = ParsedData()
                df_sorted = reg_df.sort_values('date_created')
                df_sorted = df_sorted.loc[df_sorted['registry_number'] == registry_number]
                dates = sorted(df_sorted.proc_date.unique())

                for date in dates:
                    df_sorted1 = df_sorted[(df_sorted.proc_date == date)]
                    pprint(f"processing data for date {date}")
                    if len(df_sorted1) == 1:
                        continue
                    for row in df_sorted1.iterrows():

                        index, row_data = row
                        sha = row_data['sha']
                        # print("row_data", row_data.to_dict())
                        html_path = row_data['dl_path']
                        f_date = date.strftime('%Y%m%d')
                        tab_name = row_data['tab_name']
                        registry = row_data['registry']
                        if html_path is None:
                            continue
                        with open(html_path, 'rb') as f:
                            html = f.read()
                            if registry not in ['InsolventEstate', 'Search']:
                                basic_path = os.path.join(apr_c_html_path, registry_number, f_date, f'BasicInfo_{registry}.html')
                                with open(basic_path, 'rb') as f1:
                                    html1 = f1.read()
                                    basic_info_parser = parser.BasicInfo(registry_number, html1, registry)
                                    basic_info_parser.parse()
                                    parsed_data.basic_info.update(basic_info_parser.data)

                            # PARSING SearchInfo data
                            if tab_name in ['SearchInfo']:
                                html = parser.cirilic_to_latin(html.decode())
                                search_info_parser = parser.SearchInfo(registry_number, html)
                                search_info_status = search_info_parser.parse()
                                if search_info_status == 200:
                                    parsed_data.search_info.update(search_info_parser.data)

                            # PARSING Bankruptcy data
                            if registry in ['InsolventEstate']:
                                bankruptcy_parser = parser.Bankruptcy(registry_number, html, registry)
                                if tab_name in ['BasicInfo']:
                                    bankruptcy_parser.basic_info()
                                else:
                                    bankruptcy_parser.parse(html, tab_name)

                                if bankruptcy_parser.data:
                                    parsed_data.bankruptcy.update(bankruptcy_parser.data)

                            # PARSING Business Names
                            if tab_name in ['FounderData', 'BusinessName', 'PoslovnoIme', 'Names']:
                                cname = basic_info_parser.data.get('cname')
                                if cname is None:
                                    cname = basic_info_parser.data.get('basic_name')
                                busniness_name_parser = parser.BusinessName(registry_number, html, registry, cname=cname)
                                parse_status = busniness_name_parser.parse()
                                if type(parse_status) is dict:
                                    pprint("error during parsing business name data")
                                    self.error_log(parse_status)
                                    if debug:
                                        exit()
                                    else:
                                        continue
                                parsed_data.business_name.update(busniness_name_parser.data)

                            # PARSING Business Data
                            if tab_name in ['PoslovniPodaci', 'BusinessData', 'BasicData']:
                                # business data is splited on below groups in contact_parsers
                                contact_parsers = {'BusinessData': parsed_data.business_data, 'LimitedDuration': parsed_data.limited_duration,
                                                   'ActivityPause': parsed_data.activity_pause, 'EraseDetails': parsed_data.erase_details,
                                                   'AssociationExtra':  parsed_data.association_extra, 'BankAccounts': parsed_data.bank_accounts}

                                for parser_name, data_holder in contact_parsers.items():
                                    parser_class = getattr(parser, parser_name)(registry_number, html, registry)
                                    parse_status = parser_class.parse()
                                    if type(parse_status) is dict:
                                        pprint(f"error during parsing {parser_name}")
                                        self.error_log(parse_status)
                                        if debug:
                                            exit()
                                        else:
                                            continue

                                    if parser_class.data:
                                        data_holder.update(parser_class.data)

                                    # BALANCING MISPLACED BASIC INFO DATA
                                    basic_info_additional = ['activity_start', 'activity_end', 'activity_end_date']
                                    for bia in basic_info_additional:
                                        with_basic_data = ['BusinessData', 'EraseDetails']
                                        for wbd in with_basic_data:
                                            get_activity_data = parser_class.data.get(bia)
                                            if get_activity_data is not None:
                                                if wbd in ['BusinessData']:
                                                    parser_class.data.pop(bia)
                                                    parsed_data.business_data = parser_class.data
                                                if bia == 'activity_end_date':
                                                    bia = 'activity_end'
                                                basic_info_parser.data.update({bia: get_activity_data})
                                parsed_data.basic_info.update(basic_info_parser.data)

                            # PARSING CONTACT DATA
                            if tab_name in ['ContactData', 'KontaktPodaci', 'PoslovniPodaci', 'BusinessData', 'PodaciOAdresama', 'BasicData']:
                                contact_parsers = {'Phone': parsed_data.phone, 'Fax': parsed_data.fax, 'Web': parsed_data.web_page, 'Email': parsed_data.email}
                                for parser_name, data_holder in contact_parsers.items():
                                    parser_class = getattr(parser, parser_name)(registry_number, html, registry)
                                    parse_status = parser_class.parse()
                                    if type(parse_status) is dict:
                                        pprint(f"error during parsing {parser_name} in tab {tab_name}")
                                        self.error_log(parse_status)
                                        if debug:
                                            exit()
                                        else:
                                            continue
                                    if parser_class.data:
                                        data_holder.update(parser_class.data)

                            # PARSING ADDRESS DATA
                            if tab_name in ['PodaciOAdresama', 'Headquarter', 'Address', 'AddressData', 'CentralOfficeData']:
                                address_parser = parser.AddressData(registry_number, html, registry, tab_name)
                                parse_status = address_parser.parse()
                                if type(parse_status) is dict:
                                    pprint(f"error during parsing {tab_name}")
                                    self.error_log(parse_status)
                                    if debug:
                                        exit()
                                    else:
                                        continue
                                parsed_data.address_data = address_parser.data

                            # PARSING ANNOTATIONS
                            if tab_name in ['Annotations', 'Annotation', 'Zabelezbe']:
                                annotation_parser = parser.Annotation(registry_number, html, registry)
                                parse_status = annotation_parser.parse()
                                if type(parse_status) is dict:
                                    pprint(f"error during parsing {tab_name}")
                                    self.error_log(parse_status)
                                    if debug:
                                        exit()
                                    else:
                                        continue
                                parsed_data.annotations = annotation_parser.data

                            # PARSING BRANCH OFFICES
                            if tab_name in ['BranchOffices', 'IzdvojenaMesta', 'BranchOfficeData', 'Branches']:
                                branch_offices_parser = parser.BranchOffices(registry_number, html, registry)
                                parse_status = branch_offices_parser.parse(tab_name)
                                if type(parse_status) is dict:
                                    pprint(f"error during parsing {tab_name}")
                                    self.error_log(parse_status)
                                    if debug:
                                        exit()
                                    else:
                                        continue

                                parsed_data.branch_offices = branch_offices_parser.data
                                for person in branch_offices_parser.persons:
                                    parsed_data.company_people.update({len(parsed_data.company_people): person})

                            # PARSING BASIC CAPITAL DATA
                            if tab_name in ['Capital']:
                                capital_parser = parser.Capital(registry_number, html, registry)
                                parse_status = capital_parser.parse()
                                if type(parse_status) is dict:
                                    pprint(f"error during parsing {tab_name}")
                                    self.error_log(parse_status)
                                    if debug:
                                        exit()
                                    else:
                                        continue
                                parsed_data.capital = capital_parser.data

                            # PARSING CAPITAL MEMBERS AND DATA
                            if tab_name in ['Members']:
                                members_parser = parser.Members(registry_number, html, registry)
                                parse_status = members_parser.parse()
                                if type(parse_status) is dict:
                                    pprint(f"error during parsing {tab_name}")
                                    self.error_log(parse_status)
                                    if debug:
                                        exit()
                                    else:
                                        continue

                                for person in members_parser.people:
                                    parsed_data.company_people.update({len(parsed_data.company_people): person})
                                if members_parser.shares:
                                    parsed_data.capital_shares.update(members_parser.shares)
                                if members_parser.amounts:
                                    parsed_data.capital_amounts.update(members_parser.amounts)
                                if members_parser.groups:
                                    parsed_data.capital_groups.update(members_parser.groups)

                            # PARSING REST OF COMPANY PEOPLE
                            if tab_name in ['LegalTrustees', 'OtherTrustees', 'SupervisoryBoard', 'SteeringCommittee',
                                            'ExecutiveBoard', 'Trustees', 'Managers', 'Procurists', 'Trusstess', 'Headman',
                                            'CompanyFounder', 'Establishers', 'SteeringComittee', 'Poslovodje',
                                            'LicniPodaciPreduzetnika', 'Prokuristi']:

                                company_people_parser = parser.CompanyPeople(registry_number, registry)
                                parse_status= company_people_parser.parse(html, tab_name)
                                if type(parse_status) is dict:
                                    pprint(f"error during parsing {tab_name}")
                                    self.error_log(parse_status)
                                    if debug:
                                        exit()
                                    else:
                                        continue
                                for person in company_people_parser.data:
                                    # print(tab_name, person)
                                    parsed_data.company_people.update({len(parsed_data.company_people): person})

                            # PARSING DECISIONS
                            if tab_name in ['Decrees']:
                                decisions_parser = parser.Decisions(registry_number, html, registry)
                                parse_status = decisions_parser.parse()
                                if type(parse_status) is dict:
                                    pprint(f"error during parsing {tab_name}")
                                    self.error_log(parse_status)
                                    if debug:
                                        exit()
                                    else:
                                        continue

                                for row in decisions_parser.data:
                                    parsed_data.decisions.update({len(parsed_data.decisions): row})

                            self.update('html_download', {'sha': sha}, {'parse_stat': 1, 'parse_date': datetime.now()},  schema=schema)

                    for tablename, db_data in parsed_data.__dict__.items():
                        if db_data:

                            if tablename in ['basic_info', 'bankruptcy']:
                                hashes = parser.HashCreator()
                                # print(f'{tablename}_cols', hashes.__getattribute__(f'{tablename}_cols'))
                                row_sha = parser.row_sha(db_data, hashes.__getattribute__(f'{tablename}_cols'))
                                db_data.update({'sha': row_sha})

                            if 0 not in db_data.keys():
                                db_data = {0: db_data}

                            print(tablename, db_data)

                            filters = {'registry_number': registry_number, 'archive': None}
                            merge_status = self.merge(tablename, db_data, filters, on=['sha'], schema=schema)
                            if merge_status != 200:
                                print(f"merge data for table {tablename} failed")
                                if debug:
                                    exit()
                                else:
                                    continue

                            if tablename == 'decisions':
                                for k, v in db_data.items():
                                    sha = v.get('sha')
                                    doc_link = v.get('doc_link')
                                    filters = {'sha': sha, 'archive': None}
                                    update_status = self.update(tablename,filters, {'doc_link': doc_link}, schema=schema)
                                    if update_status != 200:
                                        print(f"update for table {tablename} failed.")
                                        if debug:
                                            exit()
                                        else:
                                            continue

    def crawl_data(self, data, decisions):
        try:
            thread_number = threading.current_thread().ident
            registry_number = data.get('registry_number')

            # create directories for current registry_number and current date where we going to download htmls
            reg_num_html_path = os.path.join(apr_c_html_path, registry_number)
            reg_date_folder = os.path.join(reg_num_html_path, datetime.now().strftime('%Y%m%d'))
            html_paths = [reg_num_html_path, reg_date_folder]
            for html_path in html_paths:
                self.dir_creation_check(html_path)

            pprint("registry_number {} {}".format(thread_number, registry_number))

            if re.search('^\d+$',registry_number) is None:
                error = {'err_desc': f"Registry_number is not correct {registry_number}"}
                self.error_log(error)
                return

            # get web_link from search_info table if exists and currently have active status (1)
            web_link_data = self.select('search_info', filters={'registry_number': registry_number, 'archive': None, 'link_status': 1}, schema=schema)
            pprint(f"search_info existing data: {web_link_data}")
            if len(web_link_data) > 0:
                details_link = web_link_data[0].get('web_link')
                registry = web_link_data[0].get('registry')
                self.details_crawl(details_link, thread_number, registry_number, registry, reg_date_folder, approach='direct link', decisions=decisions)
                return

            # if there is not active web_link then we will have to login,
            # and fetch that web_link that goes to details

            # check does thread has assign an account
            thread_data = self.inuse.get(thread_number)

            # if it doesn't then acquire one
            if thread_data is None:
                acquire_status = self.acquire_account(thread_number)
                if type(acquire_status) is dict:
                    return acquire_status
            else:
                # check for how many registry number was used continuously
                account_usage = thread_data.get('usage')
                if account_usage is None:
                    thread_data.update({'usage': 1})
                    account_usage = 1

                if account_usage > 5:
                    pprint(f"releasing account for thread_id: {thread_number}")
                    self.release_account(thread_number)
                    acquire_status = self.acquire_account(thread_number)
                    if type(acquire_status) is dict:
                        return acquire_status

            account_data = self.inuse.get(thread_number)
            username = account_data.get('email')
            password = account_data.get('password')

            # +-------------------------------------------------------------+
            # |                         LOGIN - PART                        |
            # +-------------------------------------------------------------+
            sleep(random.randint(2, 5))
            sha = self.new_request(proc_id=thread_number, protocols=['http;https'], new=True)
            try:
                sleep(random.randint(5, 10))
                r = self.requests.get(thread_number)
                ses = requests.session()
                ses.headers.update(user_agent.load())
                proxies = dict()
                proxies['http'] = proxies['https'] = f"{r.ip}:{r.port}"
                # proxies['http'] = proxies['https'] = 'http://drago.kozic%40bisnode.com:B1$n0d3!@85.10.51.85:80'
                # proxies['http'] = proxies['https'] = 'http://drago.kozic%40bisnode.com:B1$n0d3!@141.98.103.235:80'
                # proxies['http'] = proxies['https'] = 'http://drago.kozic%40bisnode.com:B1$n0d3!@141.98.103.147:80'
                proxies['http'] = proxies['https'] = 'http://drago.kozic%40bisnode.com:B1$n0d3!@195.181.166.104:80'
                try:
                    print("main_url", main_url)
                    resp = ses.get(main_url, proxies=proxies, timeout=60)
                except TimeoutError:
                    pprint(f"Timeout on GET request {r.ip}:{r.port}")
                    return
                except requests.exceptions.ProxyError:
                    pprint(f"ProxyError on GET request {r.ip}:{r.port}")
                    self.bad_proxy(sha, r.proxy_data, web_page=base_url)
                    return

                status_code = resp.status_code
                if status_code != 200:
                    pprint(f"Bad status code : {status_code}")
                    self.bad_proxy(sha, r.proxy_data, web_page=base_url)
                    return

                if re.search('HTTP/1.1 400 Bad Request', str(resp.content)) is not None:
                    error = {'err_desc': f"error happened during fetching search info 400 Bad Request"}
                    self.error_log(error)
                    self.bad_proxy(sha, r.proxy_data, web_page=base_url)
                    return

                soup = BeautifulSoup(resp.content, 'lxml')
                prijava_link = soup.find('a', {'href': True}, text=re.compile('Пријави се'))
                if prijava_link is not None:
                    prijava_link = ''.join([base_url, prijava_link['href']])
                    pprint(f"prijava_link: {prijava_link}")
                    try:
                        resp = ses.get(prijava_link, proxies=proxies, timeout=60)
                    except TimeoutError:
                        pprint(f"Timeout on GET request during url: {prijava_link}")
                        return

                    status_code = resp.status_code
                    if status_code != 200:
                        pprint(f"Bad status code on prijava_link: {status_code}")
                        self.bad_proxy(sha, r.proxy_data, web_page=base_url)
                        return

                    soup = BeautifulSoup(resp.content, 'lxml')
                    login_form = soup.find('form', {'action': True, 'id': 'Form1'})
                    if login_form is not None:
                        form_action = login_form['action'].replace("./", '')
                        login_url = ''.join(['https://reid.apr.gov.rs/IDF/', form_action])
                        login_url = parse_url(login_url)
                        pprint(f"login_url: {login_url}")

                        form_data = {'__EVENTTARGET': '', '__EVENTARGUMENT': '', }
                        inputs_ids = ['ctl00_cphMain_ctl03_ImgCaptchaImage_ClientState', '_CurrentGuid_ctl00_cphMain_ctl03_ImgCaptchaImage', '__VIEWSTATE',
                                      '__VIEWSTATEGENERATOR', '__SCROLLPOSITIONX', '__SCROLLPOSITIONY', '__EVENTVALIDATION']
                        for input_id in inputs_ids:
                            input = soup.find('input', {'id': input_id, 'value': True})
                            if input:
                                form_data.update({input_id: input['value']})

                        user_input = {"ctl00$cphMain$ctl00$hdnXmlDataOriginal": "%3cSignedData%3e%3c!%5bCDATA%5b%5d%5d%3e%3c%2fSignedData%3e",
                                      "ctl00$cphMain$ctl00$hdnXmlDataSigned": "",
                                      "ctl00$cphMain$ctl00$hdnSigningDone": "",
                                      "ctl00$cphMain$ctl00$TxtUserEmail": username,
                                      "ctl00$cphMain$ctl00$TxtPassword": password,
                                      "ctl00$cphMain$ctl00$BtnLogin": "Пријави+се",
                                      "ctl00$cphMain$ctl00$ucHrdCtrl$hidRedirectUrl": "http://pretraga2.apr.gov.rs/UnifiedEntitySearch",
                                      "ctl00$cphMain$ctl00$HFCookies": "true",
                                      "ctl00$cphMain$ctl03$TxtName": "",
                                      "ctl00$cphMain$ctl03$TxtSurname": "",
                                      "ctl00$cphMain$ctl03$TxtEmail": "",
                                      "ctl00$cphMain$ctl03$DdlCountry": "-1",
                                      "ctl00$cphMain$ctl03$TxtIdentifier": "",
                                      "ctl00$cphMain$ctl03$TxtPhoneNumber": "",
                                      "ctl00$cphMain$ctl03$TxtPassword": "",
                                      "ctl00$cphMain$ctl03$TxtRepeatPassword": "",
                                      "ctl00$cphMain$ctl03$DdlSecretQuestion": "1",
                                      "ctl00$cphMain$ctl03$TxtSecretQuestion": "Име+фирме+где+сте+први+пут+били+запослени?",
                                      "ctl00$cphMain$ctl03$TxtSecretAnswer": "",
                                      "ctl00$cphMain$ctl03$hidInitReTextBox": "false",
                                      "ctl00$cphMain$ctl03$hidCheckBox": "false"}

                        form_data.update(user_input)
                        try:
                            resp = ses.post(login_url, form_data, proxies=proxies, timeout=60)
                        except TimeoutError:
                            pprint(f"Timeout on POST request during url: {login_url}")
                            return

                        status_code = resp.status_code
                        if status_code != 200:
                            pprint(f"Bad status code on login_url: {status_code}")
                            self.bad_proxy(sha, r.proxy_data, web_page=base_url)
                            return

                        account_error = re.search('Погрешна комбинација', str(resp.content), re.MULTILINE | re.IGNORECASE | re.DOTALL)
                        if account_error:
                            pprint("Pogrešna kombinacija username i login - privremeno blokiran account")
                            with open(f"account_err_{username}.html", 'wb') as f:
                               f.write(resp.content)
                            self.release_account(thread_number)
                            return

                        soup = BeautifulSoup(resp.content, 'lxml')
                        # with open(f"after_login_{registry_number}.html", 'wb') as f:
                        #     f.write(resp.content)

                        payload = dict()
                        inputs_ids = ['wa', 'wresult', 'wctx']
                        for input_id in inputs_ids:
                            input = soup.find('input', {'name': input_id, 'value': True})
                            if input:
                                payload.update({input_id: input['value']})

                        pprint(f"redirect after login payload {payload}")

                        ses.headers.update({'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
                                            'Content-Type': 'application/x-www-form-urlencoded'})
                        sleep(random.randint(2, 5))
                        try:
                            resp = ses.post('http://pretraga2.apr.gov.rs/UnifiedEntitySearch', payload, proxies=proxies)
                        except TimeoutError:
                            print(f"Timeout on POST request during url: 'http://pretraga2.apr.gov.rs/UnifiedEntitySearch'")
                            return

                        status_code = resp.status_code
                        if status_code != 200:
                            pprint(f"Bad status code on redirect url: {status_code}")
                            # self.bad_proxy(sha, r.proxy_data, web_page=base_url)
                            return

                        pprint("Login sucessfull")
                        soup = BeautifulSoup(resp.content, 'lxml')

                        #  From this point we are in and registry_number is ready to be crawled
                        # +-------------------------------------------------------------+
                        # |                         CRAWLING - PART                     |
                        # +-------------------------------------------------------------+

                        token = soup.find('input', {'name': '__RequestVerificationToken'})
                        if token is not None:
                            token = token.get('value')

                            search_url = re.search('url: "(http://pretraga2.apr.gov.rs/UnifiedEntitySearch/UnifiedEntitySearch/GetSearchResult\?SearchId=.*?&rnd=.*?)",', soup.decode(), re.DOTALL | re.MULTILINE)
                            search_url = search_url.group(1)

                            payload = {'__RequestVerificationToken': token,
                                       'rdbtnSelectInputType': 'mbr',
                                       'SearchByRegistryCodeString': registry_number}

                            pprint(f"search company payload: {payload}")
                            try:
                                resp = ses.post(search_url, payload, proxies=proxies)
                            except TimeoutError:
                                pprint(f"Timeout on POST request during url: {search_url}")
                                return

                            status_code = resp.status_code
                            if status_code != 200:
                                pprint(f"Bad status code on registry number search: {status_code}")
                                # self.bad_proxy(sha, r.proxy_data, web_page=base_url)
                                return

                            # processing data for table search_info
                            html = parser.cirilic_to_latin(resp.content.decode())
                            search_info_data = self.search_info(registry_number, html, reg_date_folder, resp.content, resp)
                            if search_info_data is None:
                                error = {'err_desc': 'search info error during processing data'}
                                self.error_log(error)
                                return

                            details_link = search_info_data.details_link
                            self.release_proxy(sha)

                            self.details_crawl(details_link, thread_number, registry_number, search_info_data.registry, reg_date_folder, decisions=decisions)

                        else:
                            server_error = re.search(b'Server Error', resp.content, re.MULTILINE | re.DOTALL | re.IGNORECASE)
                            if server_error is None:
                                pprint("Token is None")
                                with open('token_is_None.html', 'wb') as f:
                                    f.write(resp.content)
                            else:
                                pprint("Webpage Server Error")
                                # sleep(random.randint(60, 120))
                                # with open(f'server_error_{registry_number}.html', 'wb') as f:
                                #     f.write(resp.content)
                else:
                    pprint("Prijava link nema")
                    with open('no_login_link.html', 'wb') as f:
                        f.write(resp.content)

            except Exception as e:
                exc_type, exc_obj, exc_tb = sys.exc_info()
                pprint(f"error during main_url {str(e)} on line {exc_tb.tb_lineno}")

            finally:
                # release proxy back in use if it is not released it will be not be available for that web_base next 30min
                self.release_proxy(sha)

            return ''
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_desc': f'major error {str(e)} on line {exc_tb.tb_lineno}'}
            pprint(f"{error}")
            return

    def transfer(self):
        # transfer company changes from postgres to mssql - APR_SRBIJA
        print('transfer company changes from postgres to mssql - APR_SRBIJA')
        columns = ['sha', 'registry_number', 'company_type', 'change_id', 'change_date', 'change_mode_id', 'regproc_type', 'date_created']
        changes_data = self.select('company_changes', filters={'transfer': None}, columns=columns, schema=schema)
        row_shas = [x.get('sha') for x in changes_data]
        row_sha_split = [row_shas[i:i + 1000] for i in range(0, len(row_shas), 1000)]

        cnt = 0
        for row_sha in row_sha_split:
            cnt += len(row_sha)
            sys.stdout.write(f"\rSyncing data: {cnt}/{len(changes_data)}")
            results = self.apr_conn.select('CompanyChanges', filters={'sha': row_sha})
            shas = [x.get('sha') for x in results]
            row_sha_filter = [x for x in row_sha if x not in shas]
            sha_in = [x for x in row_sha if x  in shas]

            insert_data = [data for data in changes_data if data.get('sha') in row_sha_filter]
            new_insert_data = dict()
            for row in insert_data:
                new_row_sha = row.get('sha')
                if new_row_sha not in new_insert_data.keys():
                    new_insert_data.update({new_row_sha: row})
            insert_data = list(new_insert_data.values())

            if insert_data:
                insert_data = dict(enumerate(insert_data))
                insert_status = self.apr_conn.insert('CompanyChanges', insert_data)
                if insert_status != 200:
                    print("insert failed", insert_data)
                    print("shas", row_sha_filter)
                    print("sha_in", sha_in)
                    if debug:
                        exit()
                    else:
                        continue

                if insert_status == 200:
                    for row_sha in row_sha_filter:
                        update_status = self.update('company_changes', {'sha': row_sha, 'transfer': None}, {'transfer': 1, 'transfer_date': datetime.now()}, schema=schema)
                        if update_status != 200:
                            print("update for new sha failed", row_sha)
                            if debug:
                                exit()
                            else:
                                continue

            existing_sha_filter = [x for x in row_sha if x in shas]
            for row_sha in existing_sha_filter:
                update_status = self.update('company_changes', {'sha': row_sha, 'transfer': None}, {'transfer': 1, 'transfer_date': datetime.now()}, schema=schema)
                if update_status != 200:
                    print("update for existing failed", row_sha)

        # transfer company changes from postgres to mssql - BISNODE_RS
        print('\ntransfer company changes from postgres to mssql - BISNODE_RS')
        print("> assign sha to data in [Bisnode_RS].[APR_WS].[ENTITY_CHANGES]")
        mssql_col_translate = {'registry_number': 'NATIONAL_IDENTIFICATION_NUMBER'.lower(), 'change_date': 'GROUP_DATE_CHANGE'.lower(), 'change_id': 'GROUP_ID'.lower()}
        psgr_sql_translate = {v: k for k,v in mssql_col_translate.items()}
        column_order = ['registry_number', 'change_date', 'change_id']

        sql = "select * from APR_WS.ENTITY_CHANGES where sha is null"
        asign_sha = self.rs_conn.select('ENTITY_CHANGES',sql=sql)
        cnt = 0
        for row in asign_sha:
            entry_id = row.get('entry_id')
            for_hash = {psgr_sql_translate.get(k):v for k,v in row.items() if k in mssql_col_translate.values()}
            change_date = for_hash.get('change_date')
            if change_date is not None:
                change_date = change_date.strftime("%Y-%m-%d %H:%M:%S")
                for_hash.update({'change_date': change_date})
            for_hash = {x: for_hash.get(x) for x in column_order}
            sha = hashlib.sha3_256(str(for_hash).encode()).hexdigest()
            update_status = self.rs_conn.update('ENTITY_CHANGES',{'ENTRY_ID': entry_id},{'sha': sha}, schema='APR_WS', freeze=True)
            cnt += 1
            sys.stdout.write(f'\r> progress: {cnt}/{len(asign_sha)}')
        sys.stdout.write(f'\r> progress: {cnt}/{len(asign_sha)}\n')

        columns = ['sha', 'registry_number', 'company_type', 'change_id', 'change_date', 'change_mode_id', 'regproc_type', 'date_created']
        print("> insert new records into Bisnode_RS.[APR_WS].[ENTITY_CHANGES]")
        changes_data = self.select('company_changes', filters={'transfer': 1, 'change_date': f'>={datetime(2020, 1, 1, 6)}'}, columns=columns, schema=schema)
        cnt = 0
        registry_number_changes = {}
        for row in changes_data:
            registry_number = row.get('registry_number')
            table_row_sha = row.get('sha')
            for_hash = {k: v for k, v in row.items() if k in psgr_sql_translate.values()}
            change_date = for_hash.get('change_date')

            if change_date is not None:
                change_date = change_date.strftime("%Y-%m-%d %H:%M:%S")
                for_hash.update({'change_date': change_date})
            for_hash = {x: for_hash.get(x) for x in column_order}

            sha = hashlib.sha3_256(str(for_hash).encode()).hexdigest()
            # convert column names
            row = {mssql_col_translate.get(k):v for k,v in row.items() if k in psgr_sql_translate.values()}
            row.update({'sha': sha, 'date_created': datetime.now(), 'created_by': 'draganm-python'})
            filters = {'sha': sha, 'NATIONAL_IDENTIFICATION_NUMBER'.lower(): registry_number}
            self.rs_conn.primary_key = 'ENTRY_ID'.lower()
            merge_status = self.rs_conn.merge('ENTITY_CHANGES', {0: row}, filters=filters, on=['sha'], schema='APR_WS', update=False)
            if merge_status == 200:
                update_status = self.update('company_changes', {'sha': table_row_sha, 'transfer': 1}, {'transfer': 2, 'transfer_date': datetime.now()}, schema=schema)
            cnt += 1
            sys.stdout.write(f'\r> progress: {cnt}/{len(changes_data)}')
        sys.stdout.write(f'\r> progress: {cnt}/{len(changes_data)}\n')

        print("insert or update records into Bisnode_RS.[APR_WS].[BUSINESS_ENTITIES]")
        changes_data = self.select('company_changes', filters={'transfer': 2, 'change_date': f'>={datetime(2020, 1, 1, 6)}'}, columns=columns, schema=schema)
        for row in changes_data:
            registry_number = row.get('registry_number')
            change_date = row.get('change_date')
            registry_number_change = registry_number_changes.get(registry_number)
            if registry_number_change is None:
                registry_number_changes.update({registry_number: change_date})
            else:
                if registry_number_change < change_date:
                    registry_number_changes.update({registry_number: change_date})

        cnt, update, insert, incorrect, checked = 0, 0, 0, 0, 0
        for registry_number, change_date in registry_number_changes.items():
            cnt += 1
            self.rs_conn.primary_key = 'id'
            reg_num_db_data = self.rs_conn.select('BUSINESS_ENTITIES', filters={'MB': registry_number}, schema='APR_WS')
            check_status = 0
            if reg_num_db_data:
                reg_num_db_data = reg_num_db_data[0]
                last_date = datetime.now()
                date_created = reg_num_db_data.get('date_created')
                if date_created is not None:
                    last_date = pd.to_datetime(date_created, yearfirst=True)

                date_modified = reg_num_db_data.get('date_modified')
                if date_modified is not None:
                    date_modified = pd.to_datetime(date_modified, yearfirst=True)
                    last_date = max(last_date, date_modified)

                if last_date < change_date:
                    check_status += 1
                    updates = {'IS_APR_CHECKED': 0, 'STATUS_CODE': None, 'DATE_MODIFIED': datetime.now(), 'MODIFIED_BY': 'draganm-python', 'IS_PU_CHECKED': 0}
                    update_status = self.rs_conn.update('BUSINESS_ENTITIES', {'MB': registry_number}, updates, schema='APR_WS', freeze=True)
                    if update_status == 200:
                        update += 1
                        update_status = self.update('company_changes', {'registry_number': registry_number, 'transfer': 2}, {'transfer': 3, 'transfer_date': datetime.now()}, schema=schema)
                else:
                    checked += 1
            else:
                if not registry_number.__contains__('X'):
                    check_status += 1
                    insert_datas = {'MB': registry_number, 'IS_APR_CHECKED': 0, 'STATUS_CODE': None, 'DATE_CREATED': datetime.now(), 'CREATED_BY': 'draganm-python', 'IS_PU_CHECKED': 0}
                    insert_status = self.rs_conn.insert('BUSINESS_ENTITIES', {0: insert_datas}, schema='APR_WS')
                    if insert_status == 200:
                        insert += 1
                        update_status = self.update('company_changes', {'registry_number': registry_number, 'transfer': 2}, {'transfer': 3, 'transfer_date': datetime.now()}, schema=schema)
                else:
                    incorrect += 1

            if check_status == 0:
                update_status = self.update('company_changes', {'registry_number': registry_number, 'transfer': 2}, {'transfer': 3, 'transfer_date': datetime.now()}, schema=schema)
            sys.stdout.write(f'\r> progress {cnt}:{len(registry_number_changes)}, updates: {update}, insert: {insert}, incorrect: {incorrect}, checked: {checked}')

    def decisions_download(self, registry_number=None):
        for i in range(5):
            tab_name, registry = 'Decrees', 'Any'
            filters = {'doc_dl': None, 'doc_link': True}
            if registry_number is not None:
                filters.update({'registry_number': registry_number})
                rows = self.select('decisions', filters=filters, schema=schema)
            else:
                rows = self.select('decisions', filters=filters, schema=schema)
            expired_links = False
            loop_status = False

            for row in rows:
                print(row)
                sha = row.get('sha')
                doc_link = row.get('doc_link')
                registry_number = row.get('registry_number')

                loop_status = self.requests_loop(doc_link, 1, 'doc', apr_decisions, registry_number, tab_name, registry)
                if loop_status is None:
                    error = {'err_desc': 'To many bad proxies for document link'}
                    loop_status = True
                    break

                doc_dl_link, doc_dl_data = loop_status
                if doc_dl_data == 404:
                    # if data is 404 doc link has expired
                    filters = {'sha': sha, 'registry_number': registry_number, 'archive': None}
                    values = {'doc_link': None, 'doc_dl': None}
                    update_status = self.update('decisions', filters, values,  schema=schema)
                    expired_links = True
                else:
                    sha = row.get('sha')
                    filters = {'registry_number': registry_number, 'sha': sha, 'archive': None}
                    update_status = self.update('decisions',filters, doc_dl_data,  schema=schema)
                    print("final data",update_status, row, doc_dl_data)
                    print("===="*15)

            if loop_status is True:
                continue

            if expired_links is True:
                self.start_crawl(decisions=True)
                self.parse_data()
                continue
            break

    def pu_nbs_list_updates(self):
        """"""


def webservice_history(year=2016, month=1, day=1):
    # checks weeks if starting date is not Monday it will skip
    # all dates until it comes to date that starts with Monday
    # so set date of first Sunday from where you want to check data
    base_date = datetime(year, month, day, 6, 0, 0)
    total_days = (datetime.now() - base_date).days + 1
    api = AprCompanies()
    try:
        for i in range(1, total_days, 1):
            starting_date = base_date + timedelta(days=i)
            start_day_name = starting_date.strftime('%A')
            if start_day_name == 'Monday':
                end_date = starting_date + timedelta(days=6)
                end_day_name = end_date.strftime('%A')
                print(start_day_name, starting_date, end_day_name, end_date)
                api.get_company_changes(starting_date, end_date)
        api.transfer()
    finally:
        api.ws_session.close()


if __name__ == '__main__':
    # History date must always must be Sunday (local: Nedelja)
    # webservice_history(2016, 1, 3)

    # main program endless loop
    while 1:
        print("\n"+"======="*5)
        api = AprCompanies()

        # fetching and processing data from web service
        if api.client != 404:
            api.update_apr_codebooks()
            api.get_company_changes((datetime.now() - timedelta(days=4)), datetime.now())
        api.ws_session.close()

        # transferring data from psgr to mssql
        api.transfer()

        api.runtime_log.update({'status': api.runtime_status, 'errors': api.runtime_errors})
        api.info_log(api.runtime_log)

        gc.collect()
        del api
        sleep(1800)

    # api.pu_nbs_list_updates()

    # crawl gathered data
    # api = AprCompanies()
    # api.dir_creation_check()

    # html loader (if database table html_download is truncated)
    # api.html_loader()

    # data crawling
    # regnum = '06020089'  # passport je u pin 17671677
    # regnum = None
    # api.start_crawl(regnum)

    # data processing
    # api.parse_data(regnum)

    # document downloads
    # api.decisions_download(regnum)

