import re
import os
import bs4
import sys
import hashlib
import cyrtranslit
import pandas as pd
import numpy as np
import dateutil.parser
from decimal import Decimal
from datetime import datetime
from bs4 import BeautifulSoup
from crawlers_sm.serbia.apr.companies.sync import HashCreator

hash_cols = HashCreator()

# rename tab to same group name
tab_replacements = {'PodaciOAdresama': 'AddressData', "PoslovnoIme": 'BusinessName', "PoslovniPodaci": "BusinessData",
                    "BasicData": "BusinessData", "Address": "AddressData", 'Headquarter': "AddressData", "Prokuristi": 'Procurists',
                    'Poslovodje': "Headman", "LicniPodaciPreduzetnika": "CompanyFounder",'IzdvojenaMesta': "BranchOffices",
                    "Members": "CapitalMembers"}

# all known registries if any new comes it should be added here
registries = ['Enterprise', 'Enterpreneur', 'SportAssociation', 'AssociationPublic', 'Foundation',
              'InsolventEstate', 'ForeignAssociation', 'ForeignFAEPublic']


# status codes of parsing
codes = {900: "Nijedan rezultat ne odgovara izabranom kriterijumu pretrage",
         901: "InValid Captcha",
         902: "An error occurred while processing your request",
         903: "Greska servera",
         904: "Link or links could not be parsed",
         905: "Registry not mapped",
         907: "Van nadleznosti registra",
         908: "Search data is empty",
         999: "Major method error"}


def cirilic_to_latin(text):
    if text is None:
        return None

    rep = {'А': 'A', 'Б': 'B', 'В': 'V', 'Г': 'G', 'Д': 'D', 'Ђ': 'Đ', 'Е': 'E', 'Ж': 'Ž', 'З': 'Z',
           'И': 'I', 'Ј': 'J', 'К': 'K', 'Л': 'L', 'Љ': 'Lj', 'М': 'M', 'Н': 'N', 'Њ': 'Nj', 'О': 'O',
           'П': 'P', 'Р': 'R', 'С': 'S', 'Т': 'T', 'Ћ': 'Ć', 'У': 'U', 'Ф': 'F', 'Х': 'H', 'Ц': 'C',
           'Ч': 'Č', 'Џ': 'Dž', 'Ш': 'Š', 'а': 'a', 'б': 'b', 'в': 'v', 'г': 'g', 'д': 'd', 'ђ': 'đ',
           'е': 'e', 'ж': 'ž', 'з': 'z', 'и': 'i', 'ј': 'j', 'к': 'k', 'л': 'l', 'љ': 'lj', 'м': 'm',
           'н': 'n', 'њ': 'nj', 'о': 'o', 'п': 'p', 'р': 'r', 'с': 's', 'т': 't', 'ћ': 'ć', 'у': 'u',
           'ф': 'f', 'х': 'h', 'ц': 'c', 'ч': 'č', 'џ': 'dž', 'ш': 'š','à':'a','&amp;':'&',
           'Ω':'Omega',"'":'','º':'0','Å':'A','å':'a','A':'A','È':'E','Ì':'I','Ò':'O','Ù':'U',
           'è':'e','ì':'i','Á':'A','É':'E','Í':'I','Ó':'O','Ú':'U','Ý':'Y','Â':'A','Ê':'E','Î':'I','Ô':'O',
           'Û':'U','á':'a','é':'e','í':'i','ó':'o','ú':'u','ý':'y','â':'a','ê':'e','î':'i','ô':'o','û':'u',
           'Ã':'A','Ñ':'N','Õ':'O','ã':'a','ñ':'n','õ':'o','Ä':'A','Ë':'E','Ï':'I','Ü':'U','Ÿ':'Y',
           'ä':'a','ë':'e','ï':'i','ö':'o','ü':'u','ÿ':'y','Ç':'C','ç':'c','Þ':'P','þ':'p','Ø':'O','ø':'o',
            'ª': '','ß':'ss','ẞ':'SS','Œ':'OE','œ':'oe','Æ':'Ć','æ':'ć','ð':'đ','і':'i','S':'S','&quot;':'"',
           'ӧ':'o',"ʽ":"",'ќ':'k','ѕ':'s','ĭ':'l','й':'j','я':'a','ы':'i','ΡΜ':'PM','″':'','Ð':'Đ','Й':'J','đ':'đ',
           '&#231;':'c','ğ':'g','&#193;':'A','&#201;':'E','&#220;':'U','&#211;':'O','Ő':'O','Ũ':'U','β':'SS','׳':'','ž':'ž','a':'a','č':'č','&#208;':'Đ',
           'ÄŤ':'č','İ':'I','Ĺľ':'ž','ș':'s','΄':'', '&#214 ': 'Ö', '&#39;': "'"
           }


    rep = dict((re.escape(k), v) for k, v in rep.items())
    pattern = re.compile("|".join(rep.keys()))
    text = pattern.sub(lambda m: rep[re.escape(m.group(0))], text)
    return text


def type_change(text, col_name):
    int_cols = ['change_id', 'Identifikatornacinapromene',
                'Identifikatortiparegistracionogpostupka', 'TipPrivrednogSubjekta',
                'IdentifikatorTipaPoslovnogSubjekta']

    if col_name in int_cols:
        try:
            return int(text)
        except:
            return -1
    return text


def pin_fix(pin):
    if type(pin) is str:
        pin = pin.zfill(13)
    return pin


def codebooks_type1(data, codebookname):
    replace_cols = {'IdentifikatorTipaPoslovnogSubjekta': 'company_type_id', 'TipPoslovnogSubjekta': 'description_c',
                    'IdentifikatorNacinaPromene': 'company_mode_id', 'NazivNacinaPromene': 'description_c',
                    'IdentifikatorTipaRegistracionogPostupka': 'regproc_id', 'NazivTipaRegistracionogPostupka': 'description_c' }

    need_latin_translate = {'SviPodaciOTipovimaPoslovnogSubjekta': ['TipPoslovnogSubjekta'],
                            'SviPodaciOTipovimaNacinaPromene': ['NazivNacinaPromene'],
                            'SviPodaciOTipovimaRegistracionogPostupka': ['NazivTipaRegistracionogPostupka']}

    final_data = {}
    if data is not None:
        for row in data:
            row_data = row.podatak
            row_finala_data = {}
            for column in row_data:
                col_naziv = column.naziv
                col_val = type_change(column.vrednost, col_naziv)
                col_tip = column.tip

                db_name = replace_cols.get(col_naziv)
                if db_name is not None:
                    row_finala_data.update({db_name: col_val})

                if codebookname in list(need_latin_translate.keys()):
                    trans_cols = need_latin_translate.get(codebookname)
                    if col_naziv in trans_cols:
                        if col_val is not None:
                            col_val = cyrtranslit.to_latin(col_val)
                        if db_name is not None:
                            db_name = db_name.replace("_c", '_l')
                            row_finala_data.update({db_name: col_val})
            final_data.update({len(final_data): row_finala_data})
    return final_data


def company_changes(data):
    data_ready = []
    sha_row = []

    replace_cols = {'DatumPromene': 'change_date', 'Identifikatornacinapromene': 'change_mode_id',
                    'Identifikatortiparegistracionogpostupka': 'regproc_type',
                    'TipPrivrednogSubjekta': 'company_type'}

    if data is not None:
        for row in data:
            mbr = row.maticniBroj
            grupa = row.grupa
            tip = row.tip
            for record in grupa:
                change_id = type_change(record.id, 'change_id')
                podatak = record.podatak
                record_data = {'registry_number': mbr, 'change_id': change_id}
                for columns in podatak:
                    col_naziv = columns.naziv
                    col_val = type_change(columns.vrednost, col_naziv)
                    col_tip = columns.tip
                    if col_tip in ['DateTime', 'Date']:
                        col_val = dateutil.parser.parse(col_val)
                    db_name = replace_cols.get(col_naziv)
                    if db_name is not None:
                        record_data.update({db_name: col_val})

                if len(record_data) > 2:
                    sha = hashlib.sha3_256(str(record_data).encode()).hexdigest()
                    record_data.update({'sha': sha})
                    if sha not in sha_row:
                        sha_row.append(sha)
                        data_ready.append(record_data)
    return data_ready

# ---------------------------------------------------------
#                   PAGE PARSER FUNCTIONS
# ---------------------------------------------------------


def value_replace(value):
    try:
        if type(value) is str:
            replacements = {"'": "", '&amp;': '&', ";": ' '}
            for k, v in replacements.items():
                value = value.replace(k, v)
            value = re.sub(r'\s{2,}', " ", value)
            value = value.strip(":").strip()
            if value == '' or value == '-':
                value = None
        if str(value).lower() in ('nan', 'nat', 'none'):
            value = None
        return value
    except Exception as e:
        return value


def date_convert(date, search_format=r'(\d+\.\d+\.\d+).*', striptime='%d.%m.%Y', strftime='%Y-%m-%d', string=True):
    if type(date) is str:
        search_date = re.search(search_format, date, re.DOTALL | re.MULTILINE | re.IGNORECASE)
        if search_date:
            parse_date = search_date.group(1)
            if string is False:
                parse_date = datetime.strptime(parse_date, striptime)
            else:
                parse_date = datetime.strptime(parse_date, striptime).strftime(strftime)
            return parse_date
    return None


def convert_amount(amount,as_string=False):
    if amount is not None:
        try:
            amount = str(amount).replace(".", "").replace(",", ".")
            if type(amount) is str and re.search("[A-Z]", amount) is None:
                try:

                    if as_string is True:
                        amount = str(round(Decimal(amount), 4))
                    else:
                        amount = float(round(Decimal(amount), 4))

                except Exception as e:
                    pass
            return amount
        except Exception as e:
            return amount
    return amount


def dataframe_convert_data(df, db_col):
    try:
        ready_data = dict()
        for rown, data in df.iterrows():
            header, value = data[0], data[1]
            header = header.replace(":", "")
            if header is not None and value is not None:
                if header in DataTypes.dates:
                    value = date_convert(value)
                if header in DataTypes.floats and value is not None:
                        value = round(float(value), 2)
            ready_data.update({db_col.get(header): value_replace(value)})
        return ready_data
    except Exception as e:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
        raise Exception(error)


def unique_status(status_eval):
    if status_eval is None or status_eval == '':
        return 'Nepoznato'
    rex = {'aktiv': 'Aktivan', 'bris': 'Brisan', 'prin.*?likvi': 'Prinudna likvidacija',
           'proc.*?likvi': 'U procesu likvidacije', 'likvi': 'U likvidaciji',
           'ste.*?post': 'U stečajnom postupku', 'ste.a': 'U stečaju'}
    for k, v in rex.items():
        if re.search(k, status_eval, re.IGNORECASE) is not None:
            status_eval = v
            break
    return status_eval


def row_sha(data, cols=None, dates=None):
    """
    :type data: dict
    :type cols: list
    :type dates: list
    """
    copy_data = data.copy()
    if cols is not None:
        copy_data = {k: str(v) for k,v in copy_data.items() if k in cols}

    # replacing nan and Nat and None
    for k, v in copy_data.items():
        if str(v).lower() in ('nan', 'nat', 'none'):
            copy_data.update({k: str(None)})

    # order columns
    if cols is not None:
        copy_data = {k:copy_data.get(k) for k in cols}

    values = ''.join(list(copy_data.values()))
    # values = re.sub(r'[\s+]','',values, re.DOTALL|re.UNICODE)
    sha = hashlib.sha3_256(values.encode()).hexdigest()
    return sha



def tables(html, registry_number, cols, form=1,escape=True):
    data = dict()
    if type(html) is bytes:
        html = html.decode()
    html = cirilic_to_latin(html)
    soup = BeautifulSoup(html, "lxml")
    tables_ = soup.find_all("table")
    basic_names = {c: None for c in cols}
    basic_names.update({'registry_number': registry_number})

    for table in tables_:
        table_caption = table.caption
        trs = table.find_all('tr')
        for tr in trs:
            th = tr.find_all('th')
            td = tr.find_all('td')
            if len(th) == 1 and len(td) == 1:
                header = value_replace(th[0].text)
                if escape is True:
                    header_value = value_replace(td[0].text)
                else:
                    header_value = td[0].text
                if form == 1:
                    data.update({header: header_value})
                if form == 2:
                    if table_caption is not None:
                        table_caption_text = table_caption.text.strip()
                        caption_rows = data.get(table_caption_text)
                        if escape is True:
                            header_value = value_replace(header_value)
                        if caption_rows is None:
                            data.update({table_caption_text: {header: header_value}})
                            caption_rows = data.get(table_caption_text)
                        if caption_rows is not None:
                            caption_rows.update({header: header_value})
    return data

class DataTypes:
    dates = ['Datum osnivanja', 'Datum prestanka obavljanja delatnosti', 'Datum brisanja', 'Datum početka obavljanja delatnosti', 'Datum brisanja iz registra']
    floats = ['Vrednost stečajne mase']



class SearchInfo:
    def __init__(self, registry_number, html):
        self.registry_number = registry_number
        self.html = html
        self.parsing = []
        self.blackout = 0
        self.registry = None
        self.data = dict()
        self.details_link = None

        # column names and order for hash calculation
        self.search_info = ['registry_number', 'ctype', 'cname', 'cstatus', 'registry']

    def convert_to_mssql(self, data):
        postgre_to_sql = {'ctype': 'Type', 'registry': 'Registry', 'registry_number': 'RegistryNumber', 'cname': 'Name', 'cstatus': 'Status', 'sha': 'sha'}
        return {postgre_to_sql.get(k): v for k,v in data.items() if k in list(postgre_to_sql.keys())}

    def parse(self):
        try:
            # check can we get details link from search results
            regex = "<tbody>.*?<tr>.*?<td>.*?</td>.*?<td>\s*\d+\s*</td>\s+<td>.*?</td>\s+<td>.*?</td>\s+<td>\s+-\s+</td>\s+</tr>\s+</tbody>"
            if re.search(regex, self.html, re.DOTALL | re.MULTILINE | re.IGNORECASE) is not None:
                self.blackout = 1

            # pre-check most common website problems that were occured by now
            prechecks = {"Nijedan rezultat ne odgovara izabranom kriterijumu pretrage": 900,
                         "InValid Captcha": 901,
                         "An error occurred while processing your request": 902,
                         "gre.ka": 903}

            precheck_status = 200
            for k, v in prechecks.items():
                if re.search(k, self.html, re.DOTALL | re.MULTILINE | re.IGNORECASE) is not None:
                    self.parsing.append(v)
                    precheck_status = 400

            if precheck_status == 400:
                return 400

            # if details link exist, check registry type that can be found inside link
            if self.blackout == 0:
                # find link on details and extract it
                soup = BeautifulSoup(self.html, "lxml")
                link = soup.find('a', href=True, text='Detaljnije')['href']

                # if link was able to be parsed prepare it for further use
                if link is not None:
                    self.details_link = link.replace("&amp;", "&")

                # if link was unable to be parsed
                if link is None:
                    self.parsing.append(904)

                # loop trough known registries we have
                for registry in registries:
                    if re.search(registry, self.html, re.DOTALL | re.MULTILINE | re.IGNORECASE):
                        self.registry = registry
                        break

                # if registry name is not in existing list of registries than we possible don't have parser for it
                if self.registry is None:
                    self.parsing.append(905)
                    return 400

            # major block of parsing results
            html = self.html.replace("<tbody>", "").replace('</tbody>', '')
            soup0 = BeautifulSoup(html, 'lxml')
            html = soup0.find('table', {'class': "table table-condensed table-hover table-striped"}).decode()
            df = pd.read_html(html, header=0)[0]
            db_col = {'Tip subjekta': "ctype", 'Matični broj': 'registry_number', 'Poslovno ime': "cname", 'Status': 'cstatus'}

            columns = [x for x in df.columns if x in db_col.keys()]

            for rown, data in df.iterrows():
                for column in columns:
                    value = data[column]
                    try:
                        value = value.replace("&amp;", "&").replace(";", " ## ")
                    except:
                        pass
                    if column == 'Matični broj':
                        value = str(value).zfill(8)
                    self.data.update({db_col.get(column): value_replace(str(value)), "registry": self.registry})

            if self.data:
                cstatus = self.data.get('cstatus')
                if cstatus is not None:
                    status_group = unique_status(cstatus)
                    self.data.update({'status_group': status_group})

                self.parsing.append(200)
                # data ready for pstgresql
                record_sha = row_sha(self.data, self.search_info)
                self.data.update({'sha': record_sha})
                print("search info:", self.data)

                # data ready for mssql
                # print(self.convert_to_mssql(self.data))
                return 200
            else:
                self.parsing.append(908)
                return 400

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
            return error


class BasicInfo:
    def __init__(self, registry_number, html, registry):
        self.registry_number = registry_number
        self.html = html
        self.registry = registry
        self.data = {x: None for x in hash_cols.basic_info_cols}

    def convert_to_mssql(self, data):
        postgre_to_sql = {'basic_name': 'BasicName', 'cname': 'CompanyName',
                          'cstatus': 'Status', 'legal_form': 'LegalForm',
                          'registry_number': 'RegistryNumber', 'activity_start': 'ActivityStartDate',
                          'activity_end': 'ActivityEndDate', 'foun_date': 'FounDate',
                          'erase_date': 'EraseDate', 'sha': 'sha'}
        return {postgre_to_sql.get(k): v for k, v in data.items() if k in list(postgre_to_sql.keys())}

    def parse(self):
        try:
            if type(self.html) is bytes:
                self.html = self.html.decode()
            html = cirilic_to_latin(self.html)
            df = pd.read_html(html, header=None)[0]
            df = df.where((pd.notnull(df)), None)

            db_col = {'Naziv': "basic_name", 'Poslovno ime': 'cname',
                      'Naziv predstavništva': 'basic_name',
                      'Status': "cstatus", 'Pravna forma': 'legal_form',
                      'Matični broj': 'registry_number',
                      'Datum početka obavljanja delatnosti': "activity_start",
                      'Datum prestanka obavljanja delatnosti': "activity_end",
                      'Datum osnivanja': 'foun_date',
                      'Datum brisanja': "erase_date",
                      'Oblik organizovanja': 'legal_form',
                      'Naziv udruženja/saveza': 'cname',
                      'Datum brisanja iz registra': 'erase_date'}

            self.data.update(dataframe_convert_data(df, db_col))

            soup = BeautifulSoup(self.html, "lxml")
            links = [{re.search("Details/(.*?)\?", x['data-ajax-url']).group(1): x['data-ajax-url']} for x in soup.find_all('a', {'data-ajax-url': True}, href=True)]
            if not links:
                return {'err_type': "Parsing error", 'err_desc': codes.get(904), 'status': 2}
            return links

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            return error


class BusinessName:
    def __init__(self, registry_number, html, registry, cname=None):
        self.registry_number = registry_number
        self.html = html
        self.registry = registry
        self.data = dict()
        self.cname = cname

    def convert_to_mssql(self, data):
        postgre_to_sql = {'registry_number': 'RegistryNumber', 'cname': 'CompanyName',
                          'short_name': 'ShortName', 'other_name': 'CompanyNameOther', 'short_other_name': 'ShortNameOther',
                          'sha': 'sha', 'language': 'language'}
        return {postgre_to_sql.get(k): v for k, v in data.items() if k in list(postgre_to_sql.keys())}

    def col_replace(self, value, caption='basic'):
        value = value_replace(value)
        if value is not None:
            repl_values = {'Скраћено пословно име': 'short_name', 'Пословно име': 'cname', 'Jeзик': 'language', 'Скраћени назив': 'short_name'}
            for k, v in repl_values.items():
                if re.search(k, value, re.DOTALL | re.MULTILINE | re.IGNORECASE) is not None:
                    return v
            if value.strip() == 'Превод':
                if caption == 'Преводи пословног имена':
                    return 'other_name'
                if caption == 'Преводи скраћеног пословног имена':
                    return 'short_other_name'
        return value

    def parse(self):
        html = self.html.decode()
        soup = BeautifulSoup(html, "lxml")
        tables = soup.find_all("table")
        filter_cols = ['registry_number', 'cname', 'short_name', 'other_name', 'short_other_name', 'language']
        hashs = HashCreator()
        basic_names = {c: None for c in hashs.business_name_cols}
        basic_names.update({'registry_number': self.registry_number})
        final_data = dict()
        try:
            for table in tables:
                df = pd.read_html(table.decode())
                if type(df) is list:
                    df = df[0]
                    df.columns = [value_replace(c) for c in df.columns]
                    if df.empty is False:
                        if 'Jeзик' in df.columns and 'Тип назива' not in df.columns:
                            caption = table.caption.text
                            df.columns = [self.col_replace(x) for x in df.columns]
                            df.columns = [self.col_replace(x, caption=caption) for x in df.columns]
                            data = df.to_dict(orient='records')
                            for row in data:
                                language = row.get('language')
                                row = {k: value_replace(v) for k, v in row.items()}
                                if language is not None:
                                    language = cirilic_to_latin(language)
                                    row.update({'language': language})

                                basic_names_c = basic_names.copy()
                                basic_names_c.update(row)
                                language_data = final_data.get(language)
                                if language_data is None:
                                    final_data.update({language: basic_names_c})
                                else:
                                    basic_names_c = {k: v for k, v in basic_names_c.items() if v is not None}
                                    language_data.update(basic_names_c)
                        elif 'Тип назива' in df.columns:
                            col_rep = {'Пун назив': 'other_name', 'Скраћени назив': 'short_other_name', 'Назив': 'other_name'}
                            for rown, data in df.iterrows():
                                language = cirilic_to_latin(data['Језик'])
                                name_version = col_rep.get(data['Тип назива'])
                                original_other_name = data['Назив']
                                row = {'language': language, name_version: original_other_name}

                                basic_names_c = basic_names.copy()
                                basic_names_c.update(row)
                                language_data = final_data.get(language)
                                if language_data is None:
                                    final_data.update({language: basic_names_c})
                                else:
                                    basic_names_c = {k: v for k, v in basic_names_c.items() if v is not None}
                                    language_data.update(basic_names_c)
                        else:
                            df[0] = [self.col_replace(row) for row in df[0]]
                            data = dict(df.to_dict(orient='split').get('data'))
                            data = {k:v for k,v in data.items() if k in filter_cols}
                            data = {k: cirilic_to_latin(value_replace(v)) for k, v in data.items()}
                            basic_names.update(data)
            if not final_data:
                final_data = {0: basic_names}

            final_data = dict(enumerate(list(final_data.values())))
            hash_values = dict()
            no_data_rows = list()
            for k, v in final_data.items():
                company_name = v.get('cname')
                if company_name is None:
                    v.update({'cname': self.cname})
                sha = row_sha(v, hashs.business_name_cols)
                hash_values.update({k: {'sha': sha}})
                cols_with_data = ['cname', 'short_name', 'other_name', 'short_other_name']
                no_data_test = v.copy()
                no_data_test = {k:v for k,v in no_data_test.items() if v is not None and k in cols_with_data}
                if len(no_data_test) == 0:
                    no_data_rows.append(k)

            [final_data.get(k).update(v) for k, v in hash_values.items()]
            for pop_key in no_data_rows:
                if final_data.get(pop_key) is not None:
                    final_data.pop(pop_key)
            self.data = final_data
            return 200

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            return error


class BusinessData:
    def __init__(self, registry_number, html, registry):
        self.registry_number = registry_number
        self.html = html
        self.registry = registry
        self.data = dict()

    def convert_to_mssql(self, data):
        postgre_to_sql = {'registry_number': 'RegistryNumber', 'vat_number': 'VatNumber', 'rzzo': 'RZZO', 'pio': 'PIO',
                          'activity_code': 'ActivityCode', 'activity_name': 'ActivityName'}
        return {postgre_to_sql.get(k): v for k, v in data.items() if k in list(postgre_to_sql.keys())}

    def parse(self):
        try:
            data = dict()
            hashs = HashCreator()
            busniness_data_cols = hashs.busniness_data_cols
            table_data = tables(self.html, self.registry_number, busniness_data_cols)
            basic_names = {c: None for c in busniness_data_cols}
            basic_names.update({'registry_number': self.registry_number})

            for header, header_value in table_data.items():
                if header_value is not None:
                    # case when activity data is already split
                    if header.__contains__('Šifra delatnosti'):
                        data.update({'activity_code': value_replace(header_value)})
                    if header.__contains__('Opis delatnosti'):
                        data.update({'activity_name': value_replace(header_value)})

                    # case when activity data needs to be separated
                    if header.__contains__('Šifra i naziv delatnosti'):
                        activity_code, activity_name = re.search('(\d+)\s*-(.*)', header_value).groups()
                        data.update({'activity_code': value_replace(activity_code),
                                           'activity_name': value_replace(activity_name)})

                    if header.__contains__('Šifra i opis sportske delatnosti'):
                        activity_code, activity_name = re.search('(\d+)\s*(.*)', header_value).groups()
                        data.update({'activity_code': value_replace(activity_code),
                                           'activity_name': value_replace(activity_name)})

                    if header.__contains__('Datum početka obavljanja delatnosti'):
                        activity_start = date_convert(header_value)
                        if activity_start is not None:
                            data.update({'activity_start': activity_start})

                    simple_headers = {'pib': 'vat_number', 'rzzo': 'rzzo', 'pio': 'pio'}
                    for k, v in simple_headers.items():
                        if header.lower().__contains__(k):
                            data.update({v: value_replace(header_value)})
            if data:
                basic_names.update(data)
                sha = row_sha(basic_names, busniness_data_cols)
                basic_names.update({'sha': sha})
                self.data = basic_names

            return 200
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            return error


class LimitedDuration():
    def __init__(self, registry_number, html, registry):
        self.registry_number = registry_number
        self.html = html
        self.registry = registry
        self.data = dict()

    def convert_to_mssql(self, data):
        postgre_to_sql = {'registry_number': 'RegistryNumber', 'end_date': 'EndDate'}
        return {postgre_to_sql.get(k): v for k, v in data.items() if k in list(postgre_to_sql.keys())}

    def parse(self):
        try:
            hashs = HashCreator()
            limited_duration_cols = hashs.limited_duration_cols
            table_data = tables(self.html, self.registry_number, limited_duration_cols)
            basic_names = {c: None for c in limited_duration_cols}
            basic_names.update({'registry_number': self.registry_number})
            for header, header_value in table_data.items():
                if header_value is not None:
                    header_names = ['Vreme trajanja','Trajanje ograničeno do', cirilic_to_latin('Време на које је представништво основано')]
                    for header_name in header_names:
                        if header.__contains__(header_name):
                            limited_duration = None
                            eu_date = re.search('(\d+.\d+.\d+)', value_replace(header_value), re.MULTILINE | re.IGNORECASE | re.DOTALL)
                            us_date = re.search('(\d+-\d+-\d+)', value_replace(header_value), re.MULTILINE | re.IGNORECASE | re.DOTALL)
                            if eu_date is not None:
                                limited_duration = pd.to_datetime(eu_date.group(1), dayfirst=True).strftime('%Y-%m-%d')
                            if us_date is not None:
                                limited_duration = pd.to_datetime(us_date.group(1), yearfirst=True).strftime('%Y-%m-%d')
                            if limited_duration is not None:
                                basic_names.update({'end_date': limited_duration})
                                sha = row_sha(basic_names, hashs.limited_duration_cols)
                                basic_names.update({'sha': sha})
                                self.data = basic_names
            return 200
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            return error


class ActivityPause:
    def __init__(self, registry_number, html, registry):
        self.registry_number = registry_number
        self.html = html
        self.registry = registry
        self.data = dict()

    def parse(self):
        try:
            hashs = HashCreator()
            activity_pause_cols = hashs.activity_pause_cols
            table_data = tables(self.html, self.registry_number, activity_pause_cols)
            basic_names = {c: None for c in activity_pause_cols}
            basic_names.update({'registry_number': self.registry_number})
            for header, header_value in table_data.items():
                if header_value is not None:
                    if header.__contains__('Datum privremenog prekida'):
                        data = dict()
                        start_date, end_date = re.search('(\d+.\d+\.\d+).*?-(.*?)', header_value, re.DOTALL | re.MULTILINE | re.IGNORECASE).groups()
                        start_date = start_date.strip()
                        end_date = end_date.strip()
                        if start_date != '':
                            start_date = date_convert(start_date)
                            if start_date is not None:
                                data.update({'pause_start': start_date})
                        if end_date != '':
                            end_date = date_convert(end_date)
                            if end_date is not None:
                                data.update({'pause_end': end_date})
                        if data:
                            basic_names.update(data)
                            sha = row_sha(basic_names, activity_pause_cols)
                            basic_names.update({'sha': sha})
                            self.data = basic_names
            return 200
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
            return error


class ContactData:
    def __init__(self, registry_number, html, registry, keyword, cols, col_name):
        self.registry_number = registry_number
        self.html = html
        self.registry = registry
        self.data = dict()
        self.keyword = keyword
        self.cols = cols
        self.col_name = col_name

    def parse(self):
        try:
            data = list()
            cols = self.col_name
            escape = True
            if self.html.decode().__contains__('Е-пошта') :
                escape = False

            table_data = tables(self.html, self.registry_number, self.cols, escape=escape)
            basic_names = {c: None for c in self.cols}
            basic_names.update({'registry_number': self.registry_number})
            for header, header_value in table_data.items():
                if header_value is not None:

                    if header.__contains__(self.keyword):
                        if self.keyword == 'E-pošta':
                            parse_data = header_value
                        else:
                            parse_data = value_replace(header_value)
                        if parse_data is not None and parse_data not in data:
                            number = parse_data.replace(";", " | ")
                            if len(parse_data) > 0:
                                data.append(parse_data)

            for parse_d in data:
                basic_names_c = basic_names.copy()
                basic_names_c.update({self.col_name: parse_d})
                sha = row_sha(basic_names_c, self.cols)
                basic_names_c.update({'sha': sha})
                self.data.update({len(self.data): basic_names_c})

            return 200

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': f"{self.keyword}: {str(exc_type)}", 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
            return error


class Phone(ContactData):
    def __init__(self, registry_number, html, registry):
        hashs = HashCreator()
        phone_cols = hashs.phone_cols
        ContactData.__init__(self, registry_number, html, registry, 'Telefon', phone_cols, 'phone_number')


class Fax(ContactData):
    def __init__(self, registry_number, html, registry):
        hashs = HashCreator()
        fax_cols = hashs.fax_cols
        ContactData.__init__(self, registry_number, html, registry, 'Faks', fax_cols, 'fax_number')


class Web(ContactData):
    def __init__(self, registry_number, html, registry):
        hashs = HashCreator()
        web_page_cols = hashs.web_page_cols
        ContactData.__init__(self, registry_number, html, registry, 'Internet adresa', web_page_cols, 'web_url')


class Email(ContactData):
    def __init__(self, registry_number, html, registry):
        # if re.search('Е-пошта', html.decode(), re.MULTILINE) is not None:
        #     print(html.decode())
        #     print("converted")
        #     print(cirilic_to_latin(html.decode()))
        #     # exit()
        hashs = HashCreator()
        email_cols = hashs.email_cols
        ContactData.__init__(self, registry_number, html, registry, 'E-pošta', email_cols, 'email')


class BankAccounts:
    def __init__(self, registry_number, html, registry):
        self.registry_number = registry_number
        self.html = html
        self.registry = registry
        self.data = dict()

    def parse(self):
        try:
            hashs = HashCreator()
            bank_account_cols = hashs.bank_account_cols
            html = cirilic_to_latin(self.html.decode())
            basic_names = {c: None for c in bank_account_cols}
            basic_names.update({'registry_number': self.registry_number})
            bank_accounts = re.findall("\d{3}-\d{10,}-\d{1,}", html, re.MULTILINE | re.DOTALL | re.IGNORECASE)

            if bank_accounts:
                for bank_account in bank_accounts:
                    basic_names_c = basic_names.copy()
                    basic_names_c.update({'bank_number': value_replace(bank_account)})
                    sha = row_sha(basic_names_c, bank_account_cols)
                    basic_names_c.update({'sha': sha})
                    self.data.update({len(self.data): basic_names_c})
            return 200

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
            return error


class EraseDetails:
    def __init__(self, registry_number, html, registry):
        self.registry_number = registry_number
        self.html = html
        self.registry = registry
        self.data = dict()

    def parse(self):
        try:
            data = dict()
            hashs = HashCreator()
            erase_details_cols = hashs.erase_details_cols
            table_data = tables(self.html, self.registry_number, erase_details_cols)
            basic_names = {c: None for c in erase_details_cols}
            basic_names.update({'registry_number': self.registry_number})
            for header, header_value in table_data.items():
                if header_value is not None:
                    if header.__contains__('Razlog prestanka obavljanja delatnosti'):
                        data.update({'activity_end_reason': header_value.strip()})
                    if header.__contains__('Datum brisanja'):
                        data.update({'erase_date': date_convert(header_value)})
                    if header.__contains__('Datum prestanka obavljanja delatnosti'):
                        data.update({'activity_end_date': date_convert(header_value)})
            if data:
                basic_names.update(data)
                sha = row_sha(basic_names, erase_details_cols)
                basic_names.update({'sha': sha})
                self.data = basic_names
            return 200

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
            return error


class AssociationExtra:
    def __init__(self, registry_number, html, registry):
        self.registry_number = registry_number
        self.html = html
        self.registry = registry
        self.data = dict()

    def parse(self):
        try:
            data = dict()
            hashs = HashCreator()
            association_extra_cols = hashs.association_extra_cols
            table_data = tables(self.html, self.registry_number, association_extra_cols)
            basic_names = {c: None for c in association_extra_cols}
            basic_names.update({'registry_number': self.registry_number})
            for header, header_value in table_data.items():
                if header_value is not None:
                    if len(header_value) > 0:
                        if header.__contains__('Datum donošenja statuta'):
                            data.update({'statute_date': date_convert(header_value)})
                        if header.__contains__('Datum poslednjih izmena'):
                            data.update({'statute_change': date_convert(header_value)})
                        if header.__contains__('Šifra privredne delatnosti'):
                            data.update({'activity_code2': value_replace(header_value)})
                        if header.__contains__('Opis privredne delatnosti'):
                            data.update({'activity_name2': value_replace(header_value)})
                        if header.__contains__('Šifra i opis sportske delatnosti'):
                            split_activity = re.search('(\d+)\s+(.*)', header_value, re.MULTILINE |re.IGNORECASE | re.DOTALL)
                            if split_activity is not None:
                                activity_code2, activity_name2 = split_activity.groups()
                                data.update({'activity_code2': value_replace(activity_code2), 'activity_name2': activity_name2})

            if data:
                basic_names.update(data)
                sha = row_sha(basic_names, association_extra_cols)
                basic_names.update({'sha': sha})
                self.data = basic_names

            return 200

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
            return error


class AddressData:
    def __init__(self, registry_number, html, registry, tab_name):
        self.registry_number = registry_number
        self.html = html
        self.registry = registry
        self.tab_name = tab_name
        self.data = dict()

    def address_type_calc(self, value):
        if type(value) is str:
            caption_group = {'Adresa sedišta': 'CompanyHq', 'Adresa za prijem pošte': 'CompanyMail'}
            for k, v in caption_group.items():
                if value.__contains__(k):
                    return v
        return value

    def parse(self):
        try:
            data = dict()
            from1 = ['CentralOfficeData', 'Headquarter', 'Address']
            web_cols = {'Naziv opštine': 'municipality', 'Mesto': 'place', 'Ulica , Broj i slovo': 'street',
                        'Ulica, broj i slovo': 'street', 'Broj pošte': 'zip_code',
                        'Sprat, broj stana i slovo': 'address_extra', 'Broj i slovo': 'street_number', 'Opština': 'municipality',
                        'Dodatni opis': "address_extra", 'Ulica': 'street', 'Naziv mesta': 'place', 'Broj': 'street_number',
                        'Poštanski broj': 'zip_code'}

            hashs = HashCreator()
            address_data_cols = hashs.address_data_cols
            basic_names = {c: None for c in address_data_cols}
            basic_names.update({'registry_number': self.registry_number, 'zip_code': None})

            if self.tab_name not in from1:
                table_data = tables(self.html, self.registry_number, address_data_cols, form=2)
                for caption, table_rows in table_data.items():
                    for header, header_value in table_rows.items():
                        if header_value is not None:
                            header_value = value_replace(header_value)
                            for original_name, substitute in web_cols.items():
                                if header.__contains__(original_name) or header == original_name:
                                    caption_data = data.get(caption)
                                    if caption_data is None:
                                        data.update({caption: {substitute: header_value, 'address_type': self.address_type_calc(caption)}})
                                    if caption_data is not None:
                                        caption_data.update({substitute: header_value})
                                    break

                repacked = dict()
                for add_type, add_data in data.items():
                    # merging street number data with street data
                    street_number = add_data.get('street_number')
                    street = add_data.get('street')
                    if street_number is not None and street is not None:
                        street = f"{street} {street_number}"
                        add_data.update({'street': street})
                    if 'street_number' in list(add_data.keys()):
                        add_data.pop('street_number')

                    basic_names_c = basic_names.copy()
                    basic_names_c.update(add_data)
                    sha = row_sha(basic_names_c, address_data_cols)
                    basic_names_c.update({'sha': sha})
                    repacked.update({len(repacked): basic_names_c})

                if repacked:
                    self.data = repacked

            if self.tab_name in from1:
                table_data = tables(self.html, self.registry_number, address_data_cols)
                for header, header_value in table_data.items():
                    if header_value is not None:
                        header_value = value_replace(header_value)
                        for original_name, substitute in web_cols.items():
                            if header.__contains__(original_name) or header == original_name:
                                data.update({substitute: value_replace(header_value), 'address_type': 'CompanyHq'})
                                break
                if data:
                    street_number = data.get('street_number')
                    street = data.get('street')
                    if street_number is not None and street is not None:
                        street = f"{street} {street_number}"
                        data.update({'street': street})
                    if 'street_number' in list(data.keys()):
                        data.pop('street_number')

                    basic_names.update(data)
                    sha = row_sha(basic_names, address_data_cols)
                    basic_names.update({'sha': sha})
                    self.data = basic_names

            return 200

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
            return error


class Annotation:
    def __init__(self, registry_number, html, registry):
        self.registry_number = registry_number
        self.html = html
        self.registry = registry
        self.data = dict()

    def parse(self):
        try:
            data = dict()
            web_cols = {'Datum': 'post_date', 'Podnosilac': 'submitter', 'Tekst': 'text_lines'}

            hashs = HashCreator()
            annotations_cols = hashs.annotations_cols
            basic_names = {c: None for c in annotations_cols}
            basic_names.update({'registry_number': self.registry_number})
            table_data = tables(self.html, self.registry_number, annotations_cols, form=2)
            for caption, table_rows in table_data.items():
                caption_data = dict()
                for header, header_value in table_rows.items():
                    if header_value is not None:
                        for original_name, substitute in web_cols.items():
                            if header.__contains__(original_name) or header == original_name:
                                if header.__contains__('Datum'):
                                    caption_data.update({substitute: date_convert(header_value)})
                                else:
                                    caption_data.update({substitute: header_value})
                                break
                if caption_data:
                    basic_names_c = basic_names.copy()
                    basic_names_c.update(caption_data)
                    sha = row_sha(basic_names_c, annotations_cols)
                    basic_names_c.update({'sha': sha})
                    self.data.update({len(self.data): basic_names_c})
            return 200

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
            return error


class Bankruptcy:
    def __init__(self, registry_number, html, registry):
        self.registry_number = registry_number
        self.html = html
        self.registry = registry
        self.data = dict()

    def convert_to_mssql(self, data):
        postgre_to_sql = {'registry_number': 'RegistryNumber', 'vat_number': 'VatNumber',
                          'bankruptcy_status': 'BankruptcyStatus', 'bankruptcy_value': 'BankruptcyValue', 'sha': 'sha'}
        return {postgre_to_sql.get(k): v for k, v in data.items() if k in list(postgre_to_sql.keys())}

    def basic_info(self):
        try:
            if type(self.html) is bytes:
                self.html = self.html.decode()
            html = cirilic_to_latin(self.html)
            hashs = HashCreator()
            bankruptcy_cols = hashs.bankruptcy_cols
            basic_names = {c: None for c in bankruptcy_cols}
            basic_names.update({'registry_number': self.registry_number})
            df = pd.read_html(html, header=None, decimal=',', thousands='.')[0]
            db_col = {'Matični broj': 'registry_number', 'PIB': 'vat_number',
                      'Status stečajne mase': 'bankruptcy_status',
                      'Vrednost stečajne mase': 'bankruptcy_value'}
            df = df.where((pd.notnull(df)), None)
            data = dataframe_convert_data(df, db_col)
            if data:
                if 'registry_number' in list(data.keys()):
                    data.pop('registry_number')
                basic_names.update(data)
                self.data.update(basic_names)
            return 200

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
            return error

    def parse(self, html, tab_name):
        data = dict()
        if type(html) is bytes:
            html = html.decode()
        html = cirilic_to_latin(html)
        hashs = HashCreator()
        bankruptcy_cols = hashs.bankruptcy_cols
        table_data = tables(html, self.registry_number, bankruptcy_cols)
        cols = {'Debtor': {'Matični broj': 'debtor_reg_num', 'Naziv': 'debtor'},
                'Person': {'Matični broj': 'cregistry_number', 'Poslovno ime': 'cname', 'Ime': 'pname', 'Prezime': 'psurname', 'JMBG': 'pin'},
                'Solution': {'Način nastanka': 'creation_type', 'Naziv suda': 'court_name', 'Broj rešenja': 'case_number', 'Datum rešenja': 'case_date'}}

        for header, header_value in table_data.items():
            if header_value is not None:
                tab_headers = cols.get(tab_name)
                if tab_headers is not None:
                    for original_name, substitute in tab_headers.items():
                        if header.__contains__(original_name) or header == original_name:
                            if header.__contains__('Datum'):
                                data.update({substitute: date_convert(header_value)})
                            else:
                                data.update({substitute: header_value})
                            break

        if data:
            self.data.update(data)


class BranchOffices:
    def __init__(self, registry_number, html, registry):
        self.registry_number = registry_number
        self.html = html
        self.registry = registry
        self.data = dict()
        self.data_shas = list()
        self.persons = list()
        self.persons_shas = list()

    def convert_to_mssql(self, data):
        raise NotImplemented

    def people_data(self, df, branch_name, ptype='PrivatePerson'):
        pgroup = 'BranchTrustees'
        hashs = HashCreator()
        company_people_cols = hashs.company_people_cols
        people_names = {c: None for c in company_people_cols}
        people_names.update({'registry_number': self.registry_number})

        person_cols = {'Ime i prezime': 'pname', 'Funkcija': 'pfunction', 'Lični broj za strance': 'pin2',
                       'Broj pasoša': 'passport_number', 'Država izdavanja': 'passport_state', 'JMBG': 'pin',
                       'Ograničenja supotpisom': 'limitations'}

        new_col_names = list()
        for col in df.columns:
            is_na = True
            df[col] = [value_replace(c) for c in df[col]]
            for orginal, substitute in person_cols.items():
                if col.lower().__contains__(orginal.lower()) is True:
                    new_col_names.append(substitute)
                    is_na = False
                    break
            if is_na is True:
                new_col_names.append(None)

        df.columns = new_col_names
        df.dropna(axis=0, how='all', inplace=True)
        df['ptype'] = [ptype] * len(df)
        df['pgroup'] = [pgroup] * len(df)
        df.drop(columns=[None], inplace=True)
        if 'pin' in df.columns and 'pin2' in df.columns:
            df['pin'] = df.pin.combine_first(df.pin2)
        if 'pin2' in df.columns:
            df.drop(columns=['pin2'], inplace=True)
        df['branch'] = [branch_name] * len(df)

        person_data = df.to_dict('records')
        for row in person_data:
            row = {k: v for k, v in row.items() if k in company_people_cols}
            people_names_c = people_names.copy()
            people_names_c.update(row)
            sha = row_sha(people_names_c, company_people_cols)
            people_names_c.update({'sha': sha})
            if sha not in self.persons_shas:
                self.persons.append(people_names_c)
                self.persons_shas.append(sha)

    def address_clean(self, value):
        if type(value) is str:
            value = re.sub('None', '', value)
            value = value.strip()
            if value in ['/', '//', '-', '--', '']:
                value = None
        return value

    def branch_data(self, df, branch_name, convert_data=None):
        try:
            if df is not None:
                df = df.dropna(how='all', axis=1)
                if 1 not in df.columns:
                    return

            hashs = HashCreator()
            branch_offices_cols = hashs.branch_offices_cols
            basic_names = {c: None for c in branch_offices_cols}
            basic_names.update({'registry_number': self.registry_number})
            if convert_data is None:
                col_rep = {'Naziv': 'bname', 'Šifra i naziv delatnosti': 'activity_data', 'Naziv opštine': 'municipality', 'Mesto': 'place', 'Ulica, broj i slovo': 'street',
                           'Šifra i opis sportske delatnosti': 'activity_data', 'Sprat, broj stana i slovo': 'address_extra',
                           'Naziv ogranka': 'bname', 'Ulica i broj': 'street', 'Opština': 'municipality', 'Šifra delatnosti:': 'activity_code',
                           'Šifra delatnosti': 'activity_code', 'Naziv delatnosti': 'activity_name', 'Ulica': 'street', 'Broj i slovo': 'street1', 'Dodatni opis': 'description',
                           'Naziv ogranka predstavništva': 'bname'}

                convert_data = dict(df.to_dict(orient='split').get('data'))
                new_df = pd.DataFrame.from_dict({0: convert_data}, orient='index')
                new_df.columns = [col_rep.get(value_replace(col)) for col in new_df.columns]
                if 'street1' in new_df.columns and 'street' in new_df.columns:
                    new_df["street"] = [self.address_clean(x) for x in new_df["street"]]
                    new_df["street1"] = [self.address_clean(x) for x in new_df["street1"]]
                    new_df["street"] = new_df["street"].astype(str) + " " + new_df["street1"].astype(str)
                    new_df["street"] = [self.address_clean(x) for x in new_df["street"]]
                    new_df.drop(columns=['street1'], inplace=True)

                if 'activity_data' in new_df.columns:
                    activity_data = new_df['activity_data']
                    activity_codes, activity_names = [], []
                    for ad in activity_data:
                        activity_code, activity_name = None, None
                        split_data = re.search('(\d+)\s*-?(.*)', ad)
                        if split_data is not None:
                            activity_code, activity_name = split_data.groups()
                        activity_codes.append(activity_code)
                        activity_names.append(activity_name)
                    new_df.pop('activity_data')
                    new_df['activity_code'] = activity_codes
                    new_df['activity_name'] = activity_names
                new_df['branch'] = [branch_name] * len(new_df)
                if None in new_df.columns:
                    new_df.drop(columns=[None], inplace=True)
                convert_data = new_df.to_dict('records')

            if convert_data is not None:
                for row in convert_data:
                    branch_data = {k: value_replace(v) for k, v in row.items()}
                    basic_names_c = basic_names.copy()
                    basic_names_c.update(branch_data)
                    sha = row_sha(basic_names_c, branch_offices_cols)
                    basic_names_c.update({'sha': sha})
                    if sha not in self.data_shas:
                        self.data.update({len(self.data): basic_names_c})
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            raise Exception(str(error))

    def parse(self, tab_name):
        try:
            if type(self.html) is bytes:
                self.html = self.html.decode()
            html = cirilic_to_latin(self.html)

            # parser variation for BranchOffices in Sport Association registry
            if tab_name == 'BranchOffices' and self.registry in ['SportAssociation']:
                soup = BeautifulSoup(html, 'lxml')
                divs = soup.find_all('div', {'class': 'row-fluid'})
                for div in divs:
                    parent_name = None
                    try:
                        parent = div.parent['class'][0]
                        if parent == 'span12':
                            parent_name = parent
                    except KeyError:
                        pass
                    except IndexError:
                        pass

                    if parent_name is not None:
                        tables = div.find_all('table')
                        branch_name = 'Unknown'
                        for table in tables:
                            table_caption = table.caption
                            if table_caption is not None:
                                table_caption = value_replace(table_caption.text).strip('.').strip()
                            df = pd.read_html(table.decode())
                            converters = {column_name: str for column_name in df[0].dtypes.index}
                            df = pd.read_html(table.decode(), converters=converters)
                            if df:
                                df = df[0]
                                if df.empty is False:
                                    if 0 in df.columns:
                                        branch_name = table_caption
                                        self.branch_data(df,table_caption)
                                    else:
                                        self.people_data(df, branch_name)

            if tab_name == 'BranchOffices' and self.registry not in ['SportAssociation']:

                html = re.sub('<br />\s+<br />', '</table>', html, re.MULTILINE | re.DOTALL | re.IGNORECASE)
                html = re.sub('<h4', '<table class="main_group"><h4', html, re.MULTILINE | re.DOTALL | re.IGNORECASE)
                soup = BeautifulSoup(html, 'lxml')
                tables = soup.find_all('table', {'class': 'main_group'})

                for table in tables:
                    branch_name = table.h4
                    if branch_name is not None:
                        branch_name = value_replace(branch_name.text)
                    inner_tables = table.find_all('table')
                    for inner_table in inner_tables:
                        table_caption = inner_table.caption
                        if table_caption is not None:
                            table_caption = table_caption.text

                        df = pd.read_html(inner_table.decode())
                        converters = {column_name: str for column_name in df[0].dtypes.index}
                        df = pd.read_html(inner_table.decode(), converters=converters)
                        if df:
                            df = df[0]
                            if df.empty is False:
                                if 0 in df.columns:
                                     self.branch_data(df, branch_name)
                                else:
                                    ptype = 'PrivatePerson'
                                    if table_caption.lower().__contains__('fizič') is False:
                                        ptype = 'LegalPerson'
                                    self.people_data(df, branch_name, ptype=ptype)

            if tab_name in ['IzdvojenaMesta', 'BranchOfficeData']:
                soup = BeautifulSoup(html, 'lxml')
                tables = soup.find_all("table")
                for table in tables:
                    table_caption = table.caption
                    if table_caption is not None:
                        table_caption = table_caption.text.strip('.').strip()
                    df = pd.read_html(table.decode())
                    converters = {column_name: str for column_name in df[0].dtypes.index}
                    df = pd.read_html(table.decode(), converters=converters)
                    if df:
                        df = df[0]
                        if df.empty is False:
                            if 0 in df.columns:
                                self.branch_data(df, table_caption)

            if tab_name in ['Branches']:
                if re.search('table', html, re.MULTILINE | re.DOTALL | re.IGNORECASE) is not None:
                    df = pd.read_html(html, header=0)
                    converters = {column_name: str for column_name in df[0].dtypes.index}
                    df = pd.read_html(html, converters=converters)
                    if df:
                        df = df[0]
                        if df.empty is False:
                            replace = {'Naziv ogranka': 'bname', 'Opština': 'municipality',
                                       'Adresa': 'street', 'Poštanski broj': 'address_extra'}
                            df.columns = [replace.get(col) for col in df.columns]
                            df = df.where((pd.notnull(df)), None)
                            if 'bname' in df.columns:
                                branch_numbers = list()
                                for x in df['bname']:
                                    if type(x) is str:
                                        branch_number = re.search('ogranak\s+(\d+)', x, re.MULTILINE | re.DOTALL | re.IGNORECASE)
                                        if branch_number is not None:
                                            branch_numbers.append(f"Ogranak {branch_number.group(1)}")
                                        else:
                                            branch_numbers.append(None)
                                df['branch'] = branch_numbers
                                branches = df.to_dict('records')
                                self.branch_data(None, None, convert_data=branches)

            return 200

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
            return error


class Capital:
    def __init__(self, registry_number, html, registry):
        self.registry_number = registry_number
        self.registry = registry
        self.html = html
        self.data = dict()

    def parse(self):
        try:
            html = self.html
            hashs = HashCreator()
            capital_cols = hashs.capital_cols

            share_groups = {'Novčani kapital': "MoneyShare", 'Nenovčani kapital': "NonMoneyShare"}
            group_headers = ['capital_type', 'amount_data', 'event_date', 'share_in']

            if type(html) is bytes:
                html = html.decode()

            html = cirilic_to_latin(html)
            soup = BeautifulSoup(html, "lxml")
            tables_ = soup.find_all("table")
            basic_names = {c: None for c in capital_cols}
            basic_names.update({'registry_number': self.registry_number})

            for table in tables_:
                table_caption = table.caption
                if table_caption:
                    share_group = 'Unknown'
                    table_caption = table_caption.text.strip()

                    for original, substitute in share_groups.items():
                        if table_caption.__contains__(original) is True:
                            share_group = substitute

                    df = pd.read_html(table.decode())
                    if df:
                        df = df[0]
                        if df.empty is False:
                            df = df.where((pd.notnull(df)), None)
                            if len(df.columns) == len(group_headers):
                                df.columns = group_headers

                                split_amount_data = [row.split() for row in df['amount_data']]
                                amount, currency = list(), list()

                                [(amount.append(row[0]), currency.append(row[1])) for row in split_amount_data]

                                amount = [convert_amount(row) for row in amount]
                                currency = [value_replace(row) for row in currency]

                                df['share_group'] = [share_group] * len(df)
                                df['amount'] = amount
                                df['currency'] = currency
                                df['event_date'] = [date_convert(row) for row in df['event_date']]
                                df.drop(columns=['amount_data'], inplace=True)

                                capital_datas = df.to_dict('records')
                                if capital_datas:
                                    for capital_data in capital_datas:
                                        basic_names_c = basic_names.copy()
                                        basic_names_c.update(capital_data)
                                        basic_names_c = {k: value_replace(v) for k, v in basic_names_c.items()}
                                        sha = row_sha(basic_names_c, capital_cols)
                                        basic_names_c.update({'sha': sha})
                                        self.data.update({len(self.data): basic_names_c})

            return 200

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
            return error


class Members:
    def __init__(self, registry_number, html, registry):
        self.registry_number = registry_number
        self.html = html
        self.registry = registry
        self.amounts = dict()
        self.groups = dict()
        self.shares = dict()
        self.people = list()

    def parse_amounts(self, soup, group_id, owner_type, person_group_id=None):

        share_groups = {'Novčani kapital': "MoneyShare", 'Nenovčani kapital': "NonMoneyShare",
                        'Novčani ulog': 'MoneyShare', 'Nenovčani ulog': "NonMoneyShare",}
        group_headers = {'Unnamed: 0': 'capital_type', 'Vrednost': 'amount_data', 'Datum': 'event_date', 'Unnamed: 3': 'share_in'}

        tables_ = soup.find_all("table")
        amounts_data_all = list()
        for table in tables_:

            table_caption = table.caption
            if table_caption:
                share_group = 'Unknown'
                table_caption = table_caption.text.strip()
                for original, substitute in share_groups.items():
                    if table_caption.__contains__(original) is True:
                        share_group = substitute

                df = pd.read_html(table.decode())
                if df:
                    df = df[0]
                    if df.empty is False:
                        df = df.where((pd.notnull(df)), None)
                        df.columns = [group_headers.get(col) for col in df.columns]
                        if 'amount_data' in df.columns:
                            split_amount_data = [row.split() for row in df['amount_data']]
                            amount, currency = list(), list()

                            [(amount.append(row[0]), currency.append(row[1])) for row in split_amount_data]

                            amount = [convert_amount(row) for row in amount]
                            currency = [value_replace(row) for row in currency]

                            if person_group_id is not None:
                                df['person_group_id'] = [person_group_id] * len(df)
                            df['share_group'] = [share_group] * len(df)
                            df['amount'] = amount
                            df['currency'] = currency
                            df['event_date'] = [date_convert(row) for row in df['event_date']]
                            df['group_id'] = [group_id] * len(df)
                            df['owner_type'] = [owner_type] * len(df)
                            df.drop(columns=['amount_data'], inplace=True)

                            amounts_data = df.to_dict('records')
                            amounts_data_all.extend(amounts_data)
        return amounts_data_all

    def pin_fix(self, pin):
        if type(pin) is str:
            pin = pin.zfill(13)
        return pin

    def parse(self):
        try:
            p_group = 'CapitalMembers'
            hashs = HashCreator()

            company_people_cols = hashs.company_people_cols
            c_people = {c: None for c in company_people_cols}
            c_people.update({'registry_number': self.registry_number})

            capital_groups_cols = hashs.capital_groups_cols
            capital_groups = {c: None for c in capital_groups_cols}
            capital_groups.update({'registry_number': self.registry_number})

            capital_amount_cols = hashs.capital_amount_cols
            capital_amount = {c: None for c in capital_amount_cols}
            capital_amount.update({'registry_number': self.registry_number})

            capital_shares_cols = hashs.capital_shares_cols
            capital_shares = {c: None for c in capital_shares_cols}
            capital_shares.update({'registry_number': self.registry_number})

            html = self.html
            if type(html) is bytes:
                html = html.decode()

            col_rep = {'Ime i prezime': 'pname', 'JMBG/Lični broj': 'pin', 'Procenat u suvlasništvu udela': 'in_group_share',
                       'Matični broj': 'creg_num', 'Poslovno ime': 'cname', 'Broj pasoša': 'passport_number',
                       'Država izdavanja': 'passport_state', 'Udeo': 'in_group_share', 'JMBG': 'pin', 'Lični broj za strance': 'pin2'}

            html = cirilic_to_latin(html)

            html_split = html.split('<hr />')
            group_id = -1

            for html_data in html_split:
                # print(html_data)

                soup = BeautifulSoup(html_data, "lxml")

                # parsing data of group share
                h3 = soup.h3
                if h3 is not None:
                    # print(soup)
                    owner_type = 'ShareOwnerGroup'
                    group_id_search = re.search(r'(\d+)', h3.text, re.IGNORECASE |re.MULTILINE | re.DOTALL)
                    if group_id_search is not None:
                        group_id = group_id_search.group(1)
                        share_search = re.search('<th class="apr-label">Udeo</th>\s+<td class="apr-general-data">(.*?)%</td>', html_data , re.IGNORECASE | re.MULTILINE |re.DOTALL)
                        if share_search is not None:
                            capital_share = convert_amount(share_search.group(1))
                            capital_shares_c = capital_shares.copy()
                            capital_shares_c.update({'capital_share': capital_share, 'group_id': group_id, 'owner_type': owner_type})
                            sha = row_sha(capital_shares_c, capital_shares_cols)
                            capital_shares_c.update({'sha': sha})
                            self.shares.update({len(self.shares): capital_shares_c})

                        amounts_datas = self.parse_amounts(soup, group_id, owner_type)
                        if amounts_datas:
                            for amounts_data in amounts_datas:
                                capital_amount_c = capital_amount.copy()
                                capital_amount_c.update(amounts_data)
                                capital_amount_c = {k: value_replace(v) for k, v in capital_amount_c.items()}
                                sha = row_sha(capital_amount_c, capital_amount_cols)
                                capital_amount_c.update({'sha': sha})
                                self.amounts.update({len(self.amounts): capital_amount_c})

                variations = ['table', 'div']
                # variations = ['div']
                for variation in variations:
                    divs = soup.find_all(variation)
                    for div in divs:
                        sibilings = div.next_siblings
                        next_sibling_name = None
                        next_sibling_data = None
                        for sibling in sibilings:
                            next_sibling_name = sibling.name
                            if next_sibling_name == 'div':
                                next_sibling_data = sibling
                                break

                        if next_sibling_name == 'div' and next_sibling_data is not None:
                            # person data
                            person_group = None
                            person_table = div

                            if person_table is not None:
                                table_caption = person_table.caption

                                if table_caption is not None:
                                    table_caption = table_caption.text.strip()
                                    person_group_search = re.search('(\d+)', table_caption)
                                    if person_group_search is not None:
                                        group_ = person_group_search.group(1)
                                        if variation == 'table':
                                            group_id = group_
                                        else:
                                            person_group = group_

                                # print(group_id, person_group)
                                if person_group is not None or variation == 'table':

                                    df = pd.read_html(person_table.decode())
                                    if df:
                                        df = df[0]
                                        if df.empty is False:
                                            converters = {column_name: str for column_name in df.dtypes.index}
                                            df = pd.read_html(person_table.decode(), converters=converters)
                                            df = df[0]
                                            df = df.dropna(how='all', axis=1)
                                            if 1 not in df.columns:
                                                continue

                                            convert_data = dict(df.to_dict(orient='split').get('data'))
                                            new_df = pd.DataFrame.from_dict({0: convert_data}, orient='index')
                                            new_df.columns = [col_rep.get(value_replace(col)) for col in new_df.columns]
                                            if None in new_df.columns:
                                                new_df.drop(columns=[None], inplace=True)

                                            p_type = 'PrivatePerson'
                                            if 'creg_num' in new_df.columns:
                                                p_type = 'LegalPerson'

                                            if 'pin' in new_df.columns:
                                                new_df['pin'] = [pin_fix(row) for row in new_df['pin']]
                                            else:
                                                new_df['pin'] = [None]* len(new_df)

                                            if 'pin' in new_df.columns and 'pin2' in new_df.columns:
                                                new_df['pin'] = new_df.pin.combine_first(new_df.pin2)
                                            if 'pin2' in new_df.columns:
                                                new_df.drop(columns=['pin2'], inplace=True)

                                            new_df['group_id'] = [group_id] * len(new_df)
                                            new_df['pgroup'] = [p_group] * len(new_df)
                                            new_df['ptype'] = [p_type] * len(new_df)
                                            if variation == 'div':
                                                new_df['person_group_id'] = [group_id+person_group] * len(new_df)
                                            if variation == 'table':
                                                new_df['owner_type'] = ['ShareOwner'] * len(new_df)
                                            else:
                                                new_df['owner_type'] = ['CapitalGroupMember'] * len(new_df)

                                            if variation == 'table' and 'in_group_share' in new_df.columns:
                                                new_df['capital_share'] = [convert_amount(row.strip('%').strip()) for row in new_df['in_group_share']]
                                                new_df.drop(columns=['in_group_share'], inplace=True)

                                            if 'in_group_share' in new_df.columns:
                                                new_df['in_group_share'] = [convert_amount(row.strip('%').strip()) for row in new_df['in_group_share']]

                                            if 'cname' in new_df.columns:
                                                new_df.loc[new_df['cname'] == 'Akcijski kapital', ['creg_num', 'pin']] = '0'

                                            df_columns = new_df.columns
                                            if 'pin' in df_columns or 'creg_num' in df_columns or 'passport_number' in df_columns:
                                                parsed_datas = new_df.to_dict('records')
                                                for parsed_data in parsed_datas:
                                                    pin = parsed_data.get('pin')
                                                    creg_num = parsed_data.get('creg_num')
                                                    passport_number = parsed_data.get('passport_number')
                                                    if pin is not None or creg_num is not None or passport_number is not None:
                                                        # part of data for company people table
                                                        extra_cols = company_people_cols + ['ptype', 'pgroup']
                                                        c_people_data = {k: v for k, v in parsed_data.items() if k in extra_cols}
                                                        if c_people_data:
                                                            c_people_c = c_people.copy()
                                                            c_people_c.update(c_people_data)
                                                            c_people_c = {k: value_replace(v) for k, v in c_people_c.items()}
                                                            sha = row_sha(c_people_c, company_people_cols)
                                                            c_people_c.update({'sha': sha})
                                                            self.people.append(c_people_c)

                                                        # part of data for company groups
                                                        c_group_data = {k: v for k, v in parsed_data.items() if k in capital_groups_cols}
                                                        if c_group_data:
                                                            capital_groups_c = capital_groups.copy()
                                                            capital_groups_c.update(c_group_data)
                                                            capital_groups_c = {k: value_replace(v) for k, v in capital_groups_c.items()}
                                                            sha = row_sha(capital_groups_c, capital_groups_cols)
                                                            capital_groups_c.update({'sha': sha})
                                                            self.groups.update({len(self.groups): capital_groups_c})

                                                        # part of data for company shares
                                                        c_share_data = {k: v for k, v in parsed_data.items() if k in capital_shares_cols}
                                                        if c_share_data:
                                                            capital_shares_c = capital_shares.copy()
                                                            capital_shares_c.update(c_share_data)
                                                            capital_shares_c = {k:value_replace(v) for k, v in capital_shares_c.items()}
                                                            sha = row_sha(capital_shares_c, capital_shares_cols)
                                                            capital_shares_c.update({'sha': sha})
                                                            self.shares.update({len(self.shares): capital_shares_c})

                            # capital split
                            capital_split = next_sibling_data.table
                            if (person_group is not None and variation == 'div') or (h3 is None and variation =='table'):
                                if capital_split is not None:
                                    owner_type = 'CapitalGroupMember'
                                    person_group_id = None
                                    if variation == 'table':
                                        owner_type = 'ShareOwner'
                                    if variation == 'div':
                                        person_group_id = group_id + person_group
                                    amounts_datas = self.parse_amounts(soup, group_id, owner_type, person_group_id)
                                    if amounts_datas:
                                        for amounts_data in amounts_datas:
                                            capital_amount_c = capital_amount.copy()
                                            capital_amount_c.update(amounts_data)
                                            capital_amount_c = {k: value_replace(v) for k, v in capital_amount_c.items()}
                                            sha = row_sha(capital_amount_c, capital_amount_cols)
                                            capital_amount_c.update({'sha': sha})
                                            self.amounts.update({len(self.amounts): capital_amount_c})
            return 200
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
            return error


class CompanyPeople:
    def __init__(self, registry_number, registry):
        self.registry_number = registry_number
        self.registry = registry
        self.data = list()
        self.sha_data = list()

    def ptype_calc(self, caption):
        ptype = 'Unknown'
        rex = {'fizič': 'PrivatePerson', 'privred': 'LegalPerson',
               'pojedinačna proku': 'SingleProcurist', 'zajednička proku': 'CommonProcurist',
               'upravni': 'PrivatePersonSteering', 'upravitelj': 'PrivatePersonManager'}

        for k, v in rex.items():
            if caption.lower().__contains__(k):
                return v
        return ptype

    def handle_dataframe_data(self, html, limitations, ptype, pgroup, company_people_cols, c_people):
        try:
            person_cols = {"Ime i prezime": "pname", "Funkcija": "pfunction", "JMBG": "pin", "Broj pasoša i država izdavanja": "passport_data",
                           "PersonType": "ptype", "Ograničenja supotpisom": "limitations", "Lični broj za strance": 'pin2', 'Samostalno zastupa': 'represent',
                           "Poslovno ime": "cname", "Matični broj": "creg_num", "Broj pasoša / Izbegličke legitimacije": "passport_number",
                           "Država izdavanja": "passport_state", "JMBG / Lični broj": "pin", "Broj pasoša": "passport_number",
                           "Ime": "pname2", "Prezime": "psurname2", "JMBG/Broj pasoša": "pin3", "JMBG/Lični broj": "pin",
                           'Broj lične karte stranca': None, 'JMBG / Broj pasoša i država izdavanja': 'pin4',
                           'Ime i prezime / naziv': 'pname', 'JMBG / Broj pasoša': "pin3",
                           'Matični broj pravnog lica / Broj upisa u registar': "creg_num"}

            df = pd.read_html(html)

            if df:
                df = df[0]
                if df.empty is False:
                    converters = {column_name: str for column_name in df.dtypes.index}
                    df = pd.read_html(html, converters=converters)
                    df = df[0]

                    df.replace(regex={r'^s*-\s*$': None}, inplace=True)

                    if limitations:
                        df['Ograničenja supotpisom'] = [list(limitations.values())[0]] * len(df)

                    if 1 in df.columns:
                        split_data = df.to_dict('split').get('data')
                        split_data = {sd[0]: sd[1] for sd in split_data}
                        df = pd.DataFrame.from_dict({0: split_data}, orient='index')

                    # print(df.columns)
                    df.columns = [person_cols.get(col) for col in df.columns]
                    if None in df.columns:
                        df.drop(columns=[None], inplace=True)
                    if 'pin' in df.columns:
                        df['pin'] = [pin_fix(row) for row in df['pin']]
                    else:
                        df['pin'] = [None] * len(df)

                    variations = ['pin2', 'pin3', 'pin4']
                    for variation in variations:
                        if 'pin' in df.columns and variation in df.columns:
                            df['pin'] = df.pin.combine_first(df[variation])
                        if variation in df.columns:
                            df.drop(columns=[variation], inplace=True)

                    if 'passport_number' not in df.columns:
                        df['passport_number'] = [None]*len(df)

                    # pin and passport validation
                    row_nums = list(range(0, len(df)))
                    pins, passports = list(), list()
                    for row_num in row_nums:
                        pin_validation = df['pin'][row_num]
                        passport_number = df['passport_number'][row_num]
                        if type(pin_validation) is str and passport_number is None:
                            if (len(pin_validation) < 12 or len(pin_validation) > 13) and pin_validation != '0':
                                pins.append(None)
                                passports.append(pin_validation)
                            else:
                                pins.append(pin_validation)
                                passports.append(None)
                        else:
                            pins.append(pin_validation)
                            passports.append(passport_number)
                    df['pin'] = pins
                    df['passport_number'] = passports

                    if 'pname2' in df.columns and 'psurname2' in df.columns :
                        df['pname2'].fillna('')
                        df['psurname2'].fillna('')
                        df['pname'] = df['pname2'] + " " + df['psurname2']
                        df.drop(columns=['pname2', 'psurname2'], inplace=True)


                    df.replace(regex={r'^s*-\s*$': None}, inplace=True)
                    if pgroup in ['CompanyFounder', 'Establishers']:
                        df['pfunction'] = ['Osnivač'] * len(df)

                    person = df.to_dict('records')

                    # value strip, replace etc.
                    for p in person:
                        pin = p.get('pin')
                        passport_number = p.get('passport_number')
                        creg_num = p.get('creg_num')
                        pname = p.get('pname')
                        cname = p.get('cname')


                        if pin is not None or passport_number is not None or creg_num is not None:
                            ptypes = ['SingleProcurist', 'CommonProcurist', 'PrivatePersonSteering', 'PrivatePersonManager']
                            if ptype not in ptypes:
                                none_vals = ['nan', 'nat', 'none']
                                if str(creg_num).lower() not in none_vals:
                                    ptype = 'LegalPerson'
                                    if str(pname).lower() not in none_vals and  str(cname).lower() in none_vals:
                                        p.update({'pname': None, 'cname': pname})

                                if str(pin).lower() not in none_vals or str(passport_number).lower() not in none_vals:
                                    ptype = 'PrivatePerson'

                            p.update({'ptype': ptype, 'pgroup': pgroup})
                            c_people_c = c_people.copy()
                            c_people_c.update(p)
                            c_people_c = {k: value_replace(v) for k, v in c_people_c.items()}
                            sha = row_sha(c_people_c, company_people_cols)
                            if not sha in self.sha_data:
                                c_people_c.update({'sha': sha})
                                self.data.append(c_people_c)
                                self.sha_data.append(sha)
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
            return error

    def parse(self, html, tab_name):
        try:
            if tab_name in list(tab_replacements.keys()):
                tab_name = tab_replacements.get(tab_name)

            pgroup = tab_name
            hash = HashCreator()
            company_people_cols = hash.company_people_cols
            c_people = {c: None for c in company_people_cols}
            c_people.update({'registry_number': self.registry_number})

            if type(html) is bytes:
                html = html.decode()
            html = cirilic_to_latin(html)

            # print(html)
            soup = BeautifulSoup(html, 'lxml')
            tables_ = soup.find_all("table")

            # print("======="*20)
            for table in tables_:
                ptype = 'PrivatePerson'
                if not tab_name in ['SupervisoryBoard', 'SteeringCommittee', 'ExecutiveBoard', 'Trustees',
                                    'Managers', 'Trusstess', 'Headman', 'CompanyFounder']:
                    ptype = 'LegalPerson'

                table_caption = table.caption
                # print(table_caption)
                if table_caption is None:
                    self.handle_dataframe_data(table.decode(), None, ptype, pgroup, company_people_cols, c_people)

                if table_caption is not None:
                    table_caption = table_caption.text.strip()
                    ptype = self.ptype_calc(table_caption)
                    # print(table_caption,ptype)

                    thead = table.thead
                    tbody = table.tbody

                    if thead is not None and tbody is not None:
                        if tab_name == 'SteeringComittee' and self.registry == 'Foundation':
                            trs = tbody.find_all('tr')
                            for tr in trs:
                                generic_html = "<table>{0}{1}</<table>"
                                generic_html = generic_html.format(thead, tr)
                                self.handle_dataframe_data(generic_html, None, ptype, pgroup, company_people_cols, c_people)

                        trs = tbody.find_all('tr', {'class': True})
                        limitations = dict()
                        for tr in trs:
                            generic_html = "<table>{0}{1}</<table>"
                            tr_class_name = tr['class'][0]
                            if tr_class_name == 'apr-master-row':
                                siblings = tr.next_siblings
                                for sibling in siblings:
                                    if type(sibling) is bs4.element.Tag:
                                        check_class = sibling.get('class')
                                        if check_class is not None:
                                            if check_class[0] == 'apr-detail-row':
                                                limitations = tables(sibling.decode(), self.registry_number, company_people_cols)

                                generic_html= generic_html.format(thead, tr)
                                self.handle_dataframe_data(generic_html, limitations, ptype, pgroup, company_people_cols, c_people)

                        # print("------" * 20)

            return 200
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
            return error


class Decisions:
    def __init__(self, registry_number, html, registry):
        self.registry_number = registry_number
        self.registry = registry
        self.html = html
        self.data = list()
        self.sha_data = list()

    def parse(self):
        try:
            html = self.html
            if type(html) is bytes:
                html = html.decode()
            html = cirilic_to_latin(html)

            soup = BeautifulSoup(html, 'lxml')
            trs = soup.find_all('tr', {'class': 'apr-details-row'})
            col_rep = {'Broj predmeta': 'case_number', 'Vrsta akta': 'act_type', 'Datum objave': 'publication_date'}

            for tr in trs:

                thead = tr.thead
                tbody = tr.tbody

                if thead is not None and tbody is not None:
                    trs = tbody.find_all('tr')
                    for tr in trs:
                        generic_html = "<table>{0}{1}</<table>"
                        generic_html = generic_html.format(thead, tr)
                        df = pd.read_html(generic_html)
                        if df:
                            df = df[0]
                            if df.empty is False:
                                df.columns = [col_rep.get(c) for c in df.columns]
                                df.dropna(axis=0, how='all', inplace=True)
                                doc_url = tr.find('a', {'href': True})
                                if doc_url is not None:
                                    doc_url = doc_url['href'].replace('&amp;', "&")
                                df['doc_link'] = [doc_url]*len(df)
                                df['registry_number'] = [self.registry_number] * len(df)
                                if None in df.columns:
                                    df.drop(columns=[None], inplace=True)
                                if 'publication_date' in df.columns:
                                    df['publication_date2'] = [date_convert(c) for c in df.publication_date]

                                rows = df.to_dict('records')
                                for row in rows:
                                    sha_data = [row.get('case_number'), row.get('act_type'), row.get('publication_date'), self.registry_number]
                                    sha = hashlib.sha3_256(str(sha_data).encode()).hexdigest()
                                    row.update({'sha': sha})
                                    if 'publication_date' in list(row.keys()) and 'publication_date2' in list(row.keys()):
                                        publication_date2 = row.get('publication_date2')
                                        row.update({'publication_date': publication_date2})
                                        row.pop('publication_date2')

                                    if sha not in self.sha_data:
                                        self.data.append(row)
                                        self.sha_data.append(sha)

            return 200

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            error = {'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e), 'status': 2}
            print(error)
            return error


# regnum= '07559470'
#
# html_path = r'E:/WebDocs/apr/companies/html/07559470/20200219/Decrees_Enterprise.html'
#
# with open(html_path, 'rb') as f:
#     html = f.read()
#     decisions_parser = Decisions(regnum, html, 'Enterprise')
#     decisions_parser.parse()
#     for row in decisions_parser.data:
#         print(row)

