import os
import sys
import pickle
import getpass

from sqlalchemy import *
from subprocess import call
from sqlalchemy import event
from subprocess import Popen, PIPE
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base

if sys.version_info.major > 2:
    from core_framework.db_engine import DbEngine

schema = 'monitoring'

framework_folder = r'C:\Users\{}\Documents\crawler_framework'.format(getpass.getuser())
python_path = r'{}\py_config.pkl'.format(framework_folder)

BasePstg = declarative_base()

# creating schema if it doesn't exist
event.listen(BasePstg.metadata, 'before_create', DDL("CREATE SCHEMA IF NOT EXISTS {}".format(schema)))

# ======================================================================
#                               TABLES
# ======================================================================


class RsCompanies(BasePstg):
    """Current Serbian companies in monitoring by our clients"""
    __tablename__ = 'rs_companies'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('rs_com_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    registry_number = Column(String(128), index=True)  # registry number of company being monitored
    archive = Column(DateTime)  # date and time when record was closed
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


# ======================================================================
#                               LOGGERS
# ======================================================================

class CrawlerInfoLog(BasePstg):
    __tablename__ = 'l_info'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('loginfo_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    runtime_id = Column(String(64), index=True)  # hash value of single run cycle of program
    sha = Column(String(64), index=True)  # hash value of the row
    start_time = Column(DateTime)
    end_time = Column(DateTime)
    step_name = Column(String(256))
    status = Column(Integer() ) # 1 = success  2=errors
    errors = Column(Integer())  # >0 there was and error
    note = Column(String(1024))
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


class CrawlerErrorLog(BasePstg):
    __tablename__ = 'l_errors'
    __table_args__ = {'schema': schema}

    id_seq = Sequence('logerr_id_seq', metadata=BasePstg.metadata, schema=schema)
    id = Column(
        Integer, id_seq,
        server_default=id_seq.next_value(), primary_key=True)

    runtime_id = Column(String(64), index=True)  # hash value of single run cycle of program
    sha = Column(String(64), index=True)  # hash value of the row
    start_time = Column(DateTime)
    end_time = Column(DateTime)
    step_name = Column(String(256))
    err_type = Column(String(3072))
    err_line = Column(Integer())
    err_desc = Column(String(3072))
    date_created = Column(DateTime, server_default=func.now())  # date and time when record was created


def deploy(conn_id):
    db = DbEngine()
    engine = db.connect(conn_id, echo=False)
    BasePstg.metadata.drop_all(engine)
    BasePstg.metadata.create_all(engine)


if __name__ == '__main__':
    deploy(1)