import sys
import hashlib
from abc import ABC
from datetime import datetime
from core_framework.crawlers import CrawlerBase

base = 'monitoring'
schema = 'monitoring'


class Monitoring(CrawlerBase, ABC):
    def __init__(self, conn=1):
        CrawlerBase.__init__(self, base)

        # runtime log initialization
        self.runtime_errors = 0
        self.runtime_status = 1
        self.runtime_start = datetime.now()
        self.runtime_id = hashlib.sha3_256(str(self.runtime_start).encode()).hexdigest()
        self.runtime_log = {'start_time': self.runtime_start, 'step_name': 'runtime'}
        self.conn = conn

    def info_log(self, data):
        print("info_log", data)
        self.connect(self.conn, connect_args={"application_name": "serbia/monitoring/update.py"})
        sha = hashlib.sha3_256(str(data).encode()).hexdigest()
        data.update({'sha': sha, 'runtime_id': self.runtime_id, 'end_time': datetime.now()})
        self.insert('l_info', {0: data}, schema=schema)

    def error_log(self, data):
        self.connect(self.conn, connect_args={"application_name": "serbia/monitoring/update.py"})
        print("error_log", data)
        self.runtime_status = 2
        self.runtime_errors += 1
        sha = hashlib.sha3_256(str(data).encode()).hexdigest()
        data.update({'sha': sha, 'runtime_id': self.runtime_id, 'end_time': datetime.now()})
        self.insert('l_errors', {0: data}, schema=schema)

    def get_monitoring_rs(self):
        start_time = datetime.now()
        step_name = 'get_monitoring_rs'
        print(f"> {step_name}")
        log_common = {'start_time': start_time, 'step_name': step_name}

        errors = 0
        try:
            conn_status = self.connect(5)
            if conn_status != 400:
                monitoring_data = self.select('monitoring_active', view=True)
                merge_data = {}
                for record in monitoring_data:
                    merge_data.update({len(merge_data): record})
                if merge_data:
                    self.connect(self.conn, connect_args={"application_name": "serbia/monitoring/update.py"})
                    merge_status = self.merge('rs_companies', merge_data, filters={'archive': None}, on=['registry_number'],schema=schema)
                    if merge_status != 200:
                        errors += 1
                        error = "error occured during merge"
                        log_copy = log_common.copy()
                        log_copy.update({'err_desc': error})
                        self.error_log(log_copy)
            else:
                errors += 1
                error = 'Could not connect to rs database'
                log_copy = log_common.copy()
                log_copy.update({'err_desc': error})
                self.error_log(log_copy)

        except Exception as e:
            errors += 1
            exc_type, exc_obj, exc_tb = sys.exc_info()
            copy_log = log_common.copy()
            copy_log.update({'err_type': str(exc_type), 'err_line': exc_tb.tb_lineno, 'err_desc': str(e)})
            self.error_log(copy_log)

        status = 1
        if errors > 0:
            status = 2

        log_common.update({'status': status, 'errors': errors})
        self.info_log(log_common)


def update_monitoring_rs(conn=1):
    api = Monitoring(conn=conn)
    api.get_monitoring_rs()

    # final runtime log
    api.runtime_log.update({'status': api.runtime_status, 'errors': api.runtime_errors})
    api.info_log(api.runtime_log)
    api.disconnect()


if __name__ == '__main__':
    update_monitoring_rs()