import re
import hashlib
import pandas as pd
from datetime import datetime
from sys import stdout
import concurrent.futures
from core_framework.db_engine import DbEngine


class DbConnection(DbEngine):
    def __init__(self):
        DbEngine.primary_key = 'id'
        DbEngine.__init__(self)
        self.connect(2)


class PostgresConnection(DbEngine):
    def __init__(self):
        DbEngine.primary_key = 'id'
        DbEngine.__init__(self)
        self.connect(1)


class DbAprConnection(DbEngine):
    def __init__(self):
        DbEngine.primary_key = 'id'
        DbEngine.__init__(self)
        self.connect(3)


def unique_status(status_eval):
    if status_eval is None or status_eval == '':
        return 'Nepoznato'
    rex = {'aktiv': 'Aktivan', 'bris': 'Brisan', 'prin.*?likvi': 'Prinudna likvidacija',
           'proc.*?likvi': 'U procesu likvidacije', 'likvi': 'U likvidaciji',
           'ste.*?post': 'U stečajnom postupku', 'ste.a': 'U stečaju'}
    for k, v in rex.items():
        if re.search(k, status_eval, re.IGNORECASE) is not None:
            status_eval = v
            break
    return status_eval


def row_sha(data, cols=None, dates=None):
    """
    :type data: dict
    :type cols: list
    :type dates: list
    """
    copy_data = data.copy()
    # #date conversion
    # if dates is not None:
    #     for k, v in copy_data.items():
    #         if k in dates:
    #             print(k, v, type(v))
    #             if type(v) is not str and v is not None:
    #                 copy_data.update({k: v})
    #                 exit()

    if cols is not None:
        copy_data = {k: str(v).lower() for k,v in copy_data.items() if k in cols and v is not None}

    # replacing nan and Nat
    for k, v in copy_data.items():
        if str(v).lower() in ('nan', 'nat'):
            copy_data.update({k: str(None)})

    values = ''.join(list(copy_data.values()))
    values = re.sub(r'[\s+]','',values, re.DOTALL|re.UNICODE)
    # values = ''.join(sorted(values))
    sha = hashlib.sha3_256(values.encode()).hexdigest()
    return sha


db = DbConnection()
db_apr = DbAprConnection()
pstgrs = PostgresConnection()


class SearchInfoFill:
    def __init__(self):
        # SHA columns
        self.search_info = ['registry_number', 'ctype', 'cname', 'cstatus']

    def update_sha_apr_subjekti(self):
        # populate search_info_sha that will be same as sha in [APR_SRBIJA].[dbo].[SearchInfo]
        # UPDATE SHA APR_SUBJEKTI
        print('updating search_info_sha in table APR_SUBJEKTI')
        sql = '''select distinct id , MB registry_number, PRAVNA_FORMA ctype, case when NAZIV is null then NAZIV_KRATKI else NAZIV end cname, STATUS cstatus, DAT_PRO archive, DATE_CREATED date_created, DAT_CHECK  date_checked
        from Bisnode_RS.dbo.APR_SUBJEKTI where search_info_sha is null order by id'''
        data = db.select('APR_SUBJEKTI', sql=sql)
        print("total", len(data))

        # # for first update we created tmp table with cols id and sha populate that data and
        # # used that data to update search_info_sha in table APR_SUBJEKTI
        # # query for update is
        # sha_data_id = {}
        # for row in data:
        #     row_sha_ready = row_sha(row, self.search_info)
        #     id = row.get('id')
        #     sha_data_id.update({len(sha_data_id): {'id': id, 'sha': row_sha_ready}})
        # merge_status = db.merge('tmp_search_info_sha_ids', sha_data_id,filters={}, update=False)
        # print(merge_status)

        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            # Start the load operations and mark each future with its URL
            future_to_url = {executor.submit(self.update_sha_apr_subjekti_thread, row): row for row in data}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    status = future.result()
                except AttributeError as e:
                    print(str(e))
                except Exception as e:
                    print(str(e))

    def update_sha_apr_subjekti_thread(self, row):
        row_sha_ready = row_sha(row, self.search_info)
        id = row.get('id')
        x = db.update('APR_SUBJEKTI', {'ID': id, 'search_info_sha': None}, {'search_info_sha': row_sha_ready}, freeze=True, primary_key='ID')
        # print('Update status', x, row)
        return 200

    def update_search_sha_ms(self):
        print('Update sha in SearchInfo that is in APR_SRBIJA db')
        sql = '''select id, RegistryNumber registry_number, [TYPE] ctype, [Name] cname, [Status] cstatus from SearchInfo where sha is null'''
        data = db_apr.select('SearchInfo', sql=sql)

        with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
            # Start the load operations and mark each future with its URL
            future_to_url = {executor.submit(self.update_search_sha_ms_thread, row): row for row in data}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    status = future.result()
                except AttributeError as e:
                    pass
                except Exception as e:
                    print(str(e))

    def update_search_sha_ms_thread(self, row):
        row_sha_ready = row_sha(row, self.search_info)
        id = row.get('id')
        x = db_apr.update('SearchInfo', {'id': id, 'sha': None}, {'sha': row_sha_ready}, freeze=True)
        print('Update status', x, row)
        row.update({'sha': row_sha_ready})

    def archive_data_search_info(self):
        print('updating archive dates')
        # update rows that are older than last record for single registry number
        sql = '''select id, RegistryNumber registry_number, [TYPE] ctype, [Name] cname, [Status] cstatus, DateCreated date_created
         from SearchInfo a where a.Archive is null 
         and a.DateCreated >= getdate()-1 
         order by RegistryNumber, DateCreated desc'''
        data = db_apr.select('SearchInfo', sql=sql)
        sorted_data = {}
        for row in data:
            registry_number = row.get('registry_number')
            date_created = row.get('date_created')

            get_reg_data = sorted_data.get(registry_number)
            if get_reg_data is None:
                sorted_data.update({registry_number: {0: row}})
            else:
                get_reg_data.update({len(get_reg_data): row})
        cnt = 0
        for k, v in sorted_data.items():
            total_rows = len(v)
            if total_rows > 1:
                for i in range(1, total_rows):
                    previous_row_data = v.get(i-1)
                    row_data = v.get(i)
                    row_id = row_data.get('id')
                    archive_date = previous_row_data.get('date_created')
                    upd_stat = db_apr.update('SearchInfo', {'id': row_id}, {'Archive': archive_date}, freeze=True)
                    print(cnt, upd_stat, row_data)
            cnt += 1

    def first_transfer_apr_subjekti(self):
        # INSERT DATA TO POSTGRES SearchInfo FROM APR_SUBJEKTI
        print('insert into postgres search_info FROM APR_SUBJEKTI')
        check_data = pstgrs.select('search_info', schema='apr')
        if not check_data:
            # sql = '''select * from v_toSearchInfo a
            #          where not exists (select 1 from APR_SRBIJA.dbo.SearchInfo b where a.sha = b.sha and a.registry_number = b.RegistryNumber)
            #          order by date_created'''
            sql = '''
            select  MB registry_number, PRAVNA_FORMA ctype, case when NAZIV is null then NAZIV_KRATKI else NAZIV end cname, STATUS cstatus, DAT_PRO archive, DATE_CREATED date_created, DAT_CHECK  date_checked, search_info_sha sha  
            from  Bisnode_RS.dbo.APR_SUBJEKTI where DAT_PRO is null and MB not in (select RegistryNumber from APR_SRBIJA.dbo.SearchInfo) and search_info_sha is not null
            '''
            data = db.select('APR_SUBJEKTI', sql=sql)
            for row in data:
                get_status = row.get('cstatus')
                row_sha_ready = row_sha(row, self.search_info)
                row.update({'status_group': unique_status(get_status), 'sha': row_sha_ready})

            merge_data = dict(enumerate(data))
            pstgrs.merge('search_info', merge_data, filters={}, schema='apr')

    def mix_new_data(self):
        "combine new data with old data"
        print('insert into postgres search_info new data from SearchInfo')
        # regnums_sync = [x.get('registrynumber') for x in db_apr.select('SearchInfo', sql="select distinct RegistryNumber from SearchInfo where sync is null and RegistryNumber='06000037' ")]
        regnums_sync = [x.get('registrynumber') for x in db_apr.select('SearchInfo', sql="select distinct RegistryNumber from SearchInfo where sync is null ")]
        # compare only all records SearchInfo mssql with all records in search_info postgres
        print("compare only all records SearchInfo mssql with all records in search_info postgres")
        cnt = 0
        for regnum in regnums_sync:
            cnt += 1
            stdout.write(f'\r>progress: {cnt}/{len(regnums_sync)}')
            sql = f'''select RegistryNumber registry_number, [TYPE] ctype, [Name] cname, [Status] cstatus, Registry registry, DateCreated date_created, Archive archive, sha , DateChecked date_checked  
            from SearchInfo where RegistryNumber = '{regnum}' '''
            regnum_data = db_apr.select('SearchInfo', sql=sql)
            for regnum_d in regnum_data:
                get_status = regnum_d.get('cstatus')
                regnum_d.update({'status_group': unique_status(get_status)})

            regnum_data = dict(enumerate(regnum_data))
            merge_status = pstgrs.merge('search_info',regnum_data, filters={'registry_number': regnum}, schema='apr', on=['sha'], update=False)

        # get all registry numbers that have more than one record open in postgres now and add archive date
        print("\nget all registry numbers that have more than one record open in postgres now and add archive date")
        sql = '''select registry_number from (
        select registry_number, count(*) from apr.search_info where archive is null group by registry_number having count(*) > 1) a'''
        regnums_archive = [x.get('registry_number') for x in pstgrs.select('search_info', sql=sql, schema='apr')]
        cnt = 0
        for regnum in regnums_archive:
            cnt += 1
            stdout.write(f'\r>progress: {cnt}/{len(regnums_archive)}')
            sql = f"select * from apr.search_info where registry_number= '{regnum}' order by date_created desc"
            # print(sql)
            regnum_data = pstgrs.select('search_info', sql=sql, schema='apr')
            sorted_data = {}
            for row in regnum_data:
                registry_number = row.get('registry_number')
                date_created = row.get('date_created')
                get_reg_data = sorted_data.get(registry_number)
                if get_reg_data is None:
                    sorted_data.update({registry_number: {0: row}})
                else:
                    get_reg_data.update({len(get_reg_data): row})

            for k, v in sorted_data.items():
                total_rows = len(v)
                if total_rows > 1:
                    for i in range(1, total_rows):
                        previous_row_data = v.get(i - 1)
                        row_data = v.get(i)
                        row_id = row_data.get('id')
                        archive_date = previous_row_data.get('date_created')
                        # print(row_id, archive_date)
                        upd_stat = pstgrs.update('search_info', {'id': row_id}, {'archive': archive_date}, schema='apr')
                        # print(cnt, upd_stat, row_data)

        # compare only active records SearchInfo mssql with active records in search_info postgres
        print('\ncompare only active records SearchInfo mssql with active records in search_info postgres')
        cnt = 0
        for regnum in regnums_sync:
            cnt += 1
            stdout.write(f'\r>progress: {cnt}/{len(regnums_sync)}')
            sql = f'''select RegistryNumber registry_number, [TYPE] ctype, [Name] cname, [Status] cstatus, Registry registry,  DateCreated date_created, Archive archive, sha, DateChecked date_checked   
            from SearchInfo where RegistryNumber = '{regnum}' and archive is null'''
            regnum_data = db_apr.select('SearchInfo', sql=sql)
            for regnum_d in regnum_data:
                get_status = regnum_d.get('cstatus')
                regnum_d.update({'status_group': unique_status(get_status)})

            regnum_data = dict(enumerate(regnum_data))
            merge_status = pstgrs.merge('search_info',regnum_data, filters={'registry_number': regnum, 'archive': None}, schema='apr', on=['sha'])
            if merge_status == 200:
                db_apr.update('SearchInfo', {'RegistryNumber': regnum, 'sync': None}, {'sync': 1}, freeze=True)


class BasicInfoFill:
    def __init__(self):
        # SHA columns
        self.basic_info = ['registry_number', 'basic_name', 'cname', 'cstatus', 'status_group', 'legal_form', 'foun_date', 'erase_date', 'activity_start', 'activity_end']
        self.cnt, self.total = 0, 0

    def update_sha_apr_subjekti(self):
        # populate search_info_sha that will be same as sha in [APR_SRBIJA].[dbo].[SearchInfo]
        # UPDATE SHA APR_SUBJEKTI
        print('updating basic_info_sha in table APR_SUBJEKTI')
        sql = '''select distinct id , MB registry_number, NAZIV_KRATKI basic_name, NAZIV cname, STATUS cstatus, PRAVNA_FORMA legal_form,
        DAT_OSN foun_date, DAT_BRISANJA erase_date, DAT_PRESTANKA_DJEL activity_end
        from Bisnode_RS.dbo.APR_SUBJEKTI where basic_info_sha is null order by id'''
        data = db.select('APR_SUBJEKTI', sql=sql)

        # for first update we created tmp table with cols id and sha populate that data and
        # used that data to update search_info_sha in table APR_SUBJEKTI
        # query for update is
        sha_data_id = {}
        print(f"total size: {len(data)}")
        self.cnt = 0
        self.total = len(data)

        # for row in data:
        #     stdout.write(f"\rprocessing: {self.cnt}/{self.total}")
        #     row_sha_ready = row_sha(row, self.basic_info)
        #     id = row.get('id')
        #     sha_data_id.update({len(sha_data_id): {'id': id, 'sha': row_sha_ready}})
        #     self.cnt += 1
        # print("\ninserting data")
        # merge_status = db.insert('tmp_basic_info_sha_ids', sha_data_id)
        # print(merge_status)

        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            # Start the load operations and mark each future with its URL
            future_to_url = {executor.submit(self.update_sha_apr_subjekti_thread, row): row for row in data}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    status = future.result()
                except AttributeError as e:
                    print(str(e))
                except Exception as e:
                    print(str(e))

    def update_sha_apr_subjekti_thread(self, row):
        self.cnt += 1
        stdout.write(f"\r progress: {self.cnt}/{self.total}")
        row_sha_ready = row_sha(row, self.basic_info)
        id = row.get('id')
        x = db.update('APR_SUBJEKTI', {'ID': id, 'basic_info_sha': None}, {'basic_info_sha': row_sha_ready}, freeze=True, primary_key='ID')
        # print('Update status', x, row)
        return 200

    def update_basic_sha_ms(self):
        print('update sha in BasicInfo that is in APR_SRBIJA db')
        sql = '''select id, RegistryNumber registry_number, BasicName basic_name, CompanyName cname, [Status] cstatus, 
        LegalForm legal_form, FounDate foun_date, EraseDate erase_date, ActivityStartDate activity_start, ActivityEndDate activity_end 
        from BasicInfo where sha is null'''
        data = db_apr.select('BasicInfo', sql=sql)

        self.cnt = 0
        self.total = len(data)

        with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
            # Start the load operations and mark each future with its URL
            future_to_url = {executor.submit(self.update_search_sha_ms_thread, row): row for row in data}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    status = future.result()
                except AttributeError as e:
                    pass
                except Exception as e:
                    print(str(e))

    def update_search_sha_ms_thread(self, row):
        row_sha_ready = row_sha(row, self.basic_info)
        id = row.get('id')
        x = db_apr.update('BasicInfo', {'id': id, 'sha': None}, {'sha': row_sha_ready}, freeze=True)
        self.cnt += 1
        stdout.write(f"\r>progress: {self.cnt}/{self.total}")
        row.update({'sha': row_sha_ready})

    def archive_data_basic_info(self):
        print('\nupdating archive dates')
        # update rows that are older than last record for single registry number

        # check last 2 days
        sql = '''select id, RegistryNumber registry_number, BasicName basic_name, CompanyName cname, [Status] cstatus,
        LegalForm legal_form, FounDate foun_date, EraseDate erase_date, ActivityStartDate activity_start,
        ActivityEndDate activity_end, DateCreated date_created from BasicInfo a where a.Archive is null
        and a.DateCreated >= getdate()-1 order by RegistryNumber, DateCreated desc'''

        # # check all
        # sql = '''select id, RegistryNumber registry_number, BasicName basic_name, CompanyName cname, [Status] cstatus,
        # LegalForm legal_form, FounDate foun_date, EraseDate erase_date, ActivityStartDate activity_start,
        # ActivityEndDate activity_end, DateCreated date_created from BasicInfo a where a.Archive is null
        # order by RegistryNumber, DateCreated desc'''

        data = db_apr.select('BasicInfo', sql=sql)
        sorted_data = {}
        for row in data:
            registry_number = row.get('registry_number')
            date_created = row.get('date_created')

            get_reg_data = sorted_data.get(registry_number)
            if get_reg_data is None:
                sorted_data.update({registry_number: {0: row}})
            else:
                get_reg_data.update({len(get_reg_data): row})

        cnt = 0
        for k, v in sorted_data.items():
            cnt += 1
            stdout.write(f'\r> progress: {cnt}/{len(sorted_data)}')
            total_rows = len(v)
            if total_rows > 1:
                for i in range(1, total_rows):
                    previous_row_data = v.get(i-1)
                    row_data = v.get(i)
                    row_id = row_data.get('id')
                    archive_date = previous_row_data.get('date_created')
                    upd_stat = db_apr.update('BasicInfo', {'id': row_id}, {'archive': archive_date}, freeze=True)
                    print("\n", cnt, upd_stat, row_data)

    def first_transfer_apr_subjekti(self):
        # INSERT DATA TO POSTGRES BasicInfo FROM APR_SUBJEKTI
        print('insert into postgres basic_info FROM APR_SUBJEKTI')
        check_data = pstgrs.select('basic_info', schema='apr')
        if not check_data:
            sql = '''select * from v_toBasicInfo a
                     where not exists (select 1 from APR_SRBIJA.dbo.BasicInfo b where a.sha = b.sha and a.registry_number = b.RegistryNumber)
                     order by date_created'''

            sql = '''
            select MB registry_number, NAZIV_KRATKI basic_name, NAZIV cname, [STATUS] cstatus, PRAVNA_FORMA legal_form, DAT_OSN foun_date, 
            DAT_BRISANJA erase_date, DAT_PRESTANKA_DJEL activity_end, DAT_PRO archive, DATE_CREATED date_created, basic_info_sha sha 
            from  Bisnode_RS.dbo.APR_SUBJEKTI where basic_info_sha is not null and dat_pro is null
            and MB not in (select RegistryNumber from APR_SRBIJA.dbo.BasicInfo)
            '''

            data = db.select('APR_SUBJEKTI', sql=sql)
            for row in data:
                get_status = row.get('cstatus')
                row_sha_ready = row_sha(row, self.basic_info)
                row.update({'status_group': unique_status(get_status), 'sha': row_sha_ready})

            merge_data = dict(enumerate(data))
            pstgrs.merge('basic_info', merge_data, filters={}, schema='apr')

    def mix_new_data(self):
        "combine new data with old data"
        print('insert into postgres basic_info new data from BasicInfo')
        # regnums_sync = [x.get('registrynumber') for x in db_apr.select('BasicInfo', sql="select distinct RegistryNumber from BasicInfo where sync is null and RegistryNumber='06000037' ")]
        regnums_sync = [x.get('registrynumber') for x in db_apr.select('BasicInfo', sql="select distinct RegistryNumber from BasicInfo where sync is null ")]
        # compare only all records BasicInfo mssql with all records in basic_info postgres
        print("compare only all records BasicInfo mssql with all records in basic_info postgres")
        cnt = 0
        for regnum in regnums_sync:
            cnt += 1
            stdout.write(f'\r>progress: {cnt}/{len(regnums_sync)}')

            sql = f'''
            select RegistryNumber registry_number, BasicName basic_name, CompanyName cname, [Status] cstatus,
            LegalForm legal_form, FounDate foun_date, EraseDate erase_date, ActivityStartDate activity_start,
            ActivityEndDate activity_end, DateCreated date_created, Archive archive, sha  from BasicInfo a 
            where RegistryNumber = '{regnum}'
            '''
            regnum_data = db_apr.select('BasicInfo', sql=sql)
            for regnum_d in regnum_data:
                get_status = regnum_d.get('cstatus')
                regnum_d.update({'status_group': unique_status(get_status)})

            regnum_data = dict(enumerate(regnum_data))
            merge_status = pstgrs.merge('basic_info',regnum_data, filters={'registry_number': regnum}, schema='apr', on=['sha'], update=False)

        # get all registry numbers that have more than one record open in postgres now and add archive date
        print("\nget all registry numbers that have more than one record open in postgres now and add archive date")
        sql = '''select registry_number from (
        select registry_number, count(*) from apr.basic_info where archive is null group by registry_number having count(*) > 1) a'''
        regnums_archive = [x.get('registry_number') for x in pstgrs.select('basic_info', sql=sql, schema='apr')]
        cnt = 0
        for regnum in regnums_archive:
            cnt += 1
            stdout.write(f'\r>progress: {cnt}/{len(regnums_archive)}')
            sql = f"select * from apr.basic_info where registry_number= '{regnum}' order by date_created desc"
            # print(sql)
            regnum_data = pstgrs.select('basic_info', sql=sql, schema='apr')
            sorted_data = {}
            for row in regnum_data:
                registry_number = row.get('registry_number')
                date_created = row.get('date_created')
                get_reg_data = sorted_data.get(registry_number)
                if get_reg_data is None:
                    sorted_data.update({registry_number: {0: row}})
                else:
                    get_reg_data.update({len(get_reg_data): row})

            for k, v in sorted_data.items():
                total_rows = len(v)
                if total_rows > 1:
                    for i in range(1, total_rows):
                        previous_row_data = v.get(i - 1)
                        row_data = v.get(i)
                        row_id = row_data.get('id')
                        archive_date = previous_row_data.get('date_created')
                        # print(row_id, archive_date)
                        upd_stat = pstgrs.update('basic_info', {'id': row_id}, {'archive': archive_date}, schema='apr')
                        # print(cnt, upd_stat, row_data)

        # compare only active records BasicInfo mssql with active records in basic_info postgres
        print('\ncompare only active records BasicInfo mssql with active records in basic_info postgres')
        cnt = 0
        for regnum in regnums_sync:
            cnt += 1
            stdout.write(f'\r>progress: {cnt}/{len(regnums_sync)}')
            sql = f'''
            select RegistryNumber registry_number, BasicName basic_name, CompanyName cname, [Status] cstatus,
            LegalForm legal_form, FounDate foun_date, EraseDate erase_date, ActivityStartDate activity_start,
            ActivityEndDate activity_end, DateCreated date_created, Archive archive, sha  from BasicInfo a 
            where RegistryNumber = '{regnum}' and archive is null
            '''
            regnum_data = db_apr.select('BasicInfo', sql=sql)
            for regnum_d in regnum_data:
                get_status = regnum_d.get('cstatus')
                regnum_d.update({'status_group': unique_status(get_status)})

            regnum_data = dict(enumerate(regnum_data))
            merge_status = pstgrs.merge('basic_info',regnum_data, filters={'registry_number': regnum, 'archive': None}, schema='apr', on=['sha'])
            if merge_status == 200:
                db_apr.update('BasicInfo', {'RegistryNumber': regnum, 'sync': None}, {'sync': 1}, freeze=True)


class ActivityPause:
    def __init__(self):
        # SHA columns
        self.activity_cols = ['registry_number', 'pause_start', 'pause_end']
        self.cnt, self.total = 0, 0

    def add_history_records(self):
        print("Adding records from Bisnode_RS.dbo.APR_SUBJEKTI that doesnt exist in ActivityPause")
        sql = '''
        select distinct MB RegistryNumber, DAT_OD_PP PauseStartDate, DAT_DO_PP PauseEndDate, DATE_CREATED DateCreated from Bisnode_RS.dbo.APR_SUBJEKTI a
        where DAT_OD_PP is not null
        and not exists (select 1 from ActivityPause b where a.mb = b.RegistryNumber 
        and isnull(a.DAT_OD_PP, '1900-01-01') = isnull(b.PauseStartDate, '1900-01-01')) 
        '''
        data = db_apr.select('APR_SUBJEKTI', sql=sql, freeze=True)
        if data:
            ins_data = dict(enumerate(data))
            print("inserting")
            insert_status = db_apr.insert('ActivityPause', ins_data)
            print(f"insert status: {insert_status}")

    def update_sha_activity(self):
        print("Adding hash value in ActivityPause")
        sql = '''select id, RegistryNumber registry_number,  PauseStartDate pause_start, PauseEndDate pause_end from ActivityPause where sha is null'''
        data = db_apr.select('ActivityPause', sql=sql)
        for row in data:
            row_id = row.get('id')
            row_sha_ready = row_sha(row, self.activity_cols)
            update_status = db_apr.update('ActivityPause', {'id': row_id}, {'sha': row_sha_ready})
            if update_status != 200:
                print(f"error occured: {update_status}")
                exit()

    def close_pause_with_erase_date(self):
        print("CLosing activity pause periods for companies that are erased")
        sql = '''
        select * from (
        select ROW_NUMBER() over(partition by RegistryNumber order by datecreated desc) id,
        RegistryNumber, EraseDate from APR_SRBIJA.dbo.BasicInfo) a where id = 1 and EraseDate is not null
        '''
        data = db_apr.select('ActivityPause', sql=sql)

        end_dates = {row.get('registrynumber'): pd.to_datetime(row.get('erasedate'), yearfirst=True) for row in data}

        sql = '''
        SELECT registry_number, max(change_date) change_date FROM [APR_SRBIJA].[dbo].[CompanyChanges] 
        where CHANGE_id in (57, 81, 111, 1002, 1042, 1078)
        and regproc_type in (4, 14, 26, 44, 55, 60, 68, 74)
        group by registry_number
        '''
        data = db_apr.select('ActivityPause', sql=sql)
        change_dates = {row.get('registry_number'): row.get('change_date') for row in data}

        sql = '''  select * from APR_SRBIJA.[dbo].[ActivityPause] where PauseEndDate is null   '''
        data = db_apr.select('ActivityPause', sql=sql)

        cnt = 0
        for row in data:
            registry_number = row.get('registrynumber')
            pause_start_date_org = row.get('pausestartdate')
            pause_start_date = pd.to_datetime((pause_start_date_org), yearfirst=True, errors='coerce')

            end_date = end_dates.get(registry_number)
            change_date = change_dates.get(registry_number)
            if end_date is not None and type(pause_start_date) is pd._libs.tslibs.timestamps.Timestamp:
                if pause_start_date <= end_date:
                    cnt += 1
                    print(cnt, registry_number, end_date, pause_start_date)
                    update_status = db_apr.update('ActivityPause', {'RegistryNumber': registry_number, 'PauseStartDate': pause_start_date_org, 'PauseEndDate': None}, {'PauseEndDate': end_date}, freeze=True)

            if change_date is not None and type(pause_start_date) is pd._libs.tslibs.timestamps.Timestamp:
                if pause_start_date <= change_date:
                    cnt += 1
                    print(cnt, registry_number, change_date, pause_start_date)
                    update_status = db_apr.update('ActivityPause', {'RegistryNumber': registry_number, 'PauseStartDate': pause_start_date_org, 'PauseEndDate': None}, {'PauseEndDate': end_date}, freeze=True)


class AddressData:
    # AddressData is already combined in APR_SRBIJA only archive, and sha colums need to be populated
    def __init__(self):
        # SHA columns
        self.address_data = ['registry_number', 'municipality', 'place', 'street', 'address_type', 'address_extra']

    def address_archive(self):
        print('updating archive dates')
        add_types = ['CompanyHq', 'CompanyMail']
        for add_type in add_types:
            # update rows that are older than last record for single registry number
            sql = f'''select id, RegistryNumber registry_number, DateCreated date_created  from AddressData a 
            where a.Archive is null and a.AddressType = '{add_type}'
            order by RegistryNumber, DateCreated desc'''

            data = db_apr.select('AddressData', sql=sql)
            sorted_data = {}
            for row in data:
                registry_number = row.get('registry_number')
                date_created = row.get('date_created')

                get_reg_data = sorted_data.get(registry_number)
                if get_reg_data is None:
                    sorted_data.update({registry_number: {0: row}})
                else:
                    get_reg_data.update({len(get_reg_data): row})
            cnt = 0
            for k, v in sorted_data.items():
                total_rows = len(v)
                if total_rows > 1:
                    for i in range(1, total_rows):
                        previous_row_data = v.get(i - 1)
                        row_data = v.get(i)
                        row_id = row_data.get('id')
                        archive_date = previous_row_data.get('date_created')
                        upd_stat = db_apr.update('AddressData', {'id': row_id}, {'archive': archive_date}, freeze=True)
                        print(cnt, upd_stat, row_data)
                cnt += 1

    def update_sha_address(self):
        print("Adding hash value in AddressData")
        sql = '''select id, RegistryNumber registry_number, Municipality municipality, place Place, street Street, AddressType address_type, AddressExtra address_extra
        from [APR_SRBIJA].[dbo].[AddressData] where sha is null '''
        data = db_apr.select('AddressData', sql=sql)
        cnt = 0
        for row in data:
            cnt += 1
            stdout.write(f'\rprogress: {cnt}/{len(data)}')
            row_id = row.get('id')
            row = {x: row.get(x) for x in self.address_data}
            row_sha_ready = row_sha(row, self.address_data)
            update_status = db_apr.update('AddressData', {'id': row_id}, {'sha': row_sha_ready})
            if update_status != 200:
                print(f"error occured: {update_status}")
                exit()

    def insert_new_rows(self):
        print("Inserting new data postgresql in address_data")
        sql = '''select id, RegistryNumber registry_number, Municipality municipality, place Place, street Street, AddressType address_type, 
        AddressExtra address_extra, DateCreated date_created, archive, sha
        from [APR_SRBIJA].[dbo].[AddressData] where sha is not null and sync is null'''
        data = db_apr.select('AddressData', sql=sql)

        # # prvo punjenje
        # sql = '''select RegistryNumber registry_number, Municipality municipality, place Place, street Street, AddressType address_type,
        #         AddressExtra address_extra, DateCreated date_created, archive, sha
        #         from [APR_SRBIJA].[dbo].[AddressData] where sha is not null and sync is null'''
        # data = db_apr.select('AddressData', sql=sql)
        # data = dict(enumerate(data))
        # insert_status = pstgrs.insert('address_data', data, schema='apr')
        # print(insert_status)
        # exit()

        cnt = 0
        for row in data:
            cnt += 1
            stdout.write(f'\rprogress: {cnt}/{len(data)}')
            row_id = row.get('id')
            row_sha = row.get('sha')
            row.pop('id')
            merge_status = pstgrs.merge('address_data', {0: row}, filters={'sha': row_sha}, schema='apr')
            if merge_status == 200:
                update_status = db_apr.update('AddressData', {'id': row_id, 'sync': None}, {'sync': 1}, freeze=True)


class CleanInserts:
    # Inserts in postgre tables data that already has hash column and dont require additional preprocessing
    def insert_new_rows(self):
        print('inserting Annotations to annotations')
        sql = '''select id, RegistryNumber registry_number, sha, PostDate post_date, TextLines text_lines, Submitter submiter, Archive archive, DateCreated date_created 
        from Annotations
        where sha is not null and sync is null'''
        data = db_apr.select('Annotations', sql=sql)
        cnt = 0
        for row in data:
            cnt += 1
            stdout.write(f'\rprogress: {cnt}/{len(data)}')
            row_id = row.get('id')
            registry_number = row.get('registry_number')
            row.pop('id')
            merge_status = pstgrs.merge('annotations', {0: row}, filters={'registry_number': registry_number}, on=['sha'], schema='apr')
            if merge_status == 200:
                update_status = db_apr.update('Annotations', {'id': row_id, 'sync': None}, {'sync': 1}, freeze=True)

        print('insertinf Decisions into decisions')
        sql =''' select  id, RegistryNumber registry_number, CaseNumber case_number, ActType act_type, PublicationDate publication_date, sha_doc, DocDl doc_dl,
        ext doc_ext, upload doc_upl, temp_link doc_link  from Decisions  where sha is not null and sync is null '''
        data = db_apr.select('Decisions', sql=sql)
        cnt = 0
        for row in data:
            cnt += 1
            stdout.write(f'\rprogress: {cnt}/{len(data)}')
            row_id = row.get('id')
            registry_number = row.get('registry_number')
            row.pop('id')
            merge_status = pstgrs.merge('decisions', {0: row}, filters={'registry_number': registry_number}, on=['sha', 'sha_doc'], schema='apr')
            if merge_status == 200:
                update_status = db_apr.update('Decisions', {'id': row_id, 'sync': None}, {'sync': 1}, freeze=True)


class HashCreator:
    # calculates sha for mssql tables
    def __init__(self):
        self.AssociationExtra_cols = ['registry_number', 'statute_date', 'statute_change', 'activity_code2', 'activity_name2']
        self.bank_account_cols = ['registry_number', 'bank_number']
        self.bankruptcy_cols = ['registry_number', 'vat_number', 'bankruptcy_status', 'cregistry_number', 'cname', 'pname', 'psurname', 'pin',
                                'debtor_reg_num', 'debtor', 'creation_type', 'case_number', 'case_date', 'bankruptcy_value']
        self.branch_offices_cols = ['registry_number', 'branch', 'bname', 'activity_code', 'activity_name', 'municipality', 'place', 'street', 'address_extra', 'description']
        self.busniness_data_cols = ['registry_number', 'vat_number', 'rzzo', 'pio', 'activity_code', 'activity_name']
        self.business_name_cols = ['registry_number', 'cname', 'short_name', 'other_name', 'short_other_name']
        self.capital_cols = ['registry_number', 'share_group', 'capital_type', 'amount', 'currency', 'event_date', 'share_in']
        self.capital_amount_cols = ['registry_number', 'group_id', 'person_group_id', 'owner_type', 'share_group', 'capital_type', 'amount', 'currency', 'event_date', 'share_in']
        self.capital_groups_cols = ['registry_number', 'pin', 'passport_number', 'creg_num', 'group_id', 'person_group_id', 'owner_type']
        self.capital_shares_cols = ['registry_number', 'group_id', 'person_group_id', 'in_group_share', 'owner_type']
        self.company_people_cols = ['registry_number', 'pin', 'passport_num', 'passport_state', 'pname', 'pfunction', 'ptype', 'pgroup', 'represent', 'limitations', 'creg_num', 'cname', 'branch']
        self.email_cols = ['registry_number', 'email']
        self.erase_details_cols = ['registry_number', 'erase_date', 'activity_end_date', 'activity_end_reason']
        self.fax_cols = ['registry_number', 'fax_number']
        self.limited_duration_cols = ['registry_number', 'end_date']
        self.phone_cols = ['registry_number', 'phone_number']
        self.web_page_cols = ['registry_number', 'web_url']

    def assign_row_sha(self):
        tables_with_dates = {'Bankruptcy': ['case_date']}

        association_extra_sql = ''' select id, RegistryNumber registry_number, StatuteDate statute_date, StatuteDateChange statute_change,
         ActivityCode2 activity_code2, ActivityName2 activity_name2  from AssociationExtra where sha is null'''

        bank_accounts_sql = ''' select  id,  RegistryNumber registry_number, BankNumber bank_number from BankAccounts where sha is null '''

        bankruptcy_sql = ''' select  id, RegistryNumber registry_number, VatNumber vat_number, BankruptcyStatus bankruptcy_status, CompanyRegNum cregistry_number, 
        CompanyName cname, Name pname, Surname psurname, PIN pin, DebtorRegNum debtor_reg_num, Debtor debtor, CreationType creation_type,
        CaseNumber case_number, CaseDate case_date, BankruptcyValue bankruptcy_value  from Bankruptcy where sha is null '''

        branch_offices_sql = ''' select  id, RegistryNumber registry_number, Branch branch, Name bname, ActivityCode activity_code, ActivityName activity_name, 
        Municipality municipality, Place place, 
        Street street, AddressExtra address_extra, Description description from BranchOffices where sha is null '''

        busniness_data_sql = ''' select  id, RegistryNumber registry_number, VatNumber vat_number, RZZO rzzo, PIO pio, ActivityCode activity_code, ActivityName activity_name  
        from BusinessData where sha is null '''

        business_name_sql = '''select  id, RegistryNumber registry_number, CompanyName cname, ShortName short_name, CompanyNameOther other_name, ShortNameOther short_other_name  
        from BusinessName where sha is null'''

        capital_sql = ''' select  id, RegistryNumber registry_number, ShareGroup share_group, CapitalType capital_type, Amount amount, Currency currency, EventDate event_date, 
        ShareIn share_in from Capital where sha is null'''

        capital_amount_sql = ''' select  id, RegistryNumber registry_number, GroupId group_id, PersonGroupId person_group_id, OwnerType owner_type, ShareGroup share_group,
        CapitalType capital_type, Amount amount, Currency currency , EventDate event_date, ShareIn share_in  from CapitalAmounts where sha is null '''

        capital_groups_sql = ''' select   id, RegistryNumber registry_number, PIN pin, PassportNum passport_number, CompanyNumber creg_num, GroupId group_id,
        PersonGroupId person_group_id,  OwnerType owner_type from CapitalGroups where sha is null '''

        capital_shares_sql = ''' select id, RegistryNumber registry_number, GroupId group_id, PersonGroupId person_group_id, InGroupShare in_group_share, OwnerType owner_type
        from CapitalShares where sha is null '''

        company_people_sql = ''' select   id, RegistryNumber registry_number, PIN pin, PassportNum passport_num, PassportState passport_state, 
        Name pname, PersonFunction pfunction, PersonType ptype, PersonGroup pgroup, Represent represent, Limitations limitations,
         CompanyNumber creg_num, CompanyName cname, Branch branch from CompanyPeople  where sha is null '''

        email_sql = ''' select  id, RegistryNumber registry_number, Email email  from Email where sha is null  '''

        erase_details_sql = '''  select  id, RegistryNumber registry_number, EraseDate erase_date, ActivityEndDate activity_end_date, ActivityEndReason activity_end_reason 
        from EraseDetails where sha is null  '''

        fax_sql = ''' select  id, RegistryNumber registry_number,  FaxNumber fax_number from Fax where sha is null  '''
        limited_duration_sql = "select  id, RegistryNumber registry_number,  EndDate end_date from LimitedDuration where sha is null "
        phone_sql = "select  id, RegistryNumber registry_number,  PhoneNumber phone_number from Phone where sha is null "
        web_page_sql = "select  id, RegistryNumber registry_number, Url web_url  from WebPage where sha is null "


        gen_order = {'AssociationExtra': {'sql': association_extra_sql, 'cols': self.AssociationExtra_cols},
                     'BankAccounts': {'sql': bank_accounts_sql, 'cols': self.bank_account_cols},
                     'Bankruptcy': {'sql': bankruptcy_sql, 'cols': self.bankruptcy_cols},
                     'BranchOffices': {'sql': branch_offices_sql, 'cols': self.branch_offices_cols},
                     'BusinessData': {'sql': busniness_data_sql, 'cols': self.busniness_data_cols},
                     'BusinessName': {'sql': business_name_sql, 'cols': self.business_name_cols},
                     'Capital': {'sql': capital_sql, 'cols': self.capital_cols},
                     'CapitalAmounts': {'sql': capital_amount_sql, 'cols': self.capital_amount_cols},
                     'CapitalGroups': {'sql': capital_groups_sql, 'cols': self.capital_groups_cols},
                     'CapitalShares': {'sql': capital_shares_sql, 'cols': self.capital_shares_cols},
                     'CompanyPeople': {'sql': company_people_sql, 'cols': self.company_people_cols},
                     'Email': {'sql': email_sql, 'cols': self.email_cols},
                     'EraseDetails': {'sql': erase_details_sql, 'cols': self.erase_details_cols},
                     'Fax': {'sql': fax_sql, 'cols': self.fax_cols},
                     'LimitedDuration': {'sql': limited_duration_sql, 'cols': self.limited_duration_cols},
                     'Phone': {'sql': phone_sql, 'cols': self.phone_cols},
                     'WebPage': {'sql': web_page_sql, 'cols': self.web_page_cols},
                     }

        # association extra sha row assing
        for tablename, gen_data in gen_order.items():
            print(f"\nCalcluating sha for {tablename}")
            sql = gen_data.get('sql')
            cols = gen_data.get('cols')
            data = db_apr.select(tablename, sql=sql)
            cnt = 0
            for row in data:
                cnt += 1
                stdout.write(f'\rprogress: {cnt}/{len(data)}')
                row_id = row.get('id')
                if tablename in list(tables_with_dates.keys()):
                    date_columns = tables_with_dates.get(tablename)
                    for date_column in date_columns:
                        date_data = row.get(date_column)
                        lower_str = str(date_data).lower()
                        if date_data is not None and lower_str not in ('nan', 'nat'):
                            date_data = date_data.strftime("%Y-%m-%d %H:%M:%S")
                            row.update({date_column: date_data})
                row = {x: row.get(x) for x in cols}
                row_sha_ready = row_sha(row, cols)
                update_status = db_apr.update(tablename, {'id': row_id}, {'sha': row_sha_ready})
                if update_status != 200:
                    print(f"error occured: {update_status}")
                    exit()
            stdout.write(f'\rprogress: {cnt}/{len(data)}')

class ArchiveCalculator:
    # calculating arhive dates for those table who didnt have archive column
    # tables AssociationExtra, BusinessData, BusinessName, Email, EraseDetails, LimitedDuration, WebPage
    def calc_archive_date(self):
        tables = ['AssociationExtra', 'BusinessData', 'BusinessName', 'Email', 'EraseDetails', 'LimitedDuration', 'WebPage']
        for tablename in tables:
            # update rows that are older than last record for single registry number
            sql = f'''select id, RegistryNumber registry_number, DateCreated date_created  from {tablename} a 
            where a.archive is null
            order by RegistryNumber, DateCreated desc'''

            data = db_apr.select(tablename, sql=sql)
            sorted_data = {}
            for row in data:
                registry_number = row.get('registry_number')
                date_created = row.get('date_created')

                get_reg_data = sorted_data.get(registry_number)
                if get_reg_data is None:
                    sorted_data.update({registry_number: {0: row}})
                else:
                    get_reg_data.update({len(get_reg_data): row})
            cnt = 0
            for k, v in sorted_data.items():
                total_rows = len(v)
                if total_rows > 1:
                    for i in range(1, total_rows):
                        previous_row_data = v.get(i - 1)
                        row_data = v.get(i)
                        row_id = row_data.get('id')
                        archive_date = previous_row_data.get('date_created')
                        upd_stat = db_apr.update(tablename, {'id': row_id}, {'archive': archive_date}, freeze=True)
                        if upd_stat != 200:
                            print(tablename, cnt, upd_stat, row_data)
                            exit()
                        print(cnt, upd_stat, row_data)
                cnt += 1

class TransferData:
    # transfering data from mssql to postgresql
    def transferance(self):
        tablenames = ['']
        pass

# TODO: remove all duplicate hash values that are in final database known cases are search_info and basic_info tables


if __name__ == '__main__':
    # SYNC SearchInfo data
    # sif = SearchInfoFill()
    # sif.update_sha_apr_subjekti()
    # sif.update_search_sha_ms()
    # sif.archive_data_search_info()
    # sif.first_transfer_apr_subjekti()
    # sif.mix_new_data()

    # SYNC BasicInfo data
    # bif = BasicInfoFill()
    # bif.update_sha_apr_subjekti()
    # bif.update_basic_sha_ms()
    # bif.archive_data_basic_info()
    # bif.first_transfer_apr_subjekti()
    # bif.mix_new_data()

    # SYNC ActivityPause data
    # ap = ActivityPause()
    # ap.add_history_records()
    # ap.update_sha_activity()
    # ap.close_pause_with_erase_date()

    # SYNC AddressData data
    ad = AddressData()
    ad.address_archive()
    ad.update_sha_address()
    ad.insert_new_rows()

    hc = HashCreator()
    hc.assign_row_sha()

    # ac = ArchiveCalculator()
    # ac.calc_archive_date()

    # SYNC clean inserts  primjer za prebacivanje podataka
    # ci = CleanInserts()
    # ci.insert_new_rows()
