import os
import re
import sys
import random
import urllib3
import hashlib
import threading
from abc import ABC
from time import sleep
from sys import stdout
import concurrent.futures
from datetime import datetime
from urllib.parse import unquote
from core_framework.crawlers import CrawlerBase
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

base = 'https://e-oglasna.pravosudje.hr'

main_doc_path = 'E:\TuzbeDocs\\'
today = datetime.today().strftime('%Y%m%d')

server_doc_path = os.path.join('Z:\\', 'oglastrgsud')

# download dokumenata
local_dir_path = os.path.join(main_doc_path, 'download')
local_dir_path_fina = os.path.join(main_doc_path, 'download', 'fina')

local_dir_path_day = os.path.join(main_doc_path, 'downloaddaily')
local_subdir_path_day = os.path.join(main_doc_path, 'downloaddaily', today)

server_dir_path = server_doc_path
server_dir_path_fina = os.path.join(server_doc_path, 'FINA')


class Eoglasna(CrawlerBase, ABC):
    def __init__(self, first_run=True, recheck=None):
        CrawlerBase.__init__(self, base, proxy_type=1)
        self.flag_120 = False
        self.dir_creator()

    def download_dokumenata(self):
        # download part
        self.connect(4)
        download_lista = self.select('v_tuzbe_download', view=True)
        # We can use a with statement to ensure threads are cleaned up promptly
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            # Start the load operations and mark each future with its URL
            future_to_url = {executor.submit(self.download, dl_data): dl_data for dl_data in download_lista}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    data = future.result()
                except Exception as exc:
                    print('%r generated an exception: %s' % (url, exc))
                # else:
                #     print(data)

    @staticmethod
    def filemd5(filename, block_size=2 ** 20):
        f = open(filename, errors='ignore')
        md5 = hashlib.md5()
        while True:
            data = f.read(block_size)
            if not data:
                break
            data = data.encode()
            md5.update(data)
        f.close()
        return md5.hexdigest()

    @staticmethod
    def dir_creator():
        dirs = [local_dir_path, local_dir_path_fina, local_dir_path_day, local_subdir_path_day]
        for dir_path in dirs:
            if os.path.isdir(dir_path) is False:
                os.mkdir(dir_path)

    def download(self, data):
        if self.flag_120 is True:
            return

        proc_id = threading.current_thread().ident

        id = data.get('id2')
        sud_id = data.get('sud_id')
        dokument_url = data.get('dokument_url')
        nova_oglasna = data.get('nova_oglasna')

        # get today changes
        sha = self.new_request(proc_id=proc_id, protocols=['https'], new=True)
        r = self.requests.get(proc_id)
        r.request_type = 2
        r.ses.verify = False

        try:
            # url = ''.join([base, dokument_url])
            url = dokument_url

            print(url)
            doc_data = r.go(url, download_raw=True)

            if type(doc_data) is not dict:
                if int(doc_data.status_code) > 499:
                    doc_data = {"RequestError": "Page server 500 cant reach document"}

            if type(doc_data) is dict:
                # print("download nije uspio", doc_data)
                err_desc = doc_data.get('RequestError')
                print(err_desc)
            else:
                response_headers = doc_data.headers
                filename = None
                print(response_headers)
                content_disposition, content_length, content_type = None, 0, None
                if response_headers is not None:
                    content_disposition = response_headers.get('Content-Disposition')
                    filename = re.search('name="(.*?)"', content_disposition)
                    if filename is not None:
                        filename = filename.group(1)
                    content_length = response_headers.get('Content-Length')

                doc_data = doc_data.content
                print(f"filename: {filename}, content_disposition: {filename} content_length: {content_length} content_type: {content_type}")

                if sud_id == 99999:
                    local_path = local_dir_path_fina
                    server_path = server_dir_path_fina
                else:

                    # if nova_oglasna is not None and nova_oglasna == 1:
                    local_path = os.path.join(local_dir_path, 'nova', str(sud_id))
                    server_path = os.path.join(server_dir_path, 'nova', str(sud_id))
                    # else:
                    #     local_path = os.path.join(local_dir_path, 'stara', str(sud_id))
                    #     server_path = os.path.join(server_dir_path, 'stara', str(sud_id))

                if os.path.exists(local_path) is False:
                    os.makedirs(local_path)

                if filename is not None:

                    docExtension_idx = filename.rfind(".")

                    if docExtension_idx == -1:
                        print('Nema ekstenzija ' + filename , url)

                    else:
                        docExtension = filename[docExtension_idx + 1:]
                        if len(docExtension) < 6:

                            filename_short = str(id) + '.' + docExtension

                            # local dir path
                            filename_full = os.path.join(local_path, filename_short)
                            # local dir path today
                            filename_full_today = os.path.join(local_subdir_path_day, filename_short)
                            # server dir path
                            filename_server_full = os.path.join(server_path, filename_short)

                            fileLen = len(doc_data)

                            stdout.write("\rdownloading: {}, file lenght:{}".format(filename_full, fileLen))

                            if fileLen > 266:
                                # writing files
                                # writing = [filename_full, filename_full_today, filename_server_full]
                                writing = [filename_full, filename_full_today]
                                for path in writing:
                                    with open(path, 'wb') as f:
                                        f.write(doc_data)

                                hashmd5 = self.filemd5(filename_full)
                                insert_data = {0: {'date_created': datetime.now(), 'len_download': fileLen,
                                                   'file_download': filename_short, 'httpstatus': 200,
                                                   'id_objava': id, 'hashmd5': hashmd5}}

                                print("insert_data", insert_data)

                                try:
                                    self.insert('OBJAVA_FILES', insert_data, primary_key='id_objava', schema='tuzbe')
                                except Exception as e:
                                    exc_type, exc_obj, exc_tb = sys.exc_info()
                                    fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
                                    e = '======================================\n Oglasna download ERRORS: \n +error type: %s \n +py location: %s \n +error line: %s \n +error description: %s \n======================================' % (str(exc_type), os.path.abspath(fname), exc_tb.tb_lineno, str(e))
                                    print(e)

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            error_description = str(e)
            if error_description.find('No space left on device'):
                self.flag_120 = True
            e = '======================================\n Oglasna download ERRORS: \n +error type: %s \n +py location: %s \n +error line: %s \n +error description: %s \n url: %s \n======================================' % (str(exc_type), os.path.abspath(fname), exc_tb.tb_lineno, str(e), url)
            print(e)

        finally:
            # release proxy back in use if it is not released it will be not be available for that web_base next 30min
            self.release_proxy(sha)

        sleep(random.randrange(10, 15))


if __name__ == '__main__':
    api = Eoglasna()
    api.download_dokumenata()

