import threading
from time import sleep
import concurrent.futures
from random import randrange

URLS = ['http://www.foxnews.com/',
        'http://www.cnn.com/',
        'http://europe.wsj.com/',
        'http://www.bbc.co.uk/',
        'http://some-made-up-domain.com/']

class MultiThread:


    def start_crawl(self):
        with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
            future_to_url = {executor.submit(self.crawl_data, url, 60): url for url in URLS}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    data = future.result()
                except Exception as exc:
                    print('%r generated an exception: %s' % (url, exc))
                else:
                    print('%r page is %d bytes' % (url, len(data)))

    def crawl_data(self, data, timeout):
        # print(data)
        thread_id = threading.current_thread().ident

        print(thread_id, data)
        sleep(randrange(2, 5))

        return data


if __name__ == '__main__':
    api = MultiThread()
    api.start_crawl()