from abc import ABC
import multiprocessing.pool
import multiprocessing as mp


class NoDaemonProcess(mp.Process):
    @property
    def daemon(self):
        return False

    @daemon.setter
    def daemon(self, value):
        pass


class NoDaemonContext(type(multiprocessing.get_context())):
    Process = NoDaemonProcess


class MyPool(multiprocessing.pool.Pool):
    def __init__(self, *args, **kwargs):
        kwargs['context'] = NoDaemonContext()
        super(MyPool, self).__init__(*args, **kwargs)


class ProxyServer(ABC):
    def task_handler(self, task):
        task()

    def gather(self):
        print("works")

    def run(self):
        task_set = [self.gather]
        pool = MyPool(5)
        pool.map(self.task_handler, task_set)
        pool.close()
        pool.join()


if __name__ == '__main__':
    api = ProxyServer()
    api.run()